;; *************************************************************************************** ;;
;; lisp debug v0.8  : source level debugger for lisp                                             ;;
;; Copyright (C) 1998 Marc Mertens                                                         ;;
;;                                                                                         ;;
;;     This program is free software; you can redistribute it and/or modify                ;;
;;    it under the terms of the GNU General Public License as published by                 ;;
;;    the Free Software Foundation; either version 2 of the License, or                    ;;
;;    (at your option) any later version.                                                  ;;
;;                                                                                         ;;
;;    This program is distributed in the hope that it will be useful,                      ;;
;;    but WITHOUT ANY WARRANTY; without even the implied warranty of                       ;;
;;    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the                        ;;
;;    GNU General Public License for more details.                                         ;;
;;                                                                                         ;;
;;    You should have received a copy of the GNU General Public License                    ;;
;;    along with this program; if not, write to the Free Software                          ;;
;;    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA            ;;
;;                                                                                         ;;
;; Contact me on mmertens@akam.be                                                          ;;
;; ********************************************************************************************
;;
;; The following functions must be defined
;;
;; (get-unix-env "var" "default") ;; Get unix environment var, if not defined returns default
;; (get-syntax 'symbol)           ;; Get the syntaxc info of a symbol
;; (get-process-id)               ;; Get the process if of the lisp process
;; (set-current-directory dir)    ;; Set the current directory to dir

;; ****************************************************************************
;; Get the unix environment variable
;; ****************************************************************************

(defun get-unix-env (var default)
  (cond ((sys:getenv var))
        (t default)))

;; ****************************************************************************
;; Get the process of the lisp process 
;; ****************************************************************************

(defun get-process-id ()
  (excl::getpid))

;; ****************************************************************************
;; Get the syntax info of a symbol
;; ****************************************************************************

(defun get-syntax (symbol-str)
  (let ((current-package (package-name *package*))
        (package nil)
        (name nil)
        (symb nil))
    (multiple-value-setq (package name)
                         (get-syntax-split-symbol symbol-str))
    (cond (package
           (cond ((find-package package)
                  (setf symb (find-symbol (string-downcase name) package))
                  (cond (symb
                         (get-syntax-describe symb))
                        (t "No description found")))
                 (t "Package not found")))
          (t
           (setf symb (find-symbol (string-downcase name)))
           (cond (symb
                  (get-syntax-describe symb))
                 (t "No description found"))))))

;; ****************************************************************************
;; Get a string describing the systax of a symbol
;; ****************************************************************************

(defun get-syntax-describe (symbol)
  (let ((*standard-output* (make-string-output-stream)))
    (excl::briefly-describe-symbol symbol)
    (get-output-stream-string *standard-output*)))

;; *************************************************************************
;; Try to split a string representing a symbol in the package name and
;; the symbol name
;; "package::name" ==> ("package" "name")
;; "package:name"  ==> ("package" "name")
;; "name" ==> (NIL "name")
;; ************************************************************************

(defun get-syntax-split-symbol (symbol)
  (let ((pcolumn (position #\: symbol)))
    (cond (pcolumn (cond ((and (< pcolumn (1- (length symbol)))
                               (char= (char symbol (1+ pcolumn)) #\:)) 
                          (values (subseq symbol 0 pcolumn)
                                  (subseq symbol (+ pcolumn 2))))
                         (t (values (subseq symbol 0 pcolumn) (subseq symbol
                                                                      (+ pcolumn
                                                                         1))))))
          (t (values nil symbol)))))

;; ***************************************************************************
;; Function to set the current directory, accepts a string representing the 
;; current directory in the host system
;; ***************************************************************************

(defun set-current-directory (dir)
  (setf common-lisp::*default-pathname-defaults*
        (parse-namestring dir)))

;;; **************************************************************************
;;; Listen on a stream and call a callback function if input arrives 
;;; on this stream
;;; **************************************************************************

(defun register-io-handler (stream callback)
  (mp:process-run-function "callbackprocess"
                           #'(lambda ()
                                     (loop
                                      (mp:wait-for-input-available stream)
                                      (funcall callback)))))

;(defun register-io-handler (stream callback)
;  (loop
;   (mp:wait-for-input-available stream)
;   (funcall callback)))










