/*
 * CodeDocument.java
 *
 * Created on September 30, 2001, 6:22 PM
 */

package IDE.CodeEditor;

import java.awt.*;
import java.io.*;
import java.util.*;
import javax.swing.text.*;
import javax.swing.undo.*;
import IDE.Messaging.*;
import IDE.*;
import IDE.CodeEditor.Parser.*;
import IDE.Configuration.*;
import IDE.WorkArea.*;
import IDE.Utilities.*;

/**
 *
 * @author  marc
 * @version
 */
public class CodeDocument
{
    /** File associated with this document */
    private File file;
    /**  Content of this document */
    private GapContent content=new GapContent();
    /** Contains the document lines */
    private CodeLineVector lines=new CodeLineVector();
    /** Contains the root element of this document */
    private CodeElementVector elements=new CodeElementVector();
    /** Contains the leaf elements of this document */
    private LeafVector leafs=new LeafVector(5000,(byte) 100);
    /** Undo manager for this document */
    private UndoManager undomanager=new UndoManager();
    /** Parser of the document */
    protected Parser parser;
    /** The editor of this document */
    private CodeEditor editor;
    /** Indicates if a document is readonly */
    private boolean readonly=false;
    /** Flag to indicate if the document is changed since the last save */
    private boolean modified=false;
    /** Creates new CodeDocument */
    public CodeDocument(Parser parser)
    {
        this.parser=parser;
        parser.initParser(this,lines,leafs,elements);
    }
    /** Create new CodeDocument and fill it with content of file */
    public CodeDocument(Parser parser,File file)
    {
        this(parser);
        this.file=file;
        loadFile(file);
    }
    /** Set the code editor where this document is part of */
    public void setCodeEditor(CodeEditor editor)
    {
        this.editor=editor;
        lines.setEditor(editor);
    }
    /** Set the file of this document */
    public void setFile(File file)
    {
        this.file=file;
    }
    /** Get the file of this document */
    public File getFile()
    {
        return file;
    }
    /** Get a character array containing the characters in this document */
    public char[] getChars()
    {
        Segment segment=new Segment();
        try
        {
            content.getChars(0,content.length(),segment);
            return segment.array;
        }
        catch (BadLocationException e)
        {
            return new char[0];
        }
    }
    /** Get the length of the document */
    public int length()
    {
        return content.length();
    }
    /** Create a position */
    public Position createPosition(int offset)
    {
        try
        {
            return content.createPosition(offset);
        }
        catch (BadLocationException e)
        {
            MessageHandler.report("Invalid  offset during creation of position",e);
            return null;
        }
    }
    /** Get a string from begin of length length not inclusive */
    public String getString(int begin,int length)
    {
        try
        {
            return content.getString(begin,length);
        }
        catch (BadLocationException e)
        {
            return "";
        }
    }
    /** Returns the undo manager for this document */
    public UndoManager getUndoManager()
    {
        return undomanager;
    }
    /** Set the readonly flag of this document */
    public void setReadOnly(boolean readonly)
    {
        this.readonly=readonly;
    }
    /** Set the modified flag */
    public void setModified(boolean modified)
    {
        if (this.modified!=modified)
        {
            this.modified=modified;
            WorkArea.refreshTitles();
        }
    }
    /** Get the modified flag */
    public boolean isModified()
    {
        return modified;
    }
    /** Get the parser of this document */
    public Parser getParser()
    {
        return parser;
    }
    /** Get the editor of this document */
    public CodeEditor getCodeEditor()
    {
        return editor;
    }
    /** Insert a string at a certain position in the document (with undo saving) */
    public boolean insertStringAt(String str,int pos)
    {
        if (insertStringAtImp(str,pos))
        {
            undomanager.addEdit(new UndoableInsert(str,pos));
            return true;
        }
        else
        {
            return false;
        }
    }
    /** Insert a string at a certain position in the document */
    public boolean insertStringAtImp(String str,int pos)
    {
        if (readonly || str.length()==0) return false;
        try
        {
            parser.prepareParseRemoveLinesAndLeafs(pos,pos+str.length());
            content.insertString(pos,str);
            setModified(true);
            parser.prepareParseAdjustLinesAndLeafs();
            return true;
        }
        catch (BadLocationException e)
        {
            MessageHandler.report("Insert string in wrong location",e);
            return false;
        }
    }
    /** Complete a character  if needed */
    public boolean completeCharacter(char c,int pos,int modifiers)
    {
        return parser.completeCharacter(c,pos,modifiers);
    };
    /** Remove a piece of the document with (undo saving) */
    public boolean remove(int pos,int length)
    {
        String str=getString(pos,length);
        if (removeImp(pos,length))
        {
            undomanager.addEdit(new UndoableRemove(str,pos));
            return true;
        }
        else
        {
            return false;
        }
    }
    /** Remove a piece of the document */
    public boolean removeImp(int pos,int length)
    {
        if (readonly) return false;
        try
        {
            parser.prepareParseRemoveLinesAndLeafs(pos,pos+length);
            content.remove(pos,length);
            setModified(true);
            parser.prepareParseAdjustLinesAndLeafs();
            return true;
        }
        catch (BadLocationException e)
        {
            MessageHandler.report("Remove using wrong location and length",e);
            return false;
        }
    }
    /** Replace a piece of the document by str (undoable version */
    public boolean replace(int pos,int length,String str)
    {
        if (length==0) // If length == 0 we just have a insert
            return insertStringAt(str,pos);
        if (str.length()==0) // If str is empty we just have a remove
            return remove(pos,length);
        String old=getString(pos,length);
        if (replaceImp(pos,length,str,old))
        {
            undomanager.addEdit(new UndoableReplace(pos,str,old));
            return true;
        }
        else
        {
            return false;
        }
    }
    /** Replace a piece of the document by str */
    public boolean replaceImp(int pos,int length,String str,String old)
    {
        if (readonly) return false;
        if (removeImp(pos,length))
        {
            if (insertStringAtImp(str,pos))
            {
                return true;
            }
            else
            {
                insertStringAtImp(old,pos);
                return false;
            }
        }
        else
        {
            return false;
        }
    }
    /** Clear the document, not undoable */
    private void clear()
    {
        content=new GapContent();
        lines.clear();
        leafs.clear();
        elements.clear();
    }
    /** Load the contents of a file in the document */
    public void loadFile(File file)
    {
        String str;
        char[] chars;
        try
        {
            //BufferedReader reader=new BufferedReader(new FileReader(file));
            BufferedReader reader=new BufferedReader(new InputStreamReader(new FileInputStream(file),"ISO-8859-1"));
            chars=new char[(int) file.length()];
            reader.read(chars);
            str=removeTabs(new String(chars));
            clear();
            insertStringAtImp(str,0);
            setModified(false);
        }
        catch (FileNotFoundException e)
        {
            MessageHandler.warn("File "+Util.getCanonicalFileName(file)+" does not exist");
        }
        catch (IOException e)
        {
            MessageHandler.reportAndExit("Error reading file "+Util.getCanonicalFileName(file),e);
        }
    }
    /** Save the contents of the document to a file but not mark the document as not modified */
    public void saveTemporyFile(File file)
    {
        try
        {
            //BufferedWriter writer=new BufferedWriter(new FileWriter(file));
            BufferedWriter writer=new BufferedWriter(new OutputStreamWriter(new FileOutputStream(file),"ISO-8859-1"));
            Segment segment=new Segment();
            content.getChars(0,content.length(),segment);
            writer.write(segment.array);
            writer.close();
        }
        catch (BadLocationException e)
        {
        }
        catch (IOException e)
        {
            MessageHandler.report("Error writing file "+e.getMessage(),e);
        }
    }
    /** Save the contents of the document to a file and mark the document as not modified */
    public void saveFile(File file)
    {
        saveTemporyFile(file);
        setModified(false);
    }
    /** Indentate the line */
    public void indentateLine()
    {
        int pos=editor.cursor.cursorOffset();
        CodeLine line=getLineAt(pos);
        if (line!=null)
        {
            // Find the spaces needed for indentation
            int spacesNeeded=parser.spacesForIndentation(pos);
            // Find the number of spaces in the begin of the line
            int beginLine=line.begin();
            int endLine=line.end();
            String str=getString(beginLine,endLine-beginLine);
            int nrspaces=0;
            for (nrspaces=0;nrspaces<str.length() && str.charAt(nrspaces)==' ';nrspaces++);
            int delta=spacesNeeded-nrspaces;
            if (delta>0)
            {
                // We must add delta spaces at the begin of the line
                StringBuffer spaces=new StringBuffer();
                for (int i=0;i<delta;i++)
                    spaces.append(' ');
                insertStringAt(spaces.toString(),beginLine);
                editor.cursor.setCursor(pos+delta);
            }
            else if (delta<0)
            {
                // We must remove delta spaces
                remove(beginLine,-delta);
                editor.cursor.setCursor(pos+delta);
            }
        }
    }
    /**
     * Convert a string containing tabs to a string with the tabs replaced by spaces
     */
    public static String removeTabs(String str)
    {
        int i,l,n;
        StringBuffer result;
        char c;
        int tabwidth=Configuration.tabwidth;
        l=str.length();
        result=new StringBuffer(l);
        
        for(i=0;i<l;i++)
        {
            if ((c=str.charAt(i))=='\t')
            {
                // Add n spaces to the next tab position
                n=i/tabwidth;
                n=(n+1)*tabwidth-i;
                for (int j=0;j<n;j++)
                    result.append(' ');
            }
            else
            {
                result.append(c);
            }
        };
        return result.toString();
    }
    /** Get the lines of the document */
    public CodeLineVector getLines()
    {
        return lines;
    }
    /** Get the elements of the document */
    public CodeElementVector getElements()
    {
        return elements;
    }
    /** Get the leafs of the document */
    public LeafVector getLeafs()
    {
        return leafs;
    }
    /** Get the line at a certain position */
    public CodeLine getLineAt(int pos)
    {
        int i=getIndexLineAt(pos);
        if (i==-1)
            return null;
        else
            return lines.lineAt(i);
    }
    /** Get index of a line at a certain position (return -1) if not found */
    public int getIndexLineAt(int pos)
    {
        if (lines.size==0)
            return -1;
        // Use a binary search method to find the line covering pos (first line with  pos < end of line
        CodeLine[] alines=lines.lines;
        int b=0;
        int e=lines.size-1;
        int i;
        while (b<e-1)
        {
            i=b+(e-b)/2;
            if (pos<alines[i].end())
                e=i;
            else
                b=i;
        };
        if (e>0 && pos<alines[e-1].end())
            return e-1;
        else if (pos<alines[e].end())
            return e;
        else
            return -1;
    }
    
    /** Get the index of a leaf at a certain position (return -1) if not found */
    public int getIndexLeafAt(int pos)
    {
        if (leafs.size()==0)
            return -1;
        // Use a binary search method to find the leaf covering pos (first leaf with pos < end of leaf)
        int b=0;
        int e=leafs.size()-1;
        int i;
        while (b<e-1)
        {
            i=b+(e-b)/2;
            if (pos<leafs.leafAt(i).end())
                e=i;
            else
                b=i;
        };
        if (e>0 && pos<leafs.leafAt(e-1).end())
            return e-1;
        else if (pos<leafs.leafAt(e).end())
            return e;
        else
            return -1;
    }
    /** Get the leaf at a certain position */
    public CodeLeafElement getLeafAt(int pos)
    {
        int i=getIndexLeafAt(pos);
        if (i==-1)
            return null;
        else
            return leafs.leafAt(i);
    }
    /** Get the element at a certain position */
    public CodeElement getElementAt(int pos)
    {
        CodeLeafElement leaf=getLeafAt(pos);
        if (leaf!=null)
            return leaf.getParent();
        else
            return null;
    }
    /** Get the index of the element at a certain position (return -1 if no element is found) */
    public int getIndexElementAt(int pos)
    {
        if (elements.size()==0)
            return -1;
        // Use a binary search method to find the element covering pos (first leaf with pos < end of leaf)
        int b=0;
        int e=elements.size()-1;
        int i;
        while (b<e-1)
        {
            i=b+(e-b)/2;
            if (pos<elements.codeElementAt(i).end())
                e=i;
            else
                b=i;
        };
        if (e>0 && pos<elements.codeElementAt(e-1).end())
            return e-1;
        else if (pos<elements.codeElementAt(e).end())
            return e;
        else
            return -1;
    }
    /** Get the top level element at a certain position */
    public CodeElement getToplevelElementAt(int pos)
    {
        CodeLeafElement leaf=getLeafAt(pos);
        CodeElement element;
        if (leaf!=null)
        {
            for (element=leaf.getParent();element!=null &&  element.getParent()!=null;element=element.getParent());
            return element;
        }
        else
        {
            return null;
        }
    }
    /**
     * Indentate the element at position pos
     */
    public void indentateElement(int pos)
    {
        if (editor==null)
            return;
        CodeElement element=getToplevelElementAt(pos);
        if (element!=null)
        {
            String newstr=parser.indentateElement(element);
            if (newstr!=null)
            {
                editor.replace(element.begin(),element.end()-element.begin(),newstr);
            };
        };
    }
    /**
     * Indentate all the elements of this document (must be executed on the UI thread)
     */
    public void indentateAllElements()
    {
        if (editor==null)
            return;
        int m=elements.size();
        CodeElement element;
        UpdateBuffer updatebuffer=new UpdateBuffer();
        String newstr;
        /** Change the cursor */
        Main.frame.setCursor(new Cursor(Cursor.WAIT_CURSOR));
        editor.setCursor(new Cursor(Cursor.WAIT_CURSOR));
        /** For every element see if we have a new version */
        for (int i=0;i<m;i++)
        {
            element=elements.codeElementAt(i);
            newstr=parser.indentateElement(element);
            if (newstr!=null)
                updatebuffer.addUpdate(element.begin(),element.end(),newstr);
        };
        updatebuffer.doUpdate();
        /** Change the cursor back to nomal */
        Main.frame.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        editor.setCursor(new Cursor(Cursor.DEFAULT_CURSOR));
        editor.initAndRepaintVisible();
    }
    /**
     * 
     * @return
     */
    /*
    public void updateCallers(TreeSet callers)
    {
    	parser.updateCallers(this,callers);
    }
    */
    /** Get a new instance of the UpdateBuffer*/
    public UpdateBuffer createUpdateBuffer()
    {
        return new UpdateBuffer();
    }
    /** Public inner class to do a set of non overlapping replacements in one operation */
    public class UpdateBuffer
    {
        /** Contains a list of non overlapping updates sorted in reverse order of their occurances */
        private TreeMap updates=new TreeMap();
        /** Add a new update to the list of updates */
        public void addUpdate(int begin,int end,String replace)
        {
            Update update=new Update(begin,end);
            Update last;
            SortedMap tail=updates.tailMap(update);
            /* Test if this update overlaps with a earlier update */
            if (!tail.isEmpty())
            {
                last=(Update) tail.firstKey();
                if (begin<last.end)
                {
                    MessageHandler.warn("Added overlapping update to the update buffer");
                    return;
                };
            };
            updates.put(update,replace);
        }
        /** Do the updates in the document */
        public void doUpdate()
        {
            if (updates.isEmpty())
                return;
            // We use the fact that the updates are sorted from the end of the document
            Iterator iter=updates.keySet().iterator();
            Update update;
            String replace;
            String oldstr=getString(0,content.length());
            StringBuffer newcontent=new StringBuffer(oldstr);
            // Do the updates
            while (iter.hasNext())
            {
                update=(Update) iter.next();
                replace=(String) updates.get(update);
                newcontent.replace(update.begin,update.end,replace);
            };
            // Replace the contents of the document and add undo info
            String newstr=newcontent.toString();
            undomanager.addEdit(new UndoableReplace(0,newstr,oldstr));
            clear(); // Empty the document
            insertStringAtImp(newstr,0);
        }
    }
    /** Used by UpdateBuffer to store a set of updates */
    private class Update implements Comparable
    {
        public int begin;
        public int end;
        private Update(int begin,int end)
        {
            this.begin=begin;
            this.end=end;
        }
        /*
         * Compares two Updates so that Updates are sorted based on their begin, update1<update2 if update1.begin>update2.begin
         * this causes a sort of the intervals in revers order, which is needed by UpdateBuffer
         */
        public int compareTo(Object obj)
        {
            if (obj instanceof Update)
            {
                return ((Update) obj).begin-begin;
            }
            else
            {
                return -1;
            }
        }
        
    }
    /**
     * Private classes for the undo operations
     */
    /**
     * Undo for a insert
     */
    private class UndoableInsert extends AbstractUndoableEdit
    {
        /** Save of the string inserted */
        String str;
        /** Save of the offset the string is inserted */
        int offset;
        /** Creates a undoable insert instance */
        private UndoableInsert(String str,int offset)
        {
            this.str=str;
            this.offset=offset;
        }
        /** Undo the insert */
        public void undo() throws CannotUndoException
        {
            super.undo();
            if (!removeImp(offset,str.length()))
                throw new CannotUndoException();
            
        }
        /** Redo the insert */
        public void redo() throws CannotRedoException
        {
            super.redo();
            if (!insertStringAtImp(str,offset))
                throw new CannotRedoException();
        }
    }
    /**
     * Undo for a remove
     */
    private class UndoableRemove extends AbstractUndoableEdit
    {
        /** Save of the string inserted */
        private String str;
        /** Save of the offset the string is inserted */
        private int offset;
        /** Creates a undoable remove instance */
        private UndoableRemove(String str,int offset)
        {
            this.str=str;
            this.offset=offset;
        }
        /** Undo the remove */
        public void undo() throws CannotUndoException
        {
            super.undo();
            if (!insertStringAtImp(str,offset))
            {
                throw new CannotUndoException();
            };
        }
        /**  Redo the remove */
        public void redo() throws CannotRedoException
        {
            super.redo();
            if (!removeImp(offset,str.length()))
            {
                throw new CannotRedoException();
            };
        }
    }
    /**
     * Undo for a replace
     */
    private class UndoableReplace extends AbstractUndoableEdit
    {
        /** Save of the new version of the string */
        private String newstr;
        /** Save of the old version of the string */
        private String oldstr;
        /** Save of the offset */
        private int offset;
        /** Creates a undoable replace instance */
        private UndoableReplace(int offset,String newstr,String oldstr)
        {
            this.newstr=newstr;
            this.oldstr=oldstr;
            this.offset=offset;
        }
        /** Undo the replace */
        public void undo() throws CannotUndoException
        {
            super.undo();
            String removedstr=getString(offset,newstr.length());
            if (removeImp(offset,newstr.length()))
            {
                if (!insertStringAtImp(oldstr,offset))
                {
                    insertStringAtImp(removedstr,offset);
                    throw new CannotUndoException();
                };
            }
            else
            {
                throw new CannotUndoException();
            };
        }
        /** Redo the replace */
        public void redo() throws CannotRedoException
        {
            super.redo();
            String removedstr=getString(offset,oldstr.length());
            if (removeImp(offset,oldstr.length()))
            {
                if (!insertStringAtImp(newstr,offset))
                {
                    insertStringAtImp(removedstr,offset);
                    throw new CannotRedoException();
                }
            }
            else
            {
                throw new CannotRedoException();
            };
        }
    }
}
