/*
 * CodeEditor.java
 *
 * Created on October 7, 2001, 1:04 PM
 */

package IDE.CodeEditor;

import IDE.Messaging.*;
import java.awt.*;
import java.awt.datatransfer.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.event.PopupMenuEvent;
import javax.swing.event.PopupMenuListener;

import java.util.*;
import IDE.*;
import IDE.Configuration.*;
import IDE.Interaction.*;
import IDE.Help.*;
import IDE.CodeEditor.Parser.*;
import IDE.WorkArea.*;
import IDE.Project.Repositories;
import IDE.Recording.*;
import IDE.Utilities.*;
import java.io.*;

/**
 * Editor of Code its document is a CodeDocument
 * @author  marc
 * @version
 */
public class CodeEditor extends JComponent implements FocusListener,MouseListener,MouseMotionListener,KeyListener,PopupMenuListener
{
    /** Count how many editors there are created,used for generating names */
    private static int editorcount=0;
    /** Number of the editor (used for generating names) */
    private int nr;
    /** Code pane this editor is part of */
    private CodePane codepane;
    /** Document displayed by this editor */
    private CodeDocument document;
    /** Interaction pane doing debugging for this editor, if this is different from null then this editor buffer is in debug mode */
    private IInteraction interaction;
    /** Mode associated with this document */
    private Mode mode;
    /** HighLighter for this editor */
    public CodeHighLighter highlighter;
    /** Cursor used in this editor */
    public CodeCursor cursor;
    /** Begin point of a selection */
    private int beginselection=-1;
    /** End point of selection */
    private int endselection=-1;
    /** List of CodePanes that are clones of this editor and must be repainted if something in this editor changes */
    private Vector editorclones=new Vector();
    /** Avoid that a dimension object must be created each time */
    private Dimension dim=new Dimension(0,0);
    /** Height of the line numbers */
    private int linenumberheight=Configuration.lineNumbering ? getFontMetrics(Configuration.font).getHeight() : 0;
    /** Boolean set by the KeyPressed event handler to inidicate if we must process the KeyTyped event */
    private boolean processKeyTyped=true;
    /** Creates new CodeEditor, only called when cloning editor */
    private CodeEditor()
    {
        // Used only by the cloneCodeEditor function
        setDoubleBuffered(true);
        setAutoscrolls(true);
    }
    /**
     * Create a new code editor displaying document,
     * fill it up with the content of the file file,
     * add a link to the explorer if needed and tell the
     * instance in which codepane it is displayed
     */
    public CodeEditor(CodePane codepane,CodeDocument document,File file)
    {
        this(codepane,document);
        document.loadFile(file);
        document.setFile(file);
        nr=editorcount;
        Repositories.openDocument(document);
    }
    /*
     * Create a new code editor displaying document,
     * tell the instance in which codepane it is displayed
     */
    public CodeEditor(CodePane codepane,CodeDocument document)
    {
        setDoubleBuffered(true);
        setAutoscrolls(true);
        addKeyListener(this);
        addFocusListener(this);
        this.codepane=codepane;
        this.document=document;
        highlighter=new CodeHighLighter(this);
        cursor=new CodeCursor(this);
        addMouseListener(this);
        addMouseMotionListener(this);
        // Tell the document this is the primary code editor containing the document
        if (document.getCodeEditor()==null)
            document.setCodeEditor(this);
        mode=Mode.getMode("basic");
        editorclones.add(this);
        editorcount++;
        nr=editorcount;
    }
    /** Clone this code editor */
    public CodeEditor cloneCodeEditor(CodePane codepane)
    {
        CodeEditor editor=new CodeEditor();
        editor.codepane=codepane;
        editor.document=document;
        editor.highlighter=new CodeHighLighter(editor);
        editor.cursor=new CodeCursor(editor);
        editor.addMouseListener(editor);
        editor.addMouseMotionListener(editor);
        editor.addKeyListener(editor);
        editor.mode=Mode.getMode("basic");
        editor.nr=nr;
        editorclones.addElement(editor);
        editor.editorclones=editorclones;
        return editor;
    }
    /** Cleanup a editor wiithout stopping debugging (called for editor in DebugCodePane when it is closed during disableDebugging */
    public void cleanup()
    {
        // Delete the editor from the list of clones
        editorclones.removeElement(this);
        // Stop parsing this document
        document.getParser().stopParsing();
    }
    /** Ensure that the content of this editor is saved (return false if this fails)  */
    public boolean ensureSaved()
    {
        if (document.isModified())
            if (!codepane.saveAs())
                return false;
        return true;
    }
    /** Close this editor (remove it form the list of clones)  and indicate if we have saved the modified document */
    public void closeEditor()
    {
        // Remove this editor from the list of clones
        editorclones.removeElement(this);
        // Test if documents must be closed
        if (editorclones.size()==0)
        {
            Repositories.closeDocument(document);
            // Stop parsing the associated document
            document.getParser().stopParsing();
            // Stop debugging this document
            JTabbedPane interactions=Main.getInteractions();
            int m=interactions.getTabCount();
            for (int i=0;i<m;i++)
                ((AbstractInteractionPane) interactions.getComponentAt(i)).disableDebugging(this);
        };
        codepane.close(); // Close the code pane containing this editor
    }
    /** Close all the editors associated with the given document */
    public static void closeAll(CodeDocument doc)
    {
    	CodeEditor editor=doc.getCodeEditor();
    	if (editor!=null)
    	{
    		for (int i=0;i<editor.editorclones.size();i++)
    			((CodeEditor) editor.editorclones.get(i)).closeEditor();	
    	};
    }
    /**  Move the cursor to the next leaf  */
    public void moveToNextLeaf()
    {
        Parser parser=document.getParser();
        int i=document.getIndexLeafAt(cursor.cursorOffset());
        LeafVector leafs=document.getLeafs();
        for (i=i+1;i<leafs.size() && parser.isWhiteSpace(leafs.leafAt(i).getSeparatorChar());i++);
        if (i<leafs.size())
            cursor.setCursor(leafs.leafAt(i).begin());
    }
    /**  Move the cursor to the previous leaf  */
    public void moveToPrevLeaf()
    {
        Parser parser=document.getParser();
        int pos=cursor.cursorOffset();
        int i=document.getIndexLeafAt(pos);
        LeafVector leafs=document.getLeafs();
        for (i=i-1;i>=0 && (leafs.leafAt(i).end()>=pos || parser.isWhiteSpace(leafs.leafAt(i).getSeparatorChar()));i--);
        if (i>=0)
            cursor.setCursor(leafs.leafAt(i).end());
    }
    /** Move the cursor to the begin of the current list  */
    public void moveToBeginList()
    {
        int i,m;
        int pos=cursor.cursorOffset();
        CodeElement element=document.getLeafAt(pos);
        
        while (element!=null && !element.isList())
            element=element.getParent();
        if (element!=null)
        {
            if (element.begin()==pos && pos>0)
            {
                cursor.setCursor(pos-1);
                moveToBeginList();
            }
            else
            {
                cursor.setCursor(element.begin());
            };
        }
        else
        {
            moveToPrevLeaf();
        };
    }
    /** Move the cursor  to the end of the current list*/
    public void moveToEndList()
    {
        int i,m;
        CodeElement element=document.getLeafAt(cursor.cursorOffset());
        while (element!=null && !element.isList())
            element=element.getParent();
        if (element!=null)
            cursor.setCursor(element.end());
        else
            moveToNextLeaf();
    }
    /** Mode the cursor to the begin of the line */
    public void moveToBeginLine()
    {
        CodeLine line=document.getLineAt(cursor.cursorOffset());
        if (line!=null)
            cursor.setCursor(line.begin());
    }
    /** Mode the cursor to the end of the line */
    public void moveToEndLine()
    {
        CodeLine line=document.getLineAt(cursor.cursorOffset());
        if (line!=null)
            cursor.setCursor(line.end()-1);
    }
    /** Select the current leaf */
    public void selectCurrentLeaf()
    {
        CodeLeafElement leaf=document.getLeafAt(cursor.cursorOffset());
        if (leaf!=null)
            setSelection(leaf.begin(),leaf.end());
    }
    /** Select the current element */
    public void selectCurrentElement()
    {
        CodeElement element=document.getElementAt(cursor.cursorOffset());
        if (element!=null)
        {
            element=element.getParent();
            if (element!=null)
                setSelection(element.begin(),element.end());
        };
    }
    /** Select the top level element */
    public void selectToplevelElement()
    {
        CodeElement element=document.getToplevelElementAt(cursor.cursorOffset());
        if (element!=null)
            setSelection(element.begin(),element.end());
    }
    /** Select a line */
    public void selectLine()
    {
        CodeLine line=document.getLineAt(cursor.cursorOffset());
        if (line!=null)
            setSelection(line.begin(),line.end());
    }
    /** Delete the current leaf */
    public void deleteCurrentLeaf()
    {
        CodeLeafElement leaf=document.getLeafAt(cursor.cursorOffset());
        if (leaf!=null)
        {
            int b=leaf.begin();
            remove(b,leaf.end()-b);
            cursor.setCursor(b);
        };
    }
    /** Delete the current element */
    public void deleteCurrentElement()
    {
        CodeElement element=document.getElementAt(cursor.cursorOffset());
        if (element!=null)
        {
            element=element.getParent();
            if (element!=null)
            {
                int b=element.begin();
                remove(b,element.end()-b);
                cursor.setCursor(b);
            };
        };
    }
    /** Delete the top level element */
    public void deleteToplevelElement()
    {
        CodeElement element=document.getToplevelElementAt(cursor.cursorOffset());
        if (element!=null)
        {
            int b=element.begin();
            remove(b,element.end()-b);
            cursor.setCursor(b);
        };
    }
    /** Delete a line */
    public void deleteLine()
    {
        CodeLine line=document.getLineAt(cursor.cursorOffset());
        if (line!=null)
        {
            int b=line.begin();
            remove(b,line.end()-b);
            cursor.setCursor(b);
        };
    }
    /** Widen the selection */
    public void widenSelection()
    {
        int pos=(beginselection>-1) ? beginselection : cursor.cursorOffset();
        CodeLeafElement leaf=document.getLeafAt(pos);
        if (leaf!=null)
        {
            CodeElement element=leaf.getParent();
            for (;element!=null && element.begin()>=pos;element=element.getParent());
            if (element!=null)
                setSelection(element.begin(),element.end());
        };
    }
    /** Shorten the selection */
    public void shortenSelection()
    {
        if (beginselection<0)
            return;
        int pos=cursor.cursorOffset();
        CodeElement prev=document.getLeafAt(pos);
        if (prev!=null)
        {
            CodeElement next=prev.getParent();
            for (;next!=null && next.begin()>beginselection;prev=next,next=next.getParent());
            setSelection(prev.begin(),prev.end());
        };
    }
    /** Calculates the preffered size of this document */
    public Dimension getPreferredSize()
    {
        int x,y;
        CodeLineVector lines=document.getLines();
        lines.calculateSize(dim);
        x=dim.width;
        y=dim.height;
        // Correct for the linenumbering if needed
        if (Configuration.lineNumbering)
            x=x+calculateLineNumberingWidth(lines);
        return new Dimension(x,y);
    }
    /** Get the interaction responsible for debugging */
    public IInteraction getInteraction()
    {
        return interaction;
    }
    /** Set the interaction responisble for debugging */
    public void setInteraction(IInteraction interaction)
    {
        Main.setCurrent(this);
        this.interaction=interaction;
    }
    /** Set the mode for this editor */
    public void setMode(Mode mode)
    {
        this.mode=mode;
        mode.activate();
    }
    /** Set the cursor, called from within lisp */
    public void setCursor(Long pos)
    {
        cursor.setCursor(pos.intValue());
    }
    /** Calculates the linenumbering width */
    private int calculateLineNumberingWidth(LineVector lines)
    {
        int nrdigits;
        if (Configuration.lineNumbering)
        {
            if (lines.size()>10)
                nrdigits=(int) (Math.log((double) lines.size())/Math.log(10))+1;
            else
                nrdigits=2;
            char[] nines=new char[nrdigits+2];
            for (int i=0;i<nines.length;i++)
                nines[i]='9';
            String str=new String(nines);
            return getFontMetrics(getFont()).stringWidth(str);
        }
        else
        {
            return 0;
        }
    }
    /** Get the index of the line covering a position, return -1 if not found */
    public Point  getPoint(int pos)
    {
        // Find the line covering position
        int i=document.getIndexLineAt(pos);
        CodeLineVector lines=document.getLines();
        int x,y;
        if (i<0)
        {
            return new Point(0,0);
        }
        else
        {
            y=lines.getYPositionLine(i);
            x=calculateLineNumberingWidth(lines)+lines.lineAt(i).getXPosition(this,pos);
            return new Point(x,y);
        }
    }
    /** Select a piece of the text */
    public void setSelection(int b,int e)
    {
        if (b<0 || e<=b || e>document.length())
            return;
        beginselection=b;
        endselection=e;
        highlighter.removeHighLight("30select",false);
        highlighter.addHighLightRegion("30select",Configuration.colorSelection,b,e-b,true);
    }
    /** Update the selected region */
    public void updateSelection(int newendselection)
    {
        if (beginselection>=0)
        {
            if (beginselection<endselection)
            {
                if (newendselection<beginselection)
                {
                    highlighter.removeHighLightRegion("30select",beginselection,endselection-beginselection,false);
                    highlighter.addHighLightRegion("30select",Configuration.colorSelection,newendselection,beginselection-newendselection,true);
                }
                else if (newendselection<endselection)
                    highlighter.removeHighLightRegion("30select",newendselection,endselection-newendselection,true);
                else
                    highlighter.addHighLightRegion("30select",Configuration.colorSelection,endselection,newendselection-endselection,true);
            }
            else
            {
                if (newendselection<endselection)
                    highlighter.addHighLightRegion("30select",Configuration.colorSelection,newendselection,endselection-newendselection,true);
                else if (newendselection<beginselection)
                    highlighter.removeHighLightRegion("30select",endselection,newendselection-endselection,true);
                else
                {
                    highlighter.removeHighLightRegion("30select",endselection,beginselection-endselection,false);
                    highlighter.addHighLightRegion("30select",Configuration.colorSelection,beginselection,newendselection-beginselection,true);
                };
            }
        }
        endselection=newendselection;
    }
    /** Unset the selection */
    public void unsetSelection()
    {
        if (beginselection>=0)
        {
            beginselection=-1;
            endselection=-1;
            highlighter.removeHighLight("30select");
        }
    }
    /** Remove the selected text */
    public void deleteSelectedText()
    {
        highlighter.deleteHighLightedCode("30select");
        unsetSelection();
    }
    /** Check if we have selected text */
    public boolean hasSelectedText()
    {
        return beginselection>=0;
    }
    /** copy the selected text to the system clipboard */
    public void copySelectedText()
    {
        Clipboard cb=Toolkit.getDefaultToolkit().getSystemClipboard();
        String str=highlighter.getHighLightedCode("30select");
        if (str!=null)
            cb.setContents(new StringSelection(str),null);
        unsetSelection();
    }
    
    /** Cut the selected text (copy it to the clipboard and delete it */
    public void cutSelectedText()
    {
        Clipboard cb=Toolkit.getDefaultToolkit().getSystemClipboard();
        String str=highlighter.getHighLightedCode("30select");
        if (str!=null)
            cb.setContents(new StringSelection(str),null);
        highlighter.deleteHighLightedCode("30select");
        unsetSelection();
    }
    /** Paste the selection in the document */
    public void pasteSelection()
    {
        Clipboard cb=Toolkit.getDefaultToolkit().getSystemClipboard();
        int pos=cursor.cursorOffset();
        try
        {
            String str=(String) cb.getContents(null).getTransferData(DataFlavor.stringFlavor);
            if (beginselection>=0)
            {
                replace(beginselection,Math.abs(beginselection-endselection),str);
                unsetSelection();
            }
            else
                insertStringAt(str,pos);
        }
        catch (UnsupportedFlavorException e)
        {
            // Do nothing
        }
        catch (IOException e)
        {
            // Do nothing
        }
    }
    /** Get the absolute position of a point in the editor */
    public Point getAbsolutePoint(int pos)
    {
        return SwingUtilities.convertPoint(this,getPoint(pos),Main.frame);
    }
    /** Get the line numbering width */
    public int getLineNumberingWidth()
    {
        return calculateLineNumberingWidth(document.getLines());
    }
    /**  Calculates the character position in the document based on the x,y position in the editor */
    public int getPosition(int x,int y)
    {
        CodeLineVector lines=document.getLines();
        CodeLine line;
        int i=lines.getIndexLineAfterY(y)-1;
        if (i>=0)
        {
            line=lines.lineAt(i);
            return line.begin()+line.getPosition(this,x-calculateLineNumberingWidth(lines));
        }
        else if (lines.size()>0)
        {
            line=lines.lineAt(lines.size()-1);
            return line.begin()+line.getPosition(this,x-calculateLineNumberingWidth(lines));
        }
        else
            return document.length()-1;
    }
    /** Check if this editor is the current editor */
    public boolean isCurrent()
    {
        return this==Main.currentEditor();
    }
    /** Check if this is a codeeditor which is part of a AbstractInteractionPane */
    public boolean isInteractionEditor()
    {
        return codepane instanceof AbstractInteractionPane;
    }
    /** Get the document in this editor */
    public CodeDocument getDocument()
    {
        return document;
    }
    /** Get the mode of this document */
    public Mode getMode()
    {
        CodeEditor editor=getEditorForInteraction();
        if (editor!=this && editor.interaction!=null &&
        codepane instanceof AbstractInteractionPane &&
        ((AbstractInteractionPane) codepane).getInteraction()==editor.interaction)
            // We have editor either in a watch pane or a interaction pane
            return Mode.getMode("debuginteraction");
        else
            // We have a editor buffer in the work area
            return mode;
    }
    /**
     * If this is a editor part of a interaction pane or a watch pane
     * returns the current editor in the WorkArea (if there is one) otherwise return this
     */
    public CodeEditor getEditorForInteraction()
    {
        if (codepane instanceof AbstractInteractionPane)
        {
            CodeEditor current=WorkArea.getCurrentEditor();
            if (current==null)
                return this;
            else
                return current;
        }
        else
            return this;
    }
    /** Return the codepane where this editor is displayed in */
    public CodePane getCodePane()
    {
        return codepane;
    }
    /** Return the short name name of a editor */
    public String getShortName()
    {
        String modified=document.isModified() ? " *" : "";
        if (document.getFile()!=null)
            return document.getFile().getName()+modified;
        else if (codepane instanceof AbstractInteractionPane)
            return ((AbstractInteractionPane) codepane).getName();
        else
            return "New<"+nr+">"+""+modified;
    }
    /** Return the long name name of a editor */
    public String getLongName()
    {
        String modified=document.isModified() ? " *" : "";
        if (document.getFile()!=null)
            return Util.getCanonicalFileName(document.getFile())+modified;
        else if (codepane instanceof AbstractInteractionPane)
            return ((AbstractInteractionPane) codepane).getName();
        else
            return "New<"+nr+">"+""+modified;
    }
    /** Calculates the minimum size of this document */
    public Dimension getMinimumSize()
    {
        return getPreferredSize();
    }
    /** Calculates the maximum size of this document */
    public Dimension getMaximumSize()
    {
        return getPreferredSize();
    }
    // Draw part of the document
    public void paintComponent(Graphics g)
    {
        paintComponentLines(g,0,document.getLines().size());
    }
    /* Draw all the lines between beginline and endline withing the clipBounds of g */
    private void paintComponentLines(Graphics g,int beginline,int endline)
    {
        // Calculate the area that must be painted
        int xmin,ymin;
        int xmax,ymax;
        int height;
        Rectangle bounds=g.getClipBounds();
        xmin=bounds.x;
        ymin=bounds.y;
        xmax=xmin+bounds.width;
        ymax=ymin+bounds.height;
        // Paint the different lines
        CodeLineVector lines=document.getLines();
        int y=0;
        CodeLine line;
        FontMetrics fm=getFontMetrics(getFont());
        int width=calculateLineNumberingWidth(lines);
        int ibegin=Math.max(0,lines.getIndexLineAfterY(ymin)-1);
        int iend=lines.getIndexLineAfterY(ymax);
        if (iend==-1)
            iend=endline;
        else
            iend=Math.min(lines.size,Math.min(endline,iend));
        
        for (int i=ibegin;i<iend;i++)
        {
            line=lines.lineAt(i);
            y=lines.getYPositionLine(i);
            height=lines.getHeight(this,i);
            if (ymin<=y+height && i>= beginline)
            {
                if (Configuration.lineNumbering)
                {
                    // Draw the linenumbering part of a line
                    g.setFont(Configuration.font);
                    g.setColor(Configuration.colorBackgroundLine);
                    g.fillRect(0,y,width,y+height);
                    g.setColor(Configuration.colorForegroundLine);
                    g.drawString(" "+String.valueOf(i+1),0,y+fm.getMaxAscent());
                    // Draw the line itself
                    line.paintLine(this,g,y,width,xmin,xmax,height);
                }
                else
                {
                    line.paintLine(this,g,y,0,xmin,xmax,height);
                };
            };
        }
    }
    
    public void paintRegion(int begin,int end)
    {
        // Paint region is essential repainting all the lines between the
        // lines containing begin and end inclusive the lines containing
        // begin and end
        if (getGraphics()!=null)
        {
            CodeLineVector lines=document.getLines();
            int dx=0;
            int dy;
            int x=0;
            int y;
            int ibeginline,iendline;
            int m=lines.size();
            CodeLine line;
            // Set the color of the viewport
            codepane.refreshBackground();
            // Find the first line covering begin to calculate (x,y) left upper point of the region to repaint
            ibeginline=document.getIndexLineAt(begin);
            if (ibeginline==-1)
            {
                y=0;
                dx=0;
            }
            else
            {
                y=lines.getYPositionLine(ibeginline);
                line=lines.lineAt(ibeginline);
                line.calculateSize(this,dim);
                dx=dim.width;
            };
            // Find the first line not covering end  to calculate (dx,dy)
            iendline=document.getIndexLineAt(end);
            if (iendline==-1)
            {
                if (lines.size==0)
                {
                    dy=0;
                }
                else
                {
                    line=lines.lineAt(lines.size-1);
                    line.calculateSize(this,dim);
                    dy=lines.getYPositionLine(lines.size-1)-y+dim.height;
                    dx=Math.max(dx,dim.width);
                };
            }
            else
            {
                line=lines.lineAt(iendline);
                line.calculateSize(this,dim);
                dy=lines.getYPositionLine(iendline)-y+dim.height;
                dx=Math.max(dx,dim.width);
            };
            dx=dx+calculateLineNumberingWidth(lines)+10; // The 10 takes in account the cursor at end line
            // Do the repaint
            repaint(new Rectangle(x,y,dx,dy));
        }
    }
    /** Insert a character in the underlying document */
    public void insertCharacter(char c)
    {
        if (beginselection>=0)
            deleteSelectedText();
        int pos=cursor.cursorOffset();
        document.insertStringAt(String.valueOf(c),pos);
        revalidateThisAndClones();
        highlighter.removeHighLight("20hook",false);
        repaintVisible();
        cursor.moveRight();
    }
    /** Insert a character in the underlying document */
    public void insertCharacter(char c,int modifiers)
    {
        if (beginselection>=0)
            deleteSelectedText();
        int pos=cursor.cursorOffset();
        if (!document.completeCharacter(c,pos,modifiers))
        {
            document.insertStringAt(String.valueOf(c),pos);
            revalidateThisAndClones();
            highlighter.removeHighLight("20hook",false);
            repaintVisible();
            cursor.moveRight();
        }
        else
        {
            highlighter.removeHighLight("20hook",false);
            revalidateThisAndClones();
            repaintVisible();
        };
    }
    /** Insert a string in the inderlying document */
    public void insertStringAt(String str,int pos)
    {
        insertStringAt(str,pos,true);
    }
    public void insertStringAt(String str,int pos,boolean toucheditor)
    {
        if (beginselection>=0)
            deleteSelectedText();
        document.insertStringAt(str,pos);
        if (toucheditor)
            cursor.setCursor(pos+str.length());
        else
            cursor.moveCursorTo(pos+str.length());
        revalidateThisAndClones();
        highlighter.removeHighLight("20hook",false);
        repaintVisible();
    }
    /** Remove the character at the specified  position */
    public void removeCharacter(int position)
    {
        if (beginselection>=0)
        {
            deleteSelectedText();
        }
        else
        {
            if (position>=0 && position+1 < document.length())
                document.remove(position,1);
            revalidateThisAndClones();
            highlighter.removeHighLight("20hook",false);
            repaintVisible();
        };
    }
    /** Remove a piece of text at the selected position */
    public void remove(int position,int length)
    {
        int offset=cursor.cursorOffset();
        
        document.remove(position,length);
        if (offset>position+length)
            cursor.setCursor(offset-length);
        else if (offset>position)
            cursor.setCursor(position);
        revalidateThisAndClones();
        repaintVisible();
    }
    /** Replace a piece of text at a certain position */
    public void replace(int position,int length,String str)
    {
        document.replace(position,length,str);
        cursor.setCursor(cursor.cursorOffset()-length+str.length());
        revalidateThisAndClones();
        repaintVisible();
    }
    /** Revalidate this editor and all clones of it */
    protected void revalidateThisAndClones()
    {
        int m=editorclones.size();
        for (int i=0;i<m;i++)
            ((CodeEditor) editorclones.elementAt(i)).revalidate();
    }
    /** Highlight the list element covering cursor position */
    public void highlightList()
    {
        highlighter.removeHighLight("10list",false);
        CodeElement element=document.getElementAt(cursor.cursorOffset()-1);
        while (element!=null && !element.isList())
            element=element.getParent();
        if (element!=null)
        {
            // Found the list covering the the current position
            int begin=element.begin();
            int end=element.end();
            highlighter.addHighLightRegion("10list",Configuration.colorList,begin,end-begin,true);
        };
    }
    /** Repaint the visible part of the document */
    public void repaintVisible()
    {
        int m=editorclones.size();
        for (int i=0;i<m;i++)
        {
            CodeEditor editor=(CodeEditor) editorclones.elementAt(i);
            editor.repaint(editor.getCodePane().getVisibleRectangleEditor());
        };
    }
    /** Reset the sizes of the lines and repaint the visible portion of this editor (because we have changed the fonts) */
    public void initAndRepaintVisible()
    {
        CodeLineVector lines=document.getLines();
        linenumberheight=Configuration.lineNumbering ? getFontMetrics(Configuration.font).getHeight() : 0;
        lines.recalculatePositions();
        revalidate();
        codepane.revalidate();
        repaintVisible();
    }
    /** Check if the popupmenu must be displayed */
    private boolean checkPopup(MouseEvent e)
    {
        if (e.isPopupTrigger() && !Recording.isReplaying())
        {
        	removeMouseMotionListener(this);
        	mode.getPopupMenu().addPopupMenuListener(this);
            mode.getPopupMenu().show(this,e.getX(),e.getY());
            return true;
  		 }
  		 else
  		 {
  		 	return false;
  		 }
    }
    /** Implementation of different interfaces */
    public void mouseExited(MouseEvent e)
    {
    }
    
    public void mouseReleased(MouseEvent e)
    {
        checkPopup(e);
    }
    
    public void mousePressed(MouseEvent e)
    {
        if (Main.currentEditor()!=this)
            Main.setCurrent(this);
		checkPopup(e);
		if (e.getButton()==MouseEvent.BUTTON1)
		{	
        	beginselection=-1;
        	highlighter.removeHighLight("30select");
        	cursor.setCursor(getPosition(e.getX(),e.getY()));
		};
        requestFocus();
        // Ensure that its document has itself as editor (must be done if there are clones)
        document.setCodeEditor(this);
    }
    
    public void mouseClicked(MouseEvent e)
    {
    	checkPopup(e);
    	if (e.getButton()==MouseEvent.BUTTON1)
    	{
        	beginselection=-1;
        	highlighter.removeHighLight("30select");
    	}
    }
    
    public void mouseEntered(MouseEvent e)
    {
    }
    
    public void mouseDragged(MouseEvent e)
    {
        int pos=getPosition(e.getX(),e.getY());
        cursor.setCursor(pos);
        if (beginselection<0)
        {
            beginselection=cursor.cursorOffset();
            endselection=beginselection;
        }
        else
        {
            updateSelection(pos);
        }
    }
    
    public void mouseMoved(MouseEvent e)
    {
        if (interaction!=null)
            ValueWindow.displayValueLeaf(this,e.getX(),e.getY());
    }
    
    /** Override the toString method */
    public String toString()
    {
        return getLongName();
    }
    
    public void keyPressed(KeyEvent ev)
    {
        Recording.recordKeyEventPressed(ev);
        processKeyPressed(ev,false);
    }
    /** Process the key pressed in a editor, replay indicates this is a replay of a macro, because actions are covered in the macro
     * recording we have to disable processing of actions during replay
     */
    public void processKeyPressed(KeyEvent ev,boolean replay)
    {
        processKeyTyped=true;
        MessageHandler.clear();
        int keycode=ev.getKeyCode();
        if (isCursorMovement(keycode) && !ev.isControlDown() && !ev.isAltDown())
        {
            if (beginselection<0 && ev.isShiftDown())
            {
                beginselection=cursor.cursorOffset();
                endselection=beginselection;
            }
            if (keycode==KeyEvent.VK_UP)
                cursor.moveUp();
            else if (keycode==KeyEvent.VK_DOWN)
                cursor.moveDown();
            else if (keycode==KeyEvent.VK_LEFT)
                cursor.moveLeft();
            else if (keycode==KeyEvent.VK_RIGHT)
                cursor.moveRight();
            else if (keycode==KeyEvent.VK_PAGE_UP)
                cursor.pageUp();
            else if (keycode==KeyEvent.VK_PAGE_DOWN)
                cursor.pageDown();
            if (ev.isShiftDown())
                updateSelection(cursor.cursorOffset());
            else
                unsetSelection();
            processKeyTyped=false;
            ev.consume();
        }
        else if (keycode==KeyEvent.VK_G && ev.isControlDown())
        {
            // Hard coded key to cancel a keysequence
            Main.keymap.cancel();
            processKeyTyped=false;
            ev.consume();
        }
        else if (!replay && Main.keymap.execute(keycode,ev.getModifiers()))
        {
            if (!ev.isControlDown())
                unsetSelection();
            processKeyTyped=false;
            ev.consume();
        }
        else if (ev.isAltDown() && 31 <((int) ev.getKeyChar()) & ((int) ev.getKeyChar())<127)
        {
            processKeyTyped=false;
            return; // Let the system process the alter key
        }
        else
        {
            processKeyTyped=true;
        };
    }
    /** Processd keyReleased event, we should never do anything in this otherwise we will
     * have a problem with macro processing (these events are not recorded in a macro for
     * efficiency raisons
     **/
    public void keyReleased(KeyEvent ev)
    {
        char c=DeadKeySequence.checkForDeadKeys(ev);
        if (c!='\0')
        {
            insertCharacter(c); 
            processKeyTyped=false;
        };
    }
    
    /** Process the KeyTyped Event */
    public void processKeyTyped(KeyEvent ev,boolean replay)
    {
        char c=ev.getKeyChar();
        boolean enterprocessed=false;
        if (c=='\n' && codepane instanceof AbstractInteractionPane)
            enterprocessed=((AbstractInteractionPane) codepane).processEnter();
        if (!Character.isISOControl(c) || (c=='\n' && !enterprocessed))
            insertCharacter(c,ev.getModifiers());
        if (c=='\n' && Configuration.indentateOnEnter)
            document.indentateLine();
        if (!ev.isControlDown())
            unsetSelection();
    }
    /** Processd keyType event **/
    public void keyTyped(KeyEvent ev)
    {
        if (processKeyTyped)
        {
            Recording.recordKeyEventTyped(ev);
            processKeyTyped(ev,false);
            DeadKeySequence.Restart();
        };
    }
    
    /** Check if we have a cursor movement key */
    private boolean isCursorMovement(int key)
    {
        return key==KeyEvent.VK_LEFT ||
        key==KeyEvent.VK_RIGHT ||
        key==KeyEvent.VK_UP ||
        key==KeyEvent.VK_DOWN ||
        key==KeyEvent.VK_PAGE_UP ||
        key==KeyEvent.VK_PAGE_DOWN ||
        key==KeyEvent.VK_SHIFT;
    }
    
    public void focusGained(FocusEvent focusEvent)
    {
        // Enable the cursor
        cursor.setCursor(cursor.cursorOffset());
        Main.frame.setTitle(Configuration.IDEName+" "+Configuration.IDEVersion+" ("+getShortName()+")");
    }
    
    public void focusLost(FocusEvent focusEvent)
    {
        // Ensure that the cursor is not shown anymore
        if (!(focusEvent.getOppositeComponent() instanceof CodeEditor))
           	CodeCursor.cancelCursor();
    }
    
    public void popupMenuWillBecomeVisible(PopupMenuEvent e)
    {
    }
    public void popupMenuWillBecomeInvisible(PopupMenuEvent e)
    {
		addMouseMotionListener(this);
		((JPopupMenu) e.getSource()).removePopupMenuListener(this);
    }
    public void popupMenuCanceled(PopupMenuEvent e)
    {
    	addMouseMotionListener(this);
		((JPopupMenu) e.getSource()).removePopupMenuListener(this);
    }
}

