/*
 * CodeLine.java
 * Created on September 30, 2001, 8:47 PM
 */

package IDE.CodeEditor;

import java.awt.*;

/**
 * Represents a line in the code document
 * @author  marc
 * @version
 */
public class CodeLine
{
    /** Vector of LeafElements on this line */
    private LeafVector leafs=new LeafVector();
    /** Buffer the size of the line */
    private int width=-1;
    private int height=-1;
    /** Creates new CodeLine */
    public CodeLine()
    {
    }
    /** Add leaf to line */
    public void addLeaf(CodeLeafElement leaf)
    {
        leafs.addLeaf(leaf);
    }
    /** Get the leafs of this line */
    public LeafVector getLeafs()
    {
        return leafs;
    }
    /** Get the number of leafs */
    public int getLeafCount()
    {
        return leafs.size();
    }
    /** Returns the begin offset of a line */
    public int begin()
    {
        // We assume that a line always contains at least one leafelement
        return ((CodeLeafElement) leafs.firstLeaf()).begin();
    }
    /** Returns the end offset of a line */
    public int end()
    {
        // We assume that a line always contains at least one leaflement
        return ((CodeLeafElement) leafs.lastLeaf()).end();
    }
    /** Reset the size (because we have changed the fonts) */
    protected synchronized void resetSize()
    {
        height=-1;
        width=-1;
    }
    /** Get the size of a line (as a dimension)  */
    public synchronized void calculateSize(CodeEditor editor,Dimension dim)
    {
        CodeDocument document=editor.getDocument();
        if (width==-1 || height==-1)
        {
            int x=0;
            int y=0;
            int m=leafs.size();
            FontMetrics fm;
            CodeLeafElement leaf;
            for (int i=0;i<m;i++)
            {
                leaf=leafs.leafAt(i);
                fm=editor.getFontMetrics(leaf.getFont());
                x=x+fm.stringWidth(leaf.getString(document));
                y=Math.max(y,fm.getHeight());
            }
            width=x;
            height=y;
            dim.width=x;
            dim.height=y;
        }
        else
        {
            dim.width=width;
            dim.height=height;
        }
    }
    /** Tries to find the character position at screen position x */
    public int getPosition(CodeEditor editor,int x)
    {
        CodeDocument document=editor.getDocument();
        // Search for leaf covering x
        FontMetrics fm=null;
        CodeLeafElement leaf=null;
        int w;
        int i;
        int pos;
        int m=leafs.size();
        for (w=0,i=0,pos=0;i<m && w<x;i++)
        {
            leaf=leafs.leafAt(i);
            fm=editor.getFontMetrics(leaf.getFont());
            w=w+fm.stringWidth(leaf.getString(document));
            pos=pos+leaf.end()-leaf.begin();
        };
        if (leaf==null)
        {
            return 0;
        }
        else if (i<m)
        {
            // Found leave covering x (x<=w)
            String str=leaf.getString(document);
            for (i=str.length()-1;i>=0 && w-fm.charWidth(str.charAt(i))>x;i--)
            {
                w=w-fm.charWidth(str.charAt(i));
                pos--;
            };
            if (i<0)
            {
                return 0;
            }
            else
            {
                // we have found a character starting before x
                if (((float) w-((float) fm.charWidth(str.charAt(i)))/2)>x)
                {
                    return Math.min(end()-begin()-1,Math.max(0,pos-1));
                }
                else
                {
                    return pos;
                }
            }
        }
        else
        {
            // No leave covering x
            return end()-begin()-1;
        }
    }
    /** Get the X position of character at position pos in the document */
    public int getXPosition(CodeEditor editor,int pos)
    {
        CodeDocument document=editor.getDocument();
        // Add the size of all segements ending before this pos
        int x;
        FontMetrics fm;
        int i;
        int m=leafs.size();
        CodeLeafElement leaf;
        for (i=0,x=0;i<m && (leaf=leafs.leafAt(i)).end()<=pos;i++)
        {
            fm=editor.getFontMetrics(leaf.getFont());
            x=x+fm.stringWidth(leaf.getString(document));
        };
        if (i<m)
        {
            leaf=leafs.leafAt(i);
            fm=editor.getFontMetrics(leaf.getFont());
            return x+fm.stringWidth(leaf.getString(document).substring(0,Math.max(0,pos-leaf.begin())));
        }
        else
        {
            return x;
        }
    }
    /** Draw the line */
    public void paintLine(CodeEditor editor,Graphics g,int y,int xbegin,int xmin,int xmax,int lineheight)
    {
        CodeDocument document=editor.getDocument();
        int x=xbegin;
        int m=leafs.size();
        Font font;
        Color color;
        FontMetrics fm;
        int width;
        String str;
        StringBuffer strbuffer=new StringBuffer();
        CodeLeafElement leaf;
        //for (int i=0;i<m;i++)
        int lbegin,lend;
        int i=0;
        while (i<m)
        {
            leaf=leafs.leafAt(i);
            font=leaf.getFont();
            color=leaf.getColor();
            lbegin=leaf.begin();
            lend=leaf.end();
            // Try to optimize draw by creating one big string if the font and color are the same
            if (leaf instanceof CodeLeafNewLine)
                strbuffer.append(".");
            else
                strbuffer.append(leaf.getString(document));
            for (i=i+1;i<m && (leaf=leafs.leafAt(i)).getFont()==font  && leaf.getColor()==color;i++)
            {
                if (leaf instanceof CodeLeafNewLine)
                    strbuffer.append(".");
                else
                    strbuffer.append(leaf.getString(document));
                lend=leaf.end();
            };
            str=strbuffer.toString();
            strbuffer.delete(0,strbuffer.length());
            fm=editor.getFontMetrics(font);
            width=fm.stringWidth(str);
            if (xmin<=x+width)
            {
                // Draw the background of the line 
                editor.highlighter.drawBackground(g,fm,lineheight,x,y,lbegin,lend,str);
                g.setFont(font);
                g.setColor(color);
                g.drawString(str,x,y+fm.getMaxAscent());
            };
            x=x+width;
            if (x>xmax)
                break;
        };
    };
}
