/*
 * LineVector.java
 *
 * Created on October 27, 2001, 6:05 PM
 */

package IDE.CodeEditor;

import IDE.Configuration.*;
import java.awt.*;
/**
 *  Collection of lines stored in a array which can be quickly searched via x,y position
 *  also maintains the size of the document represented by these lines
 * @author  marc
 */
public class CodeLineVector extends LineVector
{
    // Avoid creating a dimension each time
    private Dimension dim=new Dimension(0,0);
    // Array to hold the link between lineindex and Y start position of the line (the lineindex is the index)
    private int[] indexToY;
    // Editor needed for the graphical context could be null
    private CodeEditor editor=null;
    // Height of the line numbers
    private int linenumberheight=0;
    // Total height of all the lines
    private int totalheight=0;
    // Total width of all the lines (= maximumum width every reached (this could be incorrect but it only influences the scrollbar size so who cares)
    private int totalwidth=0;
    // Constructors of the LineVector
    public CodeLineVector()
    {
        this(1,(byte) 1);
    }
    public CodeLineVector(int capacity, byte increment)
    {
        lines=new CodeLine[capacity];
        indexToY=new int[capacity];
        
        this.increment=increment;
        size=0;
    }
    /** Set the editor */
    public void setEditor(CodeEditor editor)
    {
        this.editor=editor;
        linenumberheight=Configuration.lineNumbering ? editor.getFontMetrics(Configuration.font).getHeight() : 0;
    }
    /** Given a y position find the index of the first line painted after the y position (its begin > y), returns -1 if there is no such line */
    public int getIndexLineAfterY(int y)
    {
        if (y<=0)
            return 1;
        if (size==0)
            return -1;
        // Use a binary search method counting on the fact that lines are painted after each other
        int b=0;
        int e=size-1;
        int i=0;
        while (b<e-1)
        {
            i=b+(e-b)/2;
            if (indexToY[i]>y)
                e=i;
            else
                b=i;
        };
        // We have now e-1<=b<=e
        if (e>0 && indexToY[e-1]>y)
            return e-1;
        else if (indexToY[e]>y)
            return e;
        else
            return -1;
    }
    /** Get the y position of line with index  i  */
    public int getYPositionLine(int i)
    {
        return indexToY[i];
    }
    /** Get the height of a line */
    public int getHeight(CodeEditor editor,int i)
    {
        if (i+1<size)
            return indexToY[i+1]-indexToY[i];
        else
        {
            lines[i].calculateSize(editor,dim);
            return dim.height;
        }
    }
    /** Calculate the total seize of all the lines (we could overestimate the total width (performance raisons) */
    public void calculateSize(Dimension dim)
    {
        dim.height=totalheight;
        dim.width=totalwidth;
    }
    /** Add all the lines in the LineVector to this LineVector (at position i) */
    public void insertAll(int i,LineVector newlines)
    {
        if (size+newlines.size<lines.length)
        {
            // Update the lines
            System.arraycopy(lines,i,lines,i+newlines.size,size-i);
            System.arraycopy(newlines.lines,0,lines,i,newlines.size);
            // Update indexToY
            System.arraycopy(indexToY,i,indexToY,i+newlines.size,size-i);
            // Update size
            size=size+newlines.size;
        }
        else
        {
            // Update the lines
            CodeLine[] templines=new CodeLine[lines.length+newlines.size];
            System.arraycopy(lines,0,templines,0,i);
            System.arraycopy(newlines.lines,0,templines,i,newlines.size);
            System.arraycopy(lines,i,templines,i+newlines.size(),size-i);
            // Update indexToY
            int[] tempIndexToY=new int[indexToY.length+newlines.size];
            System.arraycopy(indexToY,0,tempIndexToY,0,i);
            System.arraycopy(indexToY,i,tempIndexToY,i+newlines.size(),size-i);
            // Update size
            size=size+newlines.size;
            
            lines=templines;
            indexToY=tempIndexToY;
        };
        // Now we have to update the position information if there is a editor
        if (editor!=null)
        {
            int heightnewlines=0;
            int y;
            if (i>0)
            {
                lines[i-1].calculateSize(editor,dim);
                y=indexToY[i-1]+Math.max(dim.height,linenumberheight);
            }
            else
            {
                y=0;
            };
            // Update the line positions from i to i+newlines.size()-1
            for (int j=0;j<newlines.size;j++)
            {
                newlines.lines[j].calculateSize(editor,dim);
                // Update the indexToY mapping
                indexToY[i+j]=y;
                // Adjust the positions
                y=y+Math.max(linenumberheight,dim.height);
                heightnewlines=heightnewlines+Math.max(dim.height,linenumberheight);
                // Increase the maximum size if needed
                totalwidth=Math.max(totalwidth,dim.width);
            };
            // Correct the positions further on
            for (int j=i+newlines.size();j<size;j++)
                indexToY[j]=indexToY[j]+heightnewlines;
            // Corretct the total height
            totalheight=totalheight+heightnewlines;
        };
    }
    /** Recalcualte the line positions */
    public void recalculatePositions()
    {
        // Calculates the new height of the linenumbers
        linenumberheight=Configuration.lineNumbering ? editor.getFontMetrics(Configuration.font).getHeight() : 0;
        totalheight=0;
        totalwidth=0;
        // Calculates the new positions of the lines
        int y=0;
        for (int i=0;i<size;i++)
        {
            indexToY[i]=y;
            lines[i].resetSize();
            lines[i].calculateSize(editor,dim);
            y=y+Math.max(linenumberheight,dim.height);
            totalwidth=Math.max(totalwidth,dim.width);
        };
        totalheight=y;
    }
    /** Recalculate the line positions of a range of lines */
    public void recalculatePositions(int b,int e)
    {
        if (b>=0 && b<=e && e<size && editor!=null)
        {
            // Calculate the original size of the removed lines
            int totalheight=calculateTotalHeight(b,e);
            int newtotalheight=0;
            int y=indexToY[b];
            for (int i=b;i<=e;i++)
            {
                lines[i].resetSize();
                lines[i].calculateSize(editor,dim);
                indexToY[i]=y+newtotalheight;
                newtotalheight=newtotalheight+Math.max(linenumberheight,dim.height);
                totalwidth=Math.max(totalwidth,dim.width);
            };
            int delta=newtotalheight-totalheight;
            if (delta!=0)
            {
                totalheight=totalheight+delta;
                for (int i=e+1;i<size;i++)
                    indexToY[i]=indexToY[i]+delta;
            };
        }
    }
    /** Calculate the total height of the lines from b to e, b and e inclusive */
    private int calculateTotalHeight(int b,int e)
    {
        lines[e].calculateSize(editor,dim);
        return indexToY[e]-indexToY[b]+Math.max(linenumberheight,dim.height);
    }
    /** Delete a range of lines from b (inclusive) to e exclusive */
    public void removeRangeLines(int b,int e)
    {
        if (b>=0 && b<e && e<=size && editor!=null)
        {
            // Calculate the total height of the removed lines
            int totalheight=0;
            if (editor!=null)
            {
                totalheight=calculateTotalHeight(b,e-1);
            };
            // First copy all entries higher then e to position b
            if (e<size)
            {
                System.arraycopy(lines,e,lines,b,size-e);
                System.arraycopy(indexToY,e,indexToY,b,size-e);
            };
            // Next set all the element between size-(e-b) and size to null to avoid leaks
            for (int i=size-(e-b);i<size;i++)
                lines[i]=null;
            // Update all the Y values for lines between b and size-(e-b)-1
            if (editor!=null)
            {
                for (int i=size-(e-b)-1;i>=b;i--)
                    indexToY[i]=indexToY[i]-totalheight;
                this.totalheight=this.totalheight-totalheight;
            };
            // Decrease the size
            size=size-(e-b);
        }
    }
}
