/*
 * IndentationBuffer.java
 *
 * Created on March 16, 2002, 10:17 AM
 */

package IDE.CodeEditor.Parser;

import java.util.*;
import IDE.Messaging.*;

/**
 * Buffer used for construction of the indentation code.
 * The appended strings are stored in a string buffer, but also the max line length and 
 * the number of lines are maintained. Further a stack based rollback is provided.
 * @author  marc
 */
public class IndentationBuffer
{
    /** The content of the appended strings */
    private StringBuffer content=new StringBuffer();
    /** The max length of the lines in this buffer */
    private int maxLineLength=0;
    /** The number of lines in this buffer */
    private int nrLines=1;
    /** Length of the current line */
    private int currentLineLength=0;
    /** A rollback stack */
    private Stack stack=new Stack();
    /** Creates a new instance of IndentationBuffer */
    public IndentationBuffer()
    {
    }
    /** Add a string to the indentation buffer */
    public void append(String str)
    {
        // Update maxLineLength, nrLines and currentLineLength
        int m=str.length();
        for (int i=0;i<m;i++)
        {
            if (str.charAt(i)=='\n')
            {
                maxLineLength=Math.max(maxLineLength,currentLineLength);
                currentLineLength=0;
                nrLines++;
            }
            else
            {
                currentLineLength++;
            };
        };
        maxLineLength=Math.max(maxLineLength,currentLineLength);
        content.append(str);
    }
    /** Add a new line ensuring that we not add two lines */
    public void addNewLine()
    {
        int m=content.length();
        int i;
        for (i=m-1;i>=0 && content.charAt(i)==' ';i--);
        if (i>=0 && content.charAt(i)!='\n')
        {
            content.append('\n');
            maxLineLength=Math.max(maxLineLength,currentLineLength);
            currentLineLength=0;
            nrLines++;
        };
    }
    /** Add n spaces to the buffer */
    public void addSpaces(int n)
    {
        for (int i=0;i<n;i++)
            content.append(' ');
        currentLineLength=currentLineLength+n;
        maxLineLength=Math.max(maxLineLength,currentLineLength);
    }
    /** Get the number of lines added since begin or since last savepoint */
    public int getNumberLinesAfterSavePoint()
    {
        if (stack.isEmpty())
        {
            return nrLines-1;
        }
        else
        {
            UndoInfo undoinfo=(UndoInfo) stack.peek();
            return nrLines-undoinfo.nrLines;
        }
    }
    /** Get the max line length reached since the last savepoint */
    public int getMaxLineLengthAfterSavePoint()
    {
        if (stack.isEmpty())
        {
            return maxLineLength;
        }
        else
        {
            UndoInfo undoinfo=(UndoInfo) stack.peek();
            if (undoinfo.maxLineLength<maxLineLength)
            {
                // We have increased the maximum line length since the savepoint
                return maxLineLength;
            }
            else
            {
                // Calculate what the max line length reached since the savepoint 
                // without taking in account the maxLineLength reached before the 
                // checkpoint
                int m=content.length();
                int current=undoinfo.currentLineLength;
                int max=0;
                for (int i=undoinfo.start;i<m;i++)
                {
                    if (content.charAt(i)=='\n')
                    {
                        max=Math.max(max,current);
                        current=0;
                    }
                    else
                    {
                        current++;
                    };
                };
                max=Math.max(max,current);
                return max;
            }
        }
    }
    /** Get the current line length */
    public int getCurrentLineLength()
    {
        return currentLineLength;
    }
    /** Mark a point to where we can undo */
    public void setSavePoint()
    {
        stack.push(new UndoInfo());
    }
    /*
     * Rollback to the latest savepoint and removes this savepoint,
     * Do nothing (save reporting a warning) when no savepoints are on the stack   
     */
    public void rollback()
    {
        if (stack.isEmpty())
        {
            MessageHandler.warn("Invalid rollback in Indentation Buffer");
        }
        else
        {
            UndoInfo undoinfo=(UndoInfo) stack.pop();
            content.delete(undoinfo.start,content.length());
            maxLineLength=undoinfo.maxLineLength;
            nrLines=undoinfo.nrLines;
            currentLineLength=undoinfo.currentLineLength;
        }
    }
    /** Unset the latest savePoint */
    public void commitSavePoint()
    {
        if (stack.isEmpty())
        {
            MessageHandler.warn("Invalid commit in Indentation Buffer");
        }
        else
        {
            stack.pop();
        };
    }
    /** Return the string in the indentatioin buffer */
    public String toString()
    {
        return content.toString();
    }
    /** Inner class to provide the Undo information */
    private class UndoInfo
    {
        private int start; // The position in the buffer to return to
        private int maxLineLength; // The saved maxLineLength
        private int nrLines; // The saved nr of lines
        private int currentLineLength; // The current line length
        
        /** Creates a Undo information */
        private UndoInfo()
        {
            start=content.length();
            maxLineLength=IndentationBuffer.this.maxLineLength;
            nrLines=IndentationBuffer.this.nrLines;
            currentLineLength=IndentationBuffer.this.currentLineLength;
        }
    }
}
