/*
 * Parser.java
 *
 * Created on September 30, 2001, 8:36 PM
 */

package IDE.CodeEditor.Parser;

import IDE.CodeEditor.*;
import IDE.Messaging.*;
import java.util.*;
import java.lang.reflect.*;
import javax.swing.*;
import IDE.Help.*;
import IDE.Project.*;

/**
 * Do the parsing in background of a document
 * Must be subclassed to provide the language specific parsing
 * @author marc
 * @version
 */
public abstract class Parser
{
	// Repository where this parser is created
	private IRepository repository;
    // Document to parse
    protected CodeDocument document=null;
    // Lines of the document
    private CodeLineVector lines=null;
    // Leafs of the document
    protected LeafVector leafs=null;
    // Elements of the document
    protected CodeElementVector elements=null;
    // Index in lines to insert new lines
    private int iline;
    // Index in leafs to insert new leafs
    private int ileaf;
    /** Contains the timer object to do the parsing */
    private java.util.Timer timer=new java.util.Timer();
    /** TimerTask that is dooing the parsing */
    private ParseTimerTask timertask=null;
    /** Indicate if this parser should run in assynchrone or synchrone mode  */
    private boolean async=true;
    /** Constructor to create a parser */
    public Parser(boolean async)
    {
        this.async=async;
 		repository=getRepository();
    }
    /** Must be called before using the parser */
    public final void initParser(CodeDocument document,CodeLineVector lines,LeafVector leafs,CodeElementVector elements)
    {
        this.document=document;
        this.lines=lines;
        this.leafs=leafs;
        this.elements=elements;
        if (async)
        {
            // Start a TimerTask to do the parsing
            timertask=new ParseTimerTask();
            timer.schedule(timertask,200);
        };
    }
    /** Called to stop the parsing  */
    public void stopParsing()
    {
        if (timertask!=null)
            timertask.interrupt();
        timer.cancel();
        // No further parsing is done, so clean up all the further parsing
        document=null;
        lines=null;
        leafs=null;
        elements=null;
    }
    /** Interrupt the parsing, and start a new parse task */
    public void interruptParser()
    {
        if (timertask!=null)
            timertask.interrupt();
        timertask=new ParseTimerTask();
        timer.schedule(timertask,200);
    }
    /**
     *  Remove all the lines covering de region [begin,end[ and also the first line after that region,
     *  this is called before a modification of the content of a document
     */
    public final void prepareParseRemoveLinesAndLeafs(int begin,int end)
    {
        int linecount=lines.size();
        int leafcount=leafs.size();
        int ilinebegin,ilineend;
        int ileafbegin,ileafend;
        int i;
        CodeLeafElement leaf;
        if (linecount==0)
        {
            // The document is empty no lines and leafs must be removed
            iline=0;
            ileaf=0;
        }
        else
        {
            // Find the index of first line covering begin
            i=document.getIndexLineAt(begin);
            if (i!=-1)
            {
                iline=i;
                ilinebegin=i; // index of first line wich must be deleted
                // Find the index of the first line starting after end
                for (i=i+1;i<linecount && lines.lineAt(i).begin()<=end;i++);
                ilineend=i; // index of the last line which must be deleted
                // Now find the begin and end leafs which must be deleted
                i=document.getIndexLeafAt(lines.lineAt(ilinebegin).begin());
                ileaf=i;
                ileafbegin=i;
                if (ilineend<linecount)
                {
                    leaf=lines.lineAt(ilineend).getLeafs().lastLeaf();
                    for (;i<leafcount && leafs.leafAt(i)!=leaf;i++)
                    {
                        leafs.leafAt(i).setParent(null); // Mark leaf as not in use anymore
                    };
                    ileafend=i;
                }
                else
                {
                    ileafend=leafcount;
                    for (i=ileafbegin;i<ileafend;i++)
                    {
                        leafs.leafAt(i).setParent(null);
                    };
                };
                // Delete now all the lines and leafs
                lines.removeRangeLines(ilinebegin,Math.min(linecount,ilineend+1));
                leafs.removeRangeLeafs(ileafbegin,Math.min(leafcount,ileafend+1));
            }
            else
            {
                // Begin starts after the last line and thus last leaf, no lines or leafs must be removed
                iline=lines.size();
                ileaf=lines.size();
            }
        };
    }
    /**  Add the new lines and leafs to a document, this is called after the modification */
    public void prepareParseAdjustLinesAndLeafs()
    {
        int begin,end;
        int charcount;
        LineVector newlines;
        LeafVector newleafs;
        CodeLine line;
        CodeLeafElement leaf;
        int i,b;
        String str;
        char c;
        
        charcount=document.length();
        // Get the begin and end position of the region to parse
        if (leafs.size()==0)
        {
            begin=0;
            end=charcount;
        }
        else if (ileaf<leafs.size())
        {
            if (ileaf==0)
                begin=0;
            else
                begin=leafs.leafAt(ileaf-1).end();
            end=leafs.leafAt(ileaf).begin();
        }
        else
        {
            if (ileaf==0)
                begin=0;
            else
                begin=leafs.leafAt(ileaf-1).end();
            end=charcount;
        };
        // Try to guess the correct size of the newleafs and newleafs collections
        newleafs=new LeafVector((int) ((double) (end-begin)/2.3),(byte) 100);
        newlines=new LineVector((int) ((double) (end-begin)/6.0),(byte) 100);
        // Now get the new leafs and lines
        str=document.getString(begin,end-begin);
        charcount=end-begin;
        line=new CodeLine();
        for (i=0,b=0;i<charcount;i++)
        {
            c=str.charAt(i);
            if (isSeparator(c))
            {
                // Found a separator
                if (b<i)
                {
                    // There is a token for the separator
                    leaf=new CodeLeafElement(document,begin+b);
                    leaf.setLength(i-b);
                    newleafs.addLeaf(leaf);
                    line.addLeaf(leaf);
                };
                // Create the separator leaf
                if (c=='\n')
                {
                    leaf=new CodeLeafNewLine(document,begin+i);
                    ((CodeLeafSeparator) leaf).setSeparatorChar('\n');
                    newleafs.addLeaf(leaf);
                    line.addLeaf(leaf);
                    newlines.addLine(line);
                    line=new CodeLine();
                }
                else
                {
                    leaf=new CodeLeafSeparator(document,begin+i);
                    ((CodeLeafSeparator) leaf).setSeparatorChar(c);
                    newleafs.addLeaf(leaf);
                    line.addLeaf(leaf);
                };
                b=i+1;
            };
        };
        // Add final leaf and line
        if (b<i)
        {
            leaf=new CodeLeafElement(document,begin+b);
            leaf.setLength(i-b);
            newleafs.addLeaf(leaf);
            line.addLeaf(leaf);
        };
        if (line.getLeafCount()>0)
        {
            lines.addLine(line);
        };
        // Finaaly insert the new leafs and lines
        lines.insertAll(iline,newlines);
        leafs.insertAll(ileaf,newleafs);
        // Indicate that a parse must be done
        parse();
    }
    
    /** Parse part of the document from begin to end */
    private  final void parse()
    {
        if (async)
            // Intterupt the timertask dooing the parsing and start a new parsing
            interruptParser();
        else
            // Sync mode of parsing
            new ParseTimerTask().parseLeafs();
    };
    /**
     * Delete the unneeded toplevel elements,add the new elements and tell the leafs what their parents are
     * everything must be done in the thread context of the user interface in case of async parsing
     */
    private synchronized void finishParseLeafs(final int beginremove,final int endremove,final CodeElementVector newelements)
    {
        Runnable doWork=new Runnable()
        {
            public void run()
            {
                CodeEditor editor=document.getCodeEditor();;
                int m=newelements.size();
                for (int i=0;i<m;i++)
                    newelements.codeElementAt(i).acceptChilds();
                elements.removeRangeElements(beginremove,endremove);
                elements.insertAll(beginremove,newelements);
                if (editor!=null && m>0)
                {
                    int linebegin=document.getIndexLineAt(newelements.codeElementAt(0).begin());
                    int lineend=document.getIndexLineAt(newelements.codeElementAt(m-1).begin());
                    lines.recalculatePositions(linebegin,lineend);
                    editor.repaintVisible();
                    editor.highlightList();
                };
            }
        };
        try
        {
            if (async)
                SwingUtilities.invokeAndWait(doWork);
            else
                doWork.run();
        }
        catch (InterruptedException e)
        {
        }
        catch (InvocationTargetException e)
        {
            MessageHandler.report("Problem finishParseLeafs ",e);
        };
    }
    /**
     * Based on the position in the document, calculate the hint info to be displayed for code completion
     * @param pos The position in the document to calculate the HintInfo for
     * @return A instnace of HintInfo
     */
    public synchronized HintInfo getHintInfo(int pos)
    {
    	int m,ileaf;
    	CodeLeafElement leaf;
 		// Handle case of no document
		if (document==null)
			return null;
		// Get the leaf at the position pos
		ileaf=document.getIndexLeafAt(pos);
		if (ileaf<0 || ileaf>=leafs.size())
			return null; // No leaf found
		if (ileaf>0 && isSeparator(leafs.leafAt(ileaf).getSeparatorChar()) && !(isSeparator(leafs.leafAt(ileaf-1).getSeparatorChar())))
		{
			leaf=leafs.leafAt(ileaf-1);
			pos=leaf.begin();
		}
		else
		{
			leaf=leafs.leafAt(ileaf);
		}
		// Get the list enclosing this position
		Region region=findTopHooks(pos);
		if (region==null)
			return null; // No structured element at this position
		// Get the element covering this position
		ileaf=document.getIndexLeafAt(region.begin);
		CodeElementVector elements=new CodeElementVector();
		parseElement(ileaf,elements);
		if (elements.size()==0)
			return null; // No structure covering this position
		CodeStructureElement top=(CodeStructureElement) elements.codeElementAt(0); // The top element covering this position
		// Do a anlyze of the top element to get all the variables defined
		top.analyze();
		top.acceptCodeStructureChilds(); // Ensure that the hierarchy s build for the CodeStructureElements from bottom to top
		// Now get at the code element covering this position and insert a new place holder element
		// before this (this will be used by the analyzer to know the position in the structure for anlyses)
		HintMarkerElement marker=new HintMarkerElement();
		if (!(elements.codeElementAt(0) instanceof CodeStructureElement))
			return null;
		top.insertElementAt(marker,pos);
		// Now find the syntax hints
		Vector hints=new Vector();
		if (top.findHints(marker,hints))
		{
			// We have found some hint info now let build the HintInfo element
			HintInfo info=new HintInfo();
			// Build a sorted list of the code suggestions
			m=hints.size();
			info.hints=new String[m];
			for (int i=0;i<m;i++)
				info.hints[i]=createHintInfoItem(hints.elementAt(i));
			info.hints=removeDuplicatesAndSort(info.hints);
			// Fill in the syntax info
			info.syntaxhint=getSyntaxDescriptionCall(marker,document);
			// FIll in the position and leaf information
			info.leaf=leaf;
			if (isSeparator(leaf.getSeparatorChar()))
				info.iHint=0;
			else
				info.iHint=search(info.hints,leaf.getString(document).toLowerCase());
			return info; 
		}		
		else
		{
			System.out.println("No hint found");
			return null;
		}
    }
    /**
     * Removes all the duplicates and sort the list of hints (in a repository specific way)
     * @param hints
     * @return
     */
    private String[] removeDuplicatesAndSort(String[] hints)
    {
    	SortedSet set=new TreeSet(getRepository().getHintComparator());
    	for (int i=0;i<hints.length;i++)
    		set.add(hints[i]);
    	String[] result=new String[set.size()];
    	Iterator iter=set.iterator();
    	for (int i=0;i<result.length;i++)
    		result[i]=(String) iter.next();
    	return result;
    }
    private String createHintInfoItem(Object item)
    {
    	if (item instanceof CodeElement)
    		return ((CodeElement) item).getString(document);
    	else if (item instanceof String)
    		return (String) item;
    	return item.toString();
    }
	/** Search for the nearest match of a string in a sorted array of strings */
	private int search(String[] strings,String key)
	{
		key=key.toLowerCase();
		int pos=Arrays.binarySearch(strings,key,getRepository().getHintComparator());
		return (pos>0) ? pos : Math.abs(pos)-1;
	}

	/** Find the region of the most enclosing CodeStructure element of the given position */
	private Region findTopHooks(int pos)
	{
		Region region=findHooks(pos);
		int iopen=region==null ? -1 : region.begin;
		int iclose=region==null ? -1 : region.end;
		while ((region=findHooks(iopen-1))!=null && region!=null && region.begin!=0 && region.begin<=pos && pos<=region.end)
		{
			iopen=region.begin;
			iclose=region.end;
		};
		if (region!=null && region.begin==0 && 0<=pos && pos<=region.end)
			return region;
		if (iopen <= pos && pos<=iclose)
			return new Region(iopen,iclose);
		else
			return null;
	}
   /** 
     * Parse a element and add this element to elements, 
     * returns the index of the leaf after the last leafs of this element
     * Will be repeately by the parsing framework to do the parsing
     */
    public abstract int parseElement(int ileaf,CodeElementVector elements);
    /** 
     * Check if character c is a separator, remark newline should always be a separator 
     * Will be called by the tokenizer to devide a source in tokes before parsing it
     */
    public abstract boolean isSeparator(char c);
    /** Checks if character c is a white space */
    public abstract boolean isWhiteSpace(char c);
    /**
     *  Find the opening and closing hooks (the definition of a hook depends on the parser (( ) in lisp,{} in C ...
     * returns null if no closing hooks can be found
     */
    public abstract Region findHooks(int pos);
    /**  
     * Find out the correct number of spaces that must be placed at the begin of this line  covering position pos
     * to do correct indentation.
     * Called whenever a new line is inserted (or the user presses tab) to properly indentate the line containing 
     * the cursor.
     */
    public abstract int spacesForIndentation(int pos);
    /**
     * Used to create a new parser instance of the same type as the instance where this method is called on
     * async tells if the parsing has to be done in a synchronous or assyncrhonous way.
     * Used whenever we have a giving parser instance (which contains a lot of source type specific information)
     * we need a new instance which contains the same source type specific information.
     */
    public abstract Parser cloneParser(boolean async);
    /** Get the syntax description for the macro(function) referred by name */
    public abstract String getSyntaxDescription(String name);
    /** Get the syntax description for the macro(function) containing the CodeElement element */
    public abstract String getSyntaxDescriptionCall(CodeElement element,CodeDocument document);
    /** 
     * Modifies the document associated with this parser instance so that the text referred to by leaf 
     * is replaced by the text in hint. 
     * Called by the hint system when applying a hint
     */
    public abstract void replaceLeafByHint(CodeLeafElement leaf,String hint);
    /*
     * Allows the parser to do a code completion of the character c with modifiers modifiers
     * (replace for example a ( by a ()).
     * If a completion is done true is returned, otherwise false is returned.
     */
    public abstract boolean completeCharacter(char c,int pos,int modifiers);
    /** Indentate the element given */
    public abstract String indentateElement(CodeElement element);
	/** Get a list of all the callers for this document */
	//public abstract void updateCallers(CodeDocument document,TreeSet callers);
	/** Used to retrieve the repository associated with this parser */
	public abstract IRepository getRepository();
    /** Class to define a timertask to do the parsing */
    private class ParseTimerTask extends TimerTask
    {
        private boolean interrupted=false;
        public void run()
        {
            parseLeafs();
        }
        private void interrupt()
        {
            interrupted=true;
            this.cancel();
        }
        /** Give the leafs meaining by parsing them */
        public final void parseLeafs()
        {
            int m,i;
            int ileaf;
            int iremovebegin,iremoveend;
            int ifirstnewleaf;
            int elementcount;
            CodeElement element;
            CodeLeafElement leaf;
            CodeElementVector newelements=new CodeElementVector();
            // We can never be sure that the leafs are not changed during this parse causing errors
            // In all these case we must abort the parse (the modifications will cause another parse)
            try
            {
                // Find the toplevel element and leaf from where to start the parse
                if (elements.size()==0)
                {
                    // No elements yet created
                    iremovebegin=0;
                    ileaf=0;
                    ifirstnewleaf=0;
                }
                else
                {
                    // First find the first leaf wherafter we have a new leaf (indicated by having no parent)
                    m=leafs.size();
                    for (i=0;i<m && leafs.leafAt(i).getParent()!=null;i++);
                    if (i==m) return; // No new leafs added, just return
                    ifirstnewleaf=i;
                    if (interrupted) return; //If interupted do not work further
                    // i-1 is now the last leaf before the newly added leafs
                    if (i==0)
                    {
                        ileaf=0;
                        iremovebegin=0;
                    }
                    else
                    {
                        element=leafs.leafAt(i-1).getTopParent();
                        if (element==null)
                        {
                            // No elements yet
                            ileaf=0;
                            iremovebegin=0;
                        }
                        else
                        {
                            // Check now if they are incomplete elements before this element
                            m=elements.size();
                            for (i=0;i<m && elements.codeElementAt(i)!=element && elements.codeElementAt(i).isComplete();i++);
                            iremovebegin=i;
                            if (interrupted) return;
                            // Find the first leaf belonging to this element
                            ileaf=document.getIndexLeafAt(element.begin());
                        }
                    };
                };
                if (interrupted) return;
                // Do now the parsing of the now leafs
                m=leafs.size();
                elementcount=elements.size();
                while (ileaf<m && !(leafs.leafAt(ileaf).isFirstLeaf() && ileaf>ifirstnewleaf))
                    ileaf=parseElement(ileaf,newelements);
                if (interrupted) return;
                // Now we have to remove the toplevel elements, add the new elements and tell the leafs what their parents are
                if (ileaf<m)
                {
                    element=leafs.leafAt(ileaf).getTopParent();
                    if (element==null)
                        // This can normally not happen but we want to be sure
                        iremoveend=elements.size();
                    else
                        iremoveend=elements.indexOf(element,iremovebegin);
                }
                else
                {
                    iremoveend=elements.size();
                };
                // Analyze the syntax of all new elements
                if (interrupted) return;
                m=newelements.size();
                for (i=0;i<m;i++)
                {
                    if (Thread.interrupted()) return;
                    newelements.codeElementAt(i).analyze();
                };
                // Finish the parsing this will be done in the context of the userinterface thread so it must be as short as possible
                if (interrupted) return;
                finishParseLeafs(iremovebegin,iremoveend,newelements);
                // Update filenode if needed
                if (document.getFile()!=null)
                {
                    // Tell the repository that he has to refresh its contents
                    repository.analyzeElements(elements.makeClone(),document);
                };
            }
            catch (Exception e)
            {
                // We reach this because the leaf structure has changed to much during parsing
                // We have to do nothing special here
                System.out.println(e.getMessage());
                System.out.println(e.getStackTrace());
            };
        };
    }
}
