/*
 * ActionKeyMapNew.java
 *
 * Created on May 3, 2003, 1:16 PM
 */

package IDE.Configuration;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.event.*;
import javax.swing.table.*;
import IDE.Messaging.*;

/**
 *
 * Class to hold the bindings of a action to a keysequence
 * Basic rules concerning keybindings
 * 1. Each mode has its own set of keybindings (represented as a instance of ActionKeyMap)
 * 2. A action can have only one keysequence
 * 3. Duplicate KeySequences are not allowed in a ActionKeyMap
 *
 * @author  marc
 */
public class ActionKeyMap
{
    /** Static array to keep track of all the Actions */
    private static ActionKeyMap[] actionkeymappings=new ActionKeyMap[0];
    /** Hastable associating a action with a keysequence for this ActionKeyMap */
    protected HashMap actionToKeySequence=new HashMap();
    /** Hashtable associating a keysequence with a action for this ActionKeyMap */
    protected HashMap keySequenceToAction=new HashMap();
    /** String constant used to refere to the KeySequence associated with a action */
    public static final String KEYSEQUENCE="KeySequence";
    /** Compiled version of the Mapping of a KeySequence to a action */
    protected SortedMap compiledKeySequenceToAction=null;
    /** Maintains the state when checking for mapping of a key sequence to a action */
    private SortedMap keySequenceToActionNext=null;
    /** Constructor called when a new instance of the ActionKeyMapNew is made */
    public ActionKeyMap()
    {
        // Keep track of all the keymappings that exist
        ActionKeyMap[] newactionkeymappings=new ActionKeyMap[actionkeymappings.length+1];
        System.arraycopy(actionkeymappings,0,newactionkeymappings,0,actionkeymappings.length);
        newactionkeymappings[newactionkeymappings.length-1]=this;
        actionkeymappings=newactionkeymappings;
    }
    /** Restarts the state of the current typed in keysequence */
    public void cancel()
    {
        MessageHandler.say("");
        keySequenceToActionNext=null;
    }
    /**
     * See if a action that is bounded with the current state of the typed keysequence
     * must be executed and execute this action if needed.
     * Returns : true if keys must be consumed by this call (no further processing of the key is needed)
     *           false if keys must not be consumed by this call (further processing of the key is possible)
     * Will also suggest possible completions in the messageArea
     */
    public boolean execute(int keycode,int modifiers)
    {
        if (keycode==KeyEvent.VK_CONTROL || keycode==KeyEvent.VK_ALT || keycode==KeyEvent.VK_SHIFT)
            return false; // Do not react on just the Control or Alt key press
        if (compiledKeySequenceToAction==null)
            return false;
        Object found;
        if (keySequenceToActionNext==null)
        {
            // First check on a new keysequence
            found=compiledKeySequenceToAction.get(new KeyCodeModifier(keycode,modifiers));
            if (found==null)
            {
                // There is no possible keybinding
                return false;
            }
            else if (found instanceof AbstractManagedAction)
            {
                // We have found a action bound with the keysequence so far
                MessageHandler.say("");
                ((AbstractManagedAction) found).actionPerformed(null);
                keySequenceToActionNext=null;
                return true;
            }
            else
            {
                // Found must be a sorted map of the next (key,modifier) to check
                keySequenceToActionNext=(SortedMap) found;
                suggestCompletions(keySequenceToActionNext);
                return true;
            }
        }
        else
        {
            // We had a previous match
            found=keySequenceToActionNext.get(new KeyCodeModifier(keycode,modifiers));
            if (found==null)
            {
                // No match found at all
                MessageHandler.warn("Key sequence is not linked to a action");
                keySequenceToActionNext=null;
                return true;
            }
            else if (found instanceof AbstractManagedAction)
            {
                // We have found a action bound with the keysequence so far
                MessageHandler.say("");
                ((AbstractManagedAction) found).actionPerformed(null);
                keySequenceToActionNext=null;
                return true;
            }
            else
            {
                // Found must be a sorted map of the next (key,modifier) to check
                keySequenceToActionNext=(SortedMap) found;
                suggestCompletions(keySequenceToActionNext);
                return true;
            }
        }
    };
    /** Suggests the other keys which could be pressed */
    private static void suggestCompletions(SortedMap map)
    {
        StringBuffer suggestion=new StringBuffer("|==>");
        Iterator iter=map.keySet().iterator();
        KeyCodeModifier kcm;
        Object value;
        while (iter.hasNext())
        {
            kcm=(KeyCodeModifier) iter.next();
            suggestion.append(" ");
            suggestion.append(kcm.toString());
            value=map.get(kcm);
            if (value instanceof AbstractManagedAction)
            {
                suggestion.append("=");
                suggestion.append(((AbstractManagedAction) value).getValue(AbstractAction.NAME));
            };
        };
        MessageHandler.say(suggestion.toString());
    }
    /** Get the keysequences in this binding */
    public KeySequence[] getKeySequences()
    {
        KeySequence[] keysequences=new KeySequence[keySequenceToAction.size()];
        Iterator iter=keySequenceToAction.keySet().iterator();
        for (int i=0;i<keysequences.length && iter.hasNext();i++)
            keysequences[i]=(KeySequence) iter.next();
        return keysequences;
    }
    /** Given a keysequence, retrieve the action */
    public AbstractManagedAction getAction(KeySequence seq)
    {
        return (AbstractManagedAction) keySequenceToAction.get(seq);
    }
    /** Associate a action with a KeySequence */
    public void addAction(String actionname,KeySequence seq)
    {
        AbstractManagedAction action=Mode.getAction(actionname);
        if (action==null)
        {
            MessageHandler.warn("Try to bind key sequence "+seq.toString()+" to unknown action "+actionname);
        }
        else if (action.getValue(KEYSEQUENCE)!=null)
        {
            if (seq.compareTo(action.getValue(KEYSEQUENCE))!=0)
                MessageHandler.warn("Rebinding action"+actionname+" to a new keysequence "+seq.toString());
            rebindAction(action,(KeySequence) action.getValue(KEYSEQUENCE),seq);
            if (compileSequence(seq,action))
                action.putValue(KEYSEQUENCE,seq);
            else
                MessageHandler.warn("Try to bind action "+actionname+" to amiguois key sequence "+seq.toString());
        }
        else
        {
            if (compileSequence(seq,action))
                action.putValue(KEYSEQUENCE,seq);
            else
                MessageHandler.warn("Try to bind action "+actionname+" to ambiguous keysequence "+seq.toString());
        };
        action.refreshInfo();
    };
    /** Modifies the mapping of keybindings for the different modes */
    public static void changeKeyMappings()
    {
        final JDialog dialog=new JDialog(); // Dialog box used to modify the bindings
        final JTable table=new JTable();
        final JComboBox modes=new JComboBox(Mode.getModes()); // Combobox to allow you to change the keybindings
        final KeySequenceCellEditor editor=new KeySequenceCellEditor();
        table.setDefaultRenderer(KeySequence.class,new KeySequenceCellRenderer());
        table.setDefaultRenderer(AbstractManagedAction.class,new ActionCellRenderer());
        table.setDefaultEditor(KeySequence.class,editor);
        modes.setEditable(false);
        modes.setSelectedIndex(0);
        // Refill the table with new information
        final ActionKeyMapTableModel model=new ActionKeyMapTableModel((String) modes.getSelectedItem());
        table.setModel(model);
        modes.addItemListener(new ItemListener()
        {
            public void itemStateChanged(ItemEvent ev)
            {
                if (ev.getStateChange()==ItemEvent.SELECTED)
                {
                    editor.stopCellEditing(); // Ensure last test is done
                    table.removeEditor();
                    // Refill the table with new information
                    model.updateContent((String) modes.getSelectedItem());
                };
            };
        });
        JScrollPane scrollpane=new JScrollPane(table); // Scrollpane to display the keybindings
        JButton ok=new JButton(new AbstractAction("Ok") // Button used to confirm the changes made
        {
            public void actionPerformed(ActionEvent ev)
            {
                editor.stopCellEditing(); // Ensure last test is done
                table.removeEditor();
                // Save the content of the model in the action
                model.saveContent();
                dialog.dispose();
                dialog.setVisible(false);
            }
        });
        JButton cancel=new JButton(new AbstractAction("Cancel") // Button used to cancel the changes made
        {
            public void actionPerformed(ActionEvent ev)
            {
                dialog.dispose();
                dialog.setVisible(false);
            }
        });
        JPanel buttons=new JPanel(); // Panel containing the ok and cancel button
        buttons.add(ok);
        buttons.add(cancel);
        JPanel content=new JPanel(); // Content panel of the dialog box
        content.setLayout(new BorderLayout());
        content.add(modes,BorderLayout.NORTH);
        content.add(scrollpane,BorderLayout.CENTER);
        content.add(buttons,BorderLayout.SOUTH);
        dialog.setTitle("Manage the binding of keysequences to actions (press F1 to clear KeySequence)");
        dialog.setModal(true);
        dialog.setContentPane(content);
        dialog.setSize(400,500);
        dialog.setVisible(true);
    }
    /** Rebind the action in all the other keybindings if needed  (this to ensure that a action has only one keysequence) */
    protected static void rebindAction(AbstractManagedAction action,KeySequence oldseq,KeySequence newseq)
    {
        for (int i=0;i<actionkeymappings.length;i++)
        {
            // If the action is already bound to a keysequence rebind it to the new keysequence
            if (actionkeymappings[i].actionToKeySequence.get(action)!=null)
            {
                // Remove the key sequence from the compiled key sequences
                actionkeymappings[i].removeCompiledBinding(oldseq,action);
                actionkeymappings[i].actionToKeySequence.remove(action);
                actionkeymappings[i].keySequenceToAction.remove(oldseq);
                if (actionkeymappings[i].compileSequence(newseq,action))
                {
                    actionkeymappings[i].actionToKeySequence.put(action,newseq);
                    actionkeymappings[i].keySequenceToAction.put(newseq,action);
                }
                else
                {
                    MessageHandler.warn("Try to bind action "+action.getValue(AbstractAction.NAME)+" to ambiguous keysequence "+newseq.toString());
                    
                };
            };
        };
        action.putValue(KEYSEQUENCE, newseq);
        action.refreshInfo();
    }
    /**
     * Remove a binding of a KeySequence to a action from the compiled sequence
     * Returns true or false if it succeeded or not
     */
    private void removeCompiledBinding(KeySequence seq,AbstractManagedAction action)
    {
        Stack bindings=new Stack();
        Stack kcms=new Stack();
        int [][] keymods=seq.getKeys();
        int m=seq.size();
        KeyCodeModifier kcm;
        Object found;
        SortedMap seqToAction=compiledKeySequenceToAction;
        if (m>0)
        {
            for (int i=0;i<m;i++)
            {
                kcm=new KeyCodeModifier(keymods[i][0],keymods[i][1]);
                found=seqToAction.get(kcm);
                if (found==null)
                {
                    break;
                }
                else if (i+1<m)
                {
                    if (found instanceof SortedMap)
                    {
                        bindings.push(seqToAction);
                        kcms.push(kcm);
                        seqToAction=(SortedMap) found;
                    }
                    else
                    {
                        break;
                    };
                }
                else if (i==m-1)
                {
                    if (found==action)
                    {
                        // Remove the binding of the action to the key
                        seqToAction.remove(kcm);
                        // Now check if this was the last binding (traversing up the stack removing empty sorted maps)
                        while (bindings.size()>0 && seqToAction.size()==0)
                        {
                            seqToAction=(SortedMap) bindings.pop();
                            kcm=(KeyCodeModifier) kcms.pop();
                            seqToAction.remove(kcm);
                        };
                        break;
                    }
                    else
                    {
                        break;
                    }
                };
            };
        };
    }
    /*
     * Add a KeySequence to the compiled KeySequence, returning true if succeeded
     * false if failed
     */
    protected boolean compileSequence(KeySequence seq,AbstractManagedAction action)
    {
        if (compiledKeySequenceToAction==null)
            compiledKeySequenceToAction=new TreeMap();
        SortedMap keyaction=compiledKeySequenceToAction;
        Object found;
        KeyCodeModifier kcm;
        int[][] keymods=seq.getKeys();
        int m=seq.size();
        if (m>0)
        {
            for (int i=0;i+1<m;i++)
            {
                kcm=new KeyCodeModifier(keymods[i][0],keymods[i][1]);
                found=keyaction.get(kcm);
                if (found==null)
                {
                    found=new TreeMap();
                    keyaction.put(kcm,found);
                    keyaction=(SortedMap) found;
                }
                else if (found instanceof SortedMap)
                {
                    keyaction=(SortedMap) found;
                }
                else
                {
                    return false;
                };
            };
            keyaction.put(new KeyCodeModifier(keymods[m-1][0],keymods[m-1][1]), action);
            actionToKeySequence.put(action,seq);
            keySequenceToAction.put(seq,action);
        }
        return true;
    }
}
/** Private class representing a TableModel of this action to key mapping */
class ActionKeyMapTableModel extends AbstractTableModel
{
    /** Current mode */
    private Mode mode=null;
    /** Array of the abstract actions */
    private AbstractManagedAction[] actions;
    /** Array of the keySequences bound to the actions */
    private KeySequence[] keysequences;
    /** Hashtable mapping mode to array of actions */
    private HashMap modeToActions=new HashMap();
    /** Hashtable mapping mode to array of keysequences */
    private HashMap modeToKeySequences=new HashMap();
    /** Construct the ActionKeyMapTableModel */
    protected ActionKeyMapTableModel(String modename)
    {
        updateContent(modename);
    }
    /** Set the content of the table to the keymappings of the new mode  */
    protected void updateContent(String modename)
    {
        mode=Mode.getMode(modename);
        if (mode!=null)
        {
            if (modeToActions.get(mode)==null)
            {
                // We have not yet modified this mode
                ActionKeyMap map=mode.getKeyMap();
                if (map!=null)
                {
                    // Find all the actions and there keymappings in this ActionKeyMap
                    actions=new AbstractManagedAction[Mode.getActionCount()];
                    keysequences=new KeySequence[actions.length];
                    Iterator iter=Mode.getActionIterator();
                    for (int i=0;i<actions.length && iter.hasNext();i++)
                    {
                        actions[i]=(AbstractManagedAction) iter.next();
                        keysequences[i]=(KeySequence) map.actionToKeySequence.get(actions[i]);
                    };
                    fireTableDataChanged();
                    // Save the current arrays
                    modeToActions.put(mode,actions);
                    modeToKeySequences.put(mode,keysequences);
                };
            }
            else
            {
                // We have already edited this mode
                actions=(AbstractManagedAction[]) modeToActions.get(mode);
                keysequences=(KeySequence[]) modeToKeySequences.get(mode);
                fireTableDataChanged();
            };
        }
    }
    /** Saves the content for all the edited modes */
    protected void saveContent()
    {
        Iterator iter=modeToActions.keySet().iterator();
        Mode mode;
        KeySequence[] keysequences;
        AbstractManagedAction[] actions;
        while (iter.hasNext())
        {
            mode=(Mode) iter.next();
            keysequences=(KeySequence[]) modeToKeySequences.get(mode);
            actions=(AbstractManagedAction[]) modeToActions.get(mode);
            saveContent(mode, keysequences, actions);
        };
    }
    /** Check if a keysequence is not superset of one of the existing key sequences */
    protected boolean isAmbiguousKeySequence(KeySequence seq)
    {
        for (int i=0;i<keysequences.length;i++)
            if (keysequences[i]!=null && seq!=keysequences[i] && keysequences[i].isPartOf(seq))
                return true;
        return false;
    }
    /** Check if a keysequence is not part of one of the existing key sequences */
    protected boolean formsAmbiguousKeySequence(KeySequence seq)
    {
        for (int i=0;i<keysequences.length;i++)
            if (seq!=null && seq!=keysequences[i] && seq.isPartOf(keysequences[i]))
                return true;
        return false;
    }
    /** Saves the content for the current mode, return true if successfull saved */
    private void saveContent(Mode mode,KeySequence[] keysequences,AbstractManagedAction[] actions)
    {
        // Save the old content before continuing
        ActionKeyMap map=mode.getKeyMap();
        KeySequence oldseq;
        // Try to fill up the new model
        map.compiledKeySequenceToAction=new TreeMap();
        map.actionToKeySequence=new HashMap();
        map.keySequenceToAction=new HashMap();
        for (int i=0;i<actions.length;i++)
        {
            if (keysequences[i]!=null)
            {
                oldseq=(KeySequence) actions[i].getValue(ActionKeyMap.KEYSEQUENCE);
                if (oldseq!=null)
                    ActionKeyMap.rebindAction(actions[i], oldseq, keysequences[i]);
                if (map.compileSequence(keysequences[i], actions[i]))
                {
                    actions[i].putValue(ActionKeyMap.KEYSEQUENCE, keysequences[i]);
                    actions[i].refreshInfo();
                }
                else
                {
                    // Failed to bind the action to key, display a warning message
                    MessageHandler.warn("Try to bind action "+actions[i].getValue(Mode.ACTIONNAME)+" to ambiguous key sequence "+keysequences[i].toString());
                };
            };
        }
    }
    /** Returns the number of columns in the model. A
     * <code>JTable</code> uses this method to determine how many columns it
     * should create and display by default.
     *
     * @return the number of columns in the model
     * @see #getRowCount
     *
     */
    public int getColumnCount()
    {
        return 2;
    }
    
    /** Returns the number of rows in the model. A
     * <code>JTable</code> uses this method to determine how many rows it
     * should display.  This method should be quick, as it
     * is called frequently during rendering.
     *
     * @return the number of rows in the model
     * @see #getColumnCount
     *
     */
    public int getRowCount()
    {
        return actions.length;
    }
    
    /** Returns the value for the cell at <code>columnIndex</code> and
     * <code>rowIndex</code>.
     *
     * @param	rowIndex	the row whose value is to be queried
     * @param	columnIndex 	the column whose value is to be queried
     * @return	the value Object at the specified cell
     *
     */
    public Object getValueAt(int rowIndex, int columnIndex)
    {
        if (columnIndex==0)
            return actions[rowIndex];
        else
            return keysequences[rowIndex];
    }
    /** Set the value for the cell at rowIndex and columnIndex to value */
    public void setValueAt(Object value,int rowIndex,int columnIndex)
    {
        if (columnIndex==0)
            actions[rowIndex]=(AbstractManagedAction) value;
        else
            keysequences[rowIndex]=(KeySequence) value;
    }
    /** Check if a cell is editable or not */
    public boolean isCellEditable(int rowIndex,int columnIndex)
    {
        if (columnIndex==0)
            return false;
        else
            return true;
    }
    /** Get the column name */
    public String getColumnName(int columnIndex)
    {
        if (columnIndex==0)
            return "Action";
        else
            return "Keysequence";
    }
    /** Get the column type */
    public Class getColumnClass(int columnIndex)
    {
        if (columnIndex==0)
            return AbstractManagedAction.class;
        else
            return KeySequence.class;
    }
}
/*
 * Class representing a combination of a keycode and a modifier
 */
class KeyCodeModifier implements Comparable
{
    private int keycode;
    private int modifiers;
    public KeyCodeModifier(int keycode,int modifiers)
    {
        this.keycode=keycode;
        this.modifiers=modifiers;
    }
    /** Compere on KeyCodeModifier with another one */
    public int compareTo(Object o)
    {
        if (o instanceof KeyCodeModifier)
        {
            KeyCodeModifier kcm=(KeyCodeModifier) o;
            if (kcm.modifiers==modifiers)
                if (kcm.keycode==keycode)
                    return 0;
                else
                    return keycode-kcm.keycode;
            else
                return modifiers-kcm.modifiers;
        }
        else
        {
            throw new ClassCastException("Not of type KeyCodeModifier");
        }
    }
    /** Give a string representation of the keycode and modifiers */
    public String toString()
    {
        return KeyEvent.getKeyModifiersText(modifiers)+"-"+KeyEvent.getKeyText(keycode);
    }
}
/** Renderer for Actions */
class ActionCellRenderer extends JLabel implements TableCellRenderer
{
    /**  Returns the component used for drawing the cell.  This method is
     *  used to configure the renderer appropriately before drawing.
     *
     * @param	table		the <code>JTable</code> that is asking the
     * 				renderer to draw; can be <code>null</code>
     * @param	value		the value of the cell to be rendered.  It is
     * 				up to the specific renderer to interpret
     * 				and draw the value.  For example, if
     * 				<code>value</code>
     * 				is the string "true", it could be rendered as a
     * 				string or it could be rendered as a check
     * 				box that is checked.  <code>null</code> is a
     * 				valid value
     * @param	isSelected	true if the cell is to be rendered with the
     * 				selection highlighted; otherwise false
     * @param	hasFocus	if true, render cell appropriately.  For
     * 				example, put a special border on the cell, if
     * 				the cell can be edited, render in the color used
     * 				to indicate editing
     * @param	row	        the row index of the cell being drawn.  When
     * 				drawing the header, the value of
     * 				<code>row</code> is -1
     * @param	column	        the column index of the cell being drawn
     *
     */
    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column)
    {
        setText((String) ((AbstractManagedAction) value).getValue(AbstractAction.NAME));
        return this;
    }
}
/** Renderer for Actions */
class KeySequenceCellRenderer extends JLabel implements TableCellRenderer
{
    /**  Returns the component used for drawing the cell.  This method is
     *  used to configure the renderer appropriately before drawing.
     *
     * @param	table		the <code>JTable</code> that is asking the
     * 				renderer to draw; can be <code>null</code>
     * @param	value		the value of the cell to be rendered.  It is
     * 				up to the specific renderer to interpret
     * 				and draw the value.  For example, if
     * 				<code>value</code>
     * 				is the string "true", it could be rendered as a
     * 				string or it could be rendered as a check
     * 				box that is checked.  <code>null</code> is a
     * 				valid value
     * @param	isSelected	true if the cell is to be rendered with the
     * 				selection highlighted; otherwise false
     * @param	hasFocus	if true, render cell appropriately.  For
     * 				example, put a special border on the cell, if
     * 				the cell can be edited, render in the color used
     * 				to indicate editing
     * @param	row	        the row index of the cell being drawn.  When
     * 				drawing the header, the value of
     * 				<code>row</code> is -1
     * @param	column	        the column index of the cell being drawn
     *
     */
    public Component getTableCellRendererComponent(JTable table, Object value, boolean isSelected, boolean hasFocus, int row, int column)
    {
        if (value==null)
            setText("");
        else
            setText(((KeySequence) value).toString());
        return this;
    }
}
/** Cell editor for KeySequences */
class KeySequenceCellEditor extends JTextField implements TableCellEditor,KeyListener
{
    KeySequence current=null;
    KeySequence old=null;
    int column;
    int row;
    ActionKeyMapTableModel model;
    public KeySequenceCellEditor()
    {
        setEditable(false);
        addKeyListener(this);
    }
    /** Adds a listener to the list that's notified when the editor
     * stops, or cancels editing.
     *
     * @param	l		the CellEditorListener
     *
     */
    public void addCellEditorListener(CellEditorListener l)
    {
    }
    
    /** Tells the editor to cancel editing and not accept any partially
     * edited value.
     *
     */
    public void cancelCellEditing()
    {
        current=old;
        model.setValueAt(current, row, column);
        model.fireTableDataChanged();
    }
    
    /** Returns the value contained in the editor.
     * @return the value contained in the editor
     *
     */
    public Object getCellEditorValue()
    {
        return current;
    }
    
    /**  Sets an initial <code>value</code> for the editor.  This will cause
     *  the editor to <code>stopEditing</code> and lose any partially
     *  edited value if the editor is editing when this method is called. <p>
     *
     *  Returns the component that should be added to the client's
     *  <code>Component</code> hierarchy.  Once installed in the client's
     *  hierarchy this component will then be able to draw and receive
     *  user input.
     *
     * @param	table		the <code>JTable</code> that is asking the
     * 				editor to edit; can be <code>null</code>
     * @param	value		the value of the cell to be edited; it is
     * 				up to the specific editor to interpret
     * 				and draw the value.  For example, if value is
     * 				the string "true", it could be rendered as a
     * 				string or it could be rendered as a check
     * 				box that is checked.  <code>null</code>
     * 				is a valid value
     * @param	isSelected	true if the cell is to be rendered with
     * 				highlighting
     * @param	row     	the row of the cell being edited
     * @param	column  	the column of the cell being edited
     * @return	the component for editing
     *
     */
    public Component getTableCellEditorComponent(JTable table, Object value, boolean isSelected, int row, int column)
    {
        this.model=(ActionKeyMapTableModel) table.getModel();
        this.column=column;
        this.row=row;
        current=(KeySequence) value;
        if (current!=null)
        {
            old=current.copy();
            setText(current.toString());
        }
        else
        {
            old=null;
            setText("");
        };
        table.setRowSelectionInterval(row, row);
        table.setColumnSelectionInterval(column,column);
        return this;
    }
    
    /** Asks the editor if it can start editing using <code>anEvent</code>.
     * <code>anEvent</code> is in the invoking component coordinate system.
     * The editor can not assume the Component returned by
     * <code>getCellEditorComponent</code> is installed.  This method
     * is intended for the use of client to avoid the cost of setting up
     * and installing the editor component if editing is not possible.
     * If editing can be started this method returns true.
     *
     * @param	anEvent		the event the editor should use to consider
     * 				whether to begin editing or not
     * @return	true if editing can be started
     * @see #shouldSelectCell
     *
     */
    public boolean isCellEditable(EventObject anEvent)
    {
        return true;
    }
    
    /** Invoked when a key has been pressed.
     * See the class description for {@link KeyEvent} for a definition of
     * a key pressed event.
     *
     */
    public void keyPressed(KeyEvent e)
    {
        if (e.getKeyCode()==KeyEvent.VK_F1)
        {
            current=null;
            setText("");
            model.setValueAt(current, row, column);
            model.fireTableDataChanged();
        }
        else if (!(e.getKeyCode()==KeyEvent.VK_CONTROL || e.getKeyCode()==KeyEvent.VK_ALT || e.getKeyCode()==KeyEvent.VK_SHIFT))
        {
            if (current==null)
                current=new KeySequence();
            KeySequence old=current.copy();
            current.addKey(e.getKeyCode(), e.getModifiers());
            if (model.isAmbiguousKeySequence(current))
            {
                JOptionPane.showConfirmDialog(null,"Ambiguous key sequence "+current.toString()+", reverting back to original","Wrong key sequence",JOptionPane.PLAIN_MESSAGE);
                current=old;
            };
            setText(current.toString());
            model.setValueAt(current, row, column);
            model.fireTableDataChanged();
        };
    }
    
    /** Invoked when a key has been released.
     * See the class description for {@link KeyEvent} for a definition of
     * a key released event.
     *
     */
    public void keyReleased(KeyEvent e)
    {
    }
    
    /** Invoked when a key has been typed.
     * See the class description for {@link KeyEvent} for a definition of
     * a key typed event.
     *
     */
    public void keyTyped(KeyEvent e)
    {
    }
    
    /** Removes a listener from the list that's notified
     *
     * @param	l		the CellEditorListener
     *
     */
    public void removeCellEditorListener(CellEditorListener l)
    {
    }
    
    /** Returns true if the editing cell should be selected, false otherwise.
     * Typically, the return value is true, because is most cases the editing
     * cell should be selected.  However, it is useful to return false to
     * keep the selection from changing for some types of edits.
     * eg. A table that contains a column of check boxes, the user might
     * want to be able to change those checkboxes without altering the
     * selection.  (See Netscape Communicator for just such an example)
     * Of course, it is up to the client of the editor to use the return
     * value, but it doesn't need to if it doesn't want to.
     *
     * @param	anEvent		the event the editor should use to start
     * 				editing
     * @return	true if the editor would like the editing cell to be selected;
     *    otherwise returns false
     * @see #isCellEditable
     *
     */
    public boolean shouldSelectCell(EventObject anEvent)
    {
        return true;
    }
    
    /** Tells the editor to stop editing and accept any partially edited
     * value as the value of the editor.  The editor returns false if
     * editing was not stopped; this is useful for editors that validate
     * and can not accept invalid entries.
     *
     * @return	true if editing was stopped; false otherwise
     *
     */
    public boolean stopCellEditing()
    {
        if (current!=null && model.formsAmbiguousKeySequence(current))
        {
            JOptionPane.showConfirmDialog(null,
            "Ambiguous key sequence "+current.toString(),
            "Wrong key sequence",
            JOptionPane.PLAIN_MESSAGE);
            current=old;
            model.setValueAt(old,row,column);
        };
        return true;
    }
    
}