/*
 * IInteractionPane.java
 *
 * Created on November 27, 2001, 8:31 PM
 */

package IDE.Interaction;

import java.util.*;
import IDE.*;
import IDE.WorkArea.*;
import IDE.Configuration.*;
import IDE.CodeEditor.*;
import IDE.CodeEditor.Parser.*;
import java.awt.*;
import java.io.*;
import javax.swing.*;

/**
 * Interface used to implement interaction panes, these allows you to send commands to a interaction object
 * and get the result back
 * @author  marc
 */
public abstract class AbstractInteractionPane extends CodePane
{
    /** Contains the end of the last output from the underlying process */
    protected int endPositionOutput;
    /** Interaction object which input and output we are controlling */
    protected IInteraction interaction;
    /** Constaits used to signal the process we are interaction with */
    /** SIGINT signal */
    public static final String SIGINT="-SIGINT";
    /** SIGSTOP signal */
    public static final String SIGSTOP="-SIGSTOP";
    /** SIGCONT signal */
    public static final String SIGCONT="-SIGCONT";
    /** SIGQUIT signal */
    public static final String SIGQUIT="-SIGQUIT";
    /** SIGKILL signal */
    public static final String SIGKILL="-SIGKILL";
    /** Contains a list of all the recent commands saved in the command hierarchy */
    private static CommandHistory commandHistory=new CommandHistory();
    /** Create a abstract interaction pane */
    public AbstractInteractionPane(Parser parser)
    {
        super(parser);
        getEditor().setMode(Mode.getMode("interaction"));
    };
    /** Send a signal to the underlying process */
    public void signalProcess(String signal)
    {
        interaction.signalProcess(signal);
    }
    /** Add a command to the command history */
    public static void addCommandToHistory(String command)
    {
        command=removeWhiteSpaceFromBeginEnd(command);
        if (command.length()>0)
            commandHistory.addCommand(command);
    }
    /** Replace all white spaces by space and remove begin and end white space */
    private static String removeWhiteSpaceFromBeginEnd(String command)
    {
        StringBuffer buffer=new StringBuffer();
        char c;
        for (int i=0;i<command.length();i++)
        {
            c=command.charAt(i);
            if (c!='\n' && c!='\r' && c!='\t' && c!=' ')
                buffer.append(c);
            else
                buffer.append(' ');
        };
        return buffer.toString().trim();
    }
    /** Let a user choose a command and execute it in the interaction pane */
    public void chooseCommand()
    {
        CodeEditor editor=getEditor();
        Point point;
        if (editor==null)
        {
            point=null;
        }
        else
        {
            point=editor.getAbsolutePoint(editor.cursor.cursorOffset());
            SwingUtilities.convertPointToScreen(point,Main.frame);
        };
        String command=(String) ListOfChoices.showList("Command History",commandHistory.reverse(),point);
        if (command!=null)
            insertCommand(command);
    }
    /* Find the last command entered in the command pane searching from the end of the last output */
    protected String findLastCommand()
    {
        CodeDocument document=getEditor().getDocument();
        LeafVector leafs=document.getLeafs();
        Parser parser=document.getParser();
        int ileaf;
        int ielement;
        int m;
        if (document.length()<=endPositionOutput)
            return findLastCommandFromEnd();
        else
        {
            // Skip the index of the leaf covering the end of the last output
            ileaf=document.getIndexLeafAt(endPositionOutput);
            if (ileaf<-1)
                return null;
            // Parse now all the elements from this position up to the end of the document
            CodeElementVector elements=new CodeElementVector();
            m=leafs.size();
            while (ileaf<m)
                ileaf=parser.parseElement(ileaf,elements);
            StringBuffer buffer=new StringBuffer();
            m=elements.size();
            for (int i=0;i<m;i++)
                if (elements.codeElementAt(i).isComplete())
                    buffer.append(elements.codeElementAt(i).getString(document));
                else
                    return findLastCommandFromEnd();
            return buffer.toString();
        }
    }
    /** Get the last command entered in this command pane searching from the end */
    private String findLastCommandFromEnd()
    {
        CodeDocument document=getEditor().getDocument();
        LeafVector leafs=document.getLeafs();
        Parser parser=document.getParser();
        if (leafs.size()==0) return null;
        // Find the first non white space leaf from the end
        int ileaf;
        int inonwhite;
        for (ileaf=leafs.size()-1;ileaf>=0 && parser.isWhiteSpace(leafs.leafAt(ileaf).getSeparatorChar());ileaf--);
        if (ileaf<0 || ileaf>=leafs.size())
            return null;
        inonwhite=ileaf;
        // Now check if this non white space is part of a complete toplevel element
        CodeElement element=document.getToplevelElementAt(leafs.leafAt(ileaf).begin());
        if (element!=null)
        {
            if (element.isComplete())
                return element.getString(document); // Found the last command return it
            else
                return null;
        }
        else
        {
            // The parsing of the last part is not yet completed, so go to the last unparsed piece
            for (;ileaf>=0 && leafs.leafAt(ileaf).getParent()==null;ileaf--);
            ileaf=Math.max(0,ileaf);
            element=leafs.leafAt(ileaf).getTopParent();
            if (element==null)
                ileaf=0;
            else
                ileaf=document.getIndexLeafAt(element.end());
            CodeElementVector elements=new CodeElementVector();
            while (ileaf<=inonwhite)
                ileaf=parser.parseElement(ileaf,elements);
            int m=elements.size();
            if (m>1 && (element=elements.codeElementAt(m-1)).isComplete())
                return element.getString(document);
            else
                return null;
        }
    }
    /** Move the cursor to the end of the document and mark the new end */
    public void moveToEnd()
    {
        getEditor().cursor.moveToEnd();
        endPositionOutput=getEditor().cursor.cursorOffset();
    }
    /** Move the cursor to the end of the document */
    public void moveCursorToEnd()
    {
        getEditor().cursor.moveToEnd();
    }
    /** Save the state of the interaction pane */
    public static void  saveState(Writer writer)
    {
        XMLConfigClass cls=new XMLConfigClass("IDE.Interaction.AbstractInteractionPane");
        XMLConfigMethod method;
        if (commandHistory!=null)
        {
            int m=commandHistory.size();
            for (int i=0;i<m;i++)
            {
                method=new XMLConfigMethod("addCommandToHistory");
                method.addArgument((String) commandHistory.elementAt(i));
                cls.addMethod(method);
            };
            cls.serialize(writer);
        };
    }
    /** Send a command containing a return code to the interaction process */
    public void sendReturn()
    {
        // Work around for a CLISP bug
        if (interaction instanceof LispClispInteraction && Configuration.clispBug)
            return;
        // Send the return
        if (interaction!=null)
            interaction.sendCommand("\n");
    }    
    /** Return the interaction object this pane is interacting with */
    public abstract IInteraction getInteraction();
    /** Create a DebugCodePane containing the code and enable debugging for this code */
    public abstract void enableDebugging(String code);
    /** Enable debugging for document in CodeEditor*/
    public abstract void enableDebugging(CodeEditor editor);
    /** Disable debugging for document in CodeEditor */
    public abstract void disableDebugging(CodeEditor editor);
    /** Process the enter key */
    public abstract boolean processEnter();
    /**  Execute the String command */
    public abstract boolean executeCommand(String command);
    /** Insert the String command */
    public abstract void insertCommand(String command);
    /** Cleanup the interaction pane */
    public abstract void cleanup();
    /** Get the name of the interaction pane to be used */
    public abstract String getName();
}
/** class to contain the commandhistory */
class CommandHistory extends Vector
{
    protected void addCommand(String command)
    {
        if (size()>0 && command.compareTo((String) elementAt(size()-1))==0)
            return;
        if (size()>Configuration.commandHistoryLength+10)
            removeRange(0,9);
        add(command);
    }
    protected Vector reverse()
    {
        int m=size();
        Vector result=new Vector(m);
        for (int i=m-1;i>=0;i--)
            result.add(elementAt(i));
        return result;
    }
}
