/*
 * InteractionPane.java
 *
 * Created on November 7, 2001, 8:45 PM
 */

package IDE.Interaction;

import java.util.*;
import javax.swing.*;
import IDE.*;
import IDE.Messaging.*;
import IDE.WorkArea.*;
import IDE.CodeEditor.*;
import IDE.CodeEditor.Parser.*;
import IDE.Configuration.*;

/**
 * Panel to display the input and output panel for the interaction
 * @author  marc
 */
public class InteractionPane extends AbstractInteractionPane
{
    /** Indicate we must stop reading the lisp output */
    private boolean stopreading=false;
    /** List of watchpanes associated with this watch pane */
    private Vector watchpanes=new Vector();
    /** Thread to listen for output */
    private Thread outputThread;
    /** Creates new InteractionPane */
    public InteractionPane(Parser interactionparser,IInteraction interaction)
    {
        super(interactionparser);
        this.interaction=interaction;
        this.interaction.setInteractionPane(this);
        Runnable listenForOutput=new Runnable()
        {
            public void run()
            {
                listenOutput();
            }
        };
        // Starts a thread to start the process and starts listening for ouput
        outputThread=new Thread(listenForOutput);
        outputThread.setPriority(Thread.MIN_PRIORITY);
        outputThread.start();
    }
    /** Create a watchpane for this interaction pane */
    public WatchPane createWatchPane()
    {
        return new WatchPane(this);
    }
    /** Remove this watch pane */
    public void removeWatchPane(WatchPane watch)
    {
        watchpanes.remove(watch);
    }
    /** Create a watch pane, associated with this interaction pane */
    protected void addWatchPane(WatchPane watchpane)
    {
        watchpanes.add(watchpane);
    }
    /** Get the interaction object */
    public IInteraction getInteraction()
    {
        return interaction;
    }
    /** Get the name of this interaction pane */
    public String getName()
    {
        return interaction.getName();
    }
    /** Start the process and listen for output */
    private void listenOutput()
    {
        String output;
        // Start the process to listen to
        interaction.startProcess();
        try
        {
            while (true)
            {
                if (stopreading)
                    return;
                if (interaction==null)
                    return; // Indicates the closing of the interactionpane
                output=interaction.receiveErrorAndOutput();
                if (output==null)
                    break; // Indicates the closing of the interaction process
                if (output.length()>0)
                    outputString(output);
            };
        }
        catch (Exception e)
        {
            MessageHandler.report("Error dring start of the process "+e.getMessage(),e);
        };
        // There was a error
        cleanup();
    }
    /** Insert a string in the output pane */
    protected void outputString(final String output)
    {
        Runnable dowork=new Runnable()
        {
            public void run()
            {
                getEditor().insertStringAt(output,Math.max(0,getEditor().getDocument().length()-1),false);
                moveToEnd();
            }
        };
        SwingUtilities.invokeLater(dowork);
    }
    /** Process the press of the enter button */
    public boolean processEnter()
    {
        Parser parser=getEditor().getDocument().getParser();
        // Find the element to send to the interaction system
        CodeDocument document=getEditor().getDocument();
        String command=command=findLastCommand();
        if (command==null || (command.length()==1 && command.charAt(0)=='\n'))
        {
            getEditor().insertCharacter('\n',0);
            moveCursorToEnd();
            return true;
        }
        else if (command!=null)
        {
            if (Configuration.clispBug)
                command=removeLeadingWhiteSpaces(command);
            // Send the command to the lisp system for processing
            if (command.length()>0)
            {
                addCommandToHistory(command);
                interaction.sendCommand(command);
                moveToEnd();
                return true;
            }
            else
            {
                return false;
            }
        }
        else
        {
            return false;
        }
    }
    /** Remove leading characters, this to resolve a bug in CLISP */
    private String removeLeadingWhiteSpaces(String command)
    {
        StringBuffer buffer=new StringBuffer();
        int i=0;
        for (i=0;i<command.length() && (command.charAt(i)=='\n' || command.charAt(i)==' ' || command.charAt(i)=='\t' || command.charAt(i)=='\r');i++);
        for (;i<command.length();i++)
            buffer.append(command.charAt(i));
        return buffer.toString();
    }
    /** Enable the debugging mode for a document */
    public void enableDebugging(final CodeEditor editor)
    {
        final Runnable updatedocument=new Runnable()
        {
            public void run()
            {
                editor.setInteraction(interaction);
                editor.setMode(Mode.getMode("debug"));
            }
        };
        Thread thread=new Thread(new Runnable()
        {
            public void run()
            {
                if (interaction.enableDebugging(editor))
                    SwingUtilities.invokeLater(updatedocument);
            }
        });
        thread.start();
    }
    /** Enable debugging mode for a fragment of code  */
    public void enableDebugging(String code)
    {
        final Parser parser=getEditor().getDocument().getParser().cloneParser(true);
        final DebugCodePane pane=new DebugCodePane(parser,code);
        final Runnable updatedocument=new Runnable()
        {
            public void run()
            {
                CodeEditor editor=pane.getEditor();
                editor.setInteraction(interaction);
                editor.getDocument().setModified(false);
                editor.setMode(Mode.getMode("debug"));
                if (!Configuration.instrumentDebugAfter && !Configuration.instrumentCheckError)
                {
                    Mode.getAction("stopafter").setEnabled(false);
                    Mode.getAction("setnewreturnvalue").setEnabled(false);
                }
                else
                {
                    Mode.getAction("stopafter").setEnabled(true);
                    Mode.getAction("setnewreturnvalue").setEnabled(true);
                };
                WorkArea.addPane(pane);
            }
        };
        Thread thread=new Thread(new Runnable()
        {
            public void run()
            {
                if (interaction.enableDebugging(pane.getEditor()))
                    SwingUtilities.invokeLater(updatedocument);
            }
        });
        thread.start();
    }
    /** Disable the debugging mode for a document */
    public void disableDebugging(final CodeEditor editor)
    {
        final Runnable updatedocument=new Runnable()
        {
            public void run()
            {
                if (editor.getCodePane() instanceof DebugCodePane)
                {
                    WorkArea.removeWorkPaneContaining(editor);
                    editor.cleanup();
                }
                else
                {
                    editor.setInteraction(null);
                    editor.setMode(Mode.getMode("basic"));
                };
            }
        };
        final Runnable removeHighLight=new Runnable()
        {
            public void run()
            {
                editor.highlighter.removeHighLight("40breakpoint");
                editor.highlighter.removeHighLight("50nextstop");
                editor.highlighter.removeHighLight("70executedcode"); 
            }
        };
        Thread thread=new Thread(new Runnable()
        {
            public void run()
            {
                SwingUtilities.invokeLater(removeHighLight);
                interaction.disableDebugging(editor);
                SwingUtilities.invokeLater(updatedocument);
            }
        });
        thread.start();
    }
    /** Cleanup the interaction pane */
    public void cleanup()
    {
        final IInteraction saveinteraction=interaction;
        final Runnable cleanupPane=new Runnable()
        {
            public void run()
            {
                // Remove the watchnodes making use of this interaction object
                Main.getWatches().removeWatchPointsForInteraction(saveinteraction);
                // Remove all the watch panes
                for (int i=0;i<watchpanes.size();i++)
                    Main.getInteractions().remove((WatchPane) watchpanes.elementAt(i));
                // Stop reading output from lisp
                stopreading=true;
                // Stop the parsing of the code document
                getEditor().getDocument().getParser().stopParsing();
                // Remove this editor pane from the list
                Main.getInteractions().remove(InteractionPane.this);
            }
        };
        Thread thread=new Thread(new Runnable()
        {
            public void run()
            {
                // Stop the lispprocess and remove the interaction object
                if (interaction!=null)
                {
                    interaction.stopProcess();
                    interaction=null;
                };
                // Cleanup the rest
                SwingUtilities.invokeLater(cleanupPane);
            }
        });
        thread.start();
    }
    /** Execute the String command  */
    
    public boolean executeCommand(String command)
    {
        if (command!=null)
        {
            //getEditor().insertStringAt(command+"\n",getEditor().getDocument().length()-1,false);
            getEditor().insertStringAt(command,getEditor().getDocument().length()-1,false);
            moveToEnd();
            addCommandToHistory(command);
            interaction.sendCommand(command);
        };
        return true;
    }
    
    /** Insert the command String */
    public void insertCommand(String command)
    {
        getEditor().insertStringAt(command,getEditor().getDocument().length()-1,false);
        moveCursorToEnd();
    }
}
