/*
 * LispClispInteraction.java
 *
 * Created on March 23, 2002, 10:22 PM
 */

package IDE.Interaction;

import IDE.*;
import IDE.Messaging.*; 
import IDE.LispConnection.*;
import IDE.Configuration.*;
import Install.*;
import javax.swing.*;

import java.io.*;
import java.net.*;


/**
 * Represents a interaction with a ACL lisp process
 * @author  marc
 */
public class LispAclWindowsInteraction extends LispInteraction
{
    /** Number of ACL instances */
    private static int count;
    /** Name of the interaction window */
    private String name;
    /** Creates a new instance of LispClispInteraction */
    public LispAclWindowsInteraction()
    {
        super(InteractionCommands.getInteractionCommand(LispAclWindowsInteraction.class.getName()),"(load \"acl"+"/"+"debugger.fasl\")\n");
        count++; 
        name="ACL "+count; 
    }
    /** Override the getName() emthod */
    public String getName()
    {
        return name;
    }
    /** Get the extension of compiled code */
    public String getCompiledExtension()
    {
        return "fasl";
    }
    
	/** 
	 * Retrieve output from the lisp process, overwrite the method in LispInteraction
	 * because we have no stderr
	 */
	public String receiveErrorAndOutput()
	{
    	
		StringBuffer output=new StringBuffer();
		int n;
		try
		{
			while (true)
			{
				if (stoplisten)
					return "";
				if (lispoutput==null)
				{
					// Listener is not yet active
					Thread.sleep(50);
				}
				else if (lispoutput.available()==0)
				{
					if (output.length()>0)
						return output.toString();
					// No output so we should block
					Thread.sleep(50);
				}
				else if (lispoutput.available()>0)
				{
					output.append((char) lispoutput.read());
				};
			}
		}
		catch (InterruptedException e)
		{
			return null;
		}
		catch (IOException e)
		{
			cleanupAftererror("IO Error during read from lispprocess ",e);
			return null;
		}
	}

    /**
     *  Send a command to ACL overwrite the method in LispInteraction because
     *  line feeds has to be processed in a different way
     */
    public void sendCommand(final String command)
    {
        Runnable doSend=new Runnable()
        {
            public void run()
            {
                try
                {
                    MessageHandler.say("Lisp command : "+command);
                    interactionpane.outputString("\n");
                    lispinput.write(command);
                    lispinput.flush();
                }
                catch (IOException e)
                {
                    cleanupAftererror("IOError during write to lispprocess error "+e.getMessage(),e);
                };
            }
        };
        Thread thread=new Thread(doSend);
        thread.start();
    }
    /** Generates a file representing the directory where the lisp files are installed in */
    public File lispSubDir(String installdir)
    {
        return new File(installdir,"acl");
    }
	/** Start the process to interact with  */
	public void startProcess()
	{
		final int portnr=getPortNr();
		if (portnr<0)
			return;
		final int listenportnr=getPortNr();
		if (listenportnr<0)
			return;
		final Runnable setWatchResult=new Runnable()
		{
			public void run()
			{
				// Set a watchpoint for the result of a call
				Main.getWatches().addWatchPoint("Result("+getName()+")","DEBUGGER::**values**",LispAclWindowsInteraction.this);
			}
		};
		Thread createlispconnection=new Thread(new Runnable()
		{
			public void run()
			{
				Object result;
				try
				{
					// Create the lisp connection
					lispconnection=new LispConnection(portnr,portnr+1);
					MessageHandler.say("Socket connection to lisp process established");
					// Pass the interaction object to lisp
					lispconnection.callLisp("DEBUGGER::set-interaction",new Object[]
					{LispAclWindowsInteraction.this},true);
					MessageHandler.say("Creating interaction object finished");
					// Set the current directory
					setCurrentDirectory(Interactions.currentDirectory);
					// Get the process id of the LISP process
					if (OsSystem.processManagmentIsPossible())
					{
						result=lispconnection.callLisp("(DEBUGGER::get-process-id)",true);
						if (result instanceof Long)
							processid=((Long) result).intValue();
					};
					// Set a watchpoint for the result of a call
					SwingUtilities.invokeLater(setWatchResult);
					// Successfull completed
				}
				catch (Exception e)
				{
					// MessageHandler.report("Failed to establish socket connection to lisp process",e);
					cleanupAftererror("Failed to establish socket connection to lisp process",e);
				};
			}
		});
		Thread startListener=new Thread(new Runnable()
		{
			public void run()
			{
				try
				{
					// Create a server socket
					ServerSocket serversocket=new ServerSocket(listenportnr);
					// Wait on Lisp Client connection
					Socket clientsocket=serversocket.accept();
					clientsocket.setTcpNoDelay(true);
					lispoutput=clientsocket.getInputStream();
					lispinput=new OutputStreamWriter(clientsocket.getOutputStream());
					// Start the startup command
					if (Configuration.lispStartupFile.length()>0 && new File(Configuration.lispStartupFile).exists())
						sendCommand("(load \""+Configuration.lispStartupFile+"\")");					
				}
				catch (IOException e)
				{
					MessageHandler.report("Failed to establish socket connection with ACL listener ",e);
				}
			}
		});
		try
		{
			// Create a socket connection to listen for the listener of the lisp subprocess
			startListener.start();
			// Create the connection with the lisp system
			createlispconnection.start();
			// Wait some seconds to allow the system to start listening
			Thread.sleep(100);
			// Start the Lisp command
			String[] tmpcmds=InteractionCommands.parseCommandString(startcommand);
			String[] cmds=new String[tmpcmds.length+6];
			System.arraycopy(tmpcmds,0,cmds,0,tmpcmds.length);
			cmds[tmpcmds.length]="-L";
			cmds[tmpcmds.length+1]=Configuration.installDirectory+File.separator+"acl"+File.separator+"aclstartup.fasl";
			cmds[tmpcmds.length+2]="--";
			cmds[tmpcmds.length+3]=String.valueOf(listenportnr);
			cmds[tmpcmds.length+4]=String.valueOf(portnr);
			cmds[tmpcmds.length+5]=String.valueOf(portnr+1);
			// Start the lisp process
			process=Runtime.getRuntime().exec(cmds,null,new File(Configuration.installDirectory));
		}
		catch (IOException e)
		{
			cleanupAftererror("Could not run the command <"+startcommand+">",e);
		}
		catch (Exception e)
		{
			cleanupAftererror("Lisp startup interrupted",e);
		}
	}
}
