/*
 * LispConnection.java
 *
 * Created on June 17, 2000, 5:36 PM
 */

package IDE.LispConnection;

/**
 *
 * @author  root
 * @version 0.1
 * Handles a connection between Lisp and Java
 */

import java.io.*;
import java.net.*;
import java.util.*;
import javax.swing.*;
import java.lang.reflect.*;
import IDE.Messaging.*;


public class LispConnection implements Runnable
{
    private BufferedOutputStream callOut; //Sends a call to lisp via this stream
    private BufferedInputStream callIn;   //Get result call via this stream
    private ServerSocket callServer;
    private Socket callClient;
    private BufferedOutputStream calledOut; //Receive LISP calls via this stream
    private BufferedInputStream calledIn;  //Send result call via this
    private ServerSocket calledServer;
    private Socket calledClient;
    private Thread thread;
    private Hashtable integerToObject; // java objects referenced to by LISP
    private Hashtable objectToInteger;
    private int objectcounter;
    
    /* Classes to represent base types of  JAVA *, numbers convert from Long or Double */
    
    private class BaseType
    {
        protected Object getValue()
        {
            return new Integer(0);
        }
        
        protected Class getType()
        {
            return Integer.TYPE;
        }
    }
    
    private class ByteType extends BaseType
    {
        Byte value;
        
        ByteType(Object val)
        {
            value=new Byte((byte) ((Long) val).longValue());
        }
        
        protected Object getValue()
        {
            return value;
        }
        protected Class getType()
        {
            return Byte.TYPE;
        }
    }
    
    private class ShortType extends BaseType
    {
        Short value;
        
        ShortType(Object val)
        {
            value=new Short((short) ((Long) val).longValue());
        }
        
        protected Object getValue()
        {
            return value;
        }
        
        protected Class getType()
        {
            return Short.TYPE;
        }
    }
    
    private class IntType extends BaseType
    {
        Integer value;
        
        IntType(Object val)
        {
            value=new Integer((int) ((Long) val).longValue());
        }
        
        protected Object getValue()
        {
            return value;
        }
        
        protected Class getType()
        {
            return Integer.TYPE;
        }
    }
    
    private class LongType extends BaseType
    {
        Long value;
        
        LongType(Object val)
        {
            value=(Long) val;
        }
        
        protected Object getValue()
        {
            return value;
        }
        
        protected Class getType()
        {
            return Long.TYPE;
        }
    }
    
    private class FloatType extends BaseType
    {
        Float value;
        
        FloatType(Object val)
        {
            value=new Float((float) ((Double) val).doubleValue());
        }
        
        protected Object getValue()
        {
            return value;
        }
        
        protected Class getType()
        {
            return Double.TYPE;
        }
    }
    
    private class DoubleType extends BaseType
    {
        Double value;
        
        DoubleType(Object val)
        {
            value=(Double) val;
        }
        
        protected Object getValue()
        {
            return value;
        }
        
        protected Class getType()
        {
            return Double.TYPE;
        }
    }
    
    private class CharType extends BaseType
    {
        Character value;
        
        CharType(Object val)
        {
            value=(Character) val;
        }
        
        protected Object getValue()
        {
            return value;
        }
        
        protected Class getType()
        {
            return Character.TYPE;
        }
    }
    
    private class BooleanType extends BaseType
    {
        Boolean value;
        
        BooleanType(Object val)
        {
            value=(Boolean) val;
        }
        
        protected Object getValue()
        {
            return value;
        }
        
        protected Class getType()
        {
            return Boolean.TYPE;
        }
    }
    
    
    
    
    
    /** Creates new LispConnection */
 
    public LispConnection(int socket1,int socket2) throws LispConnectionException
    {
        // Initiate the hashtable containing objects in use by the LISP system
        objectToInteger=new Hashtable(100);
        integerToObject=new Hashtable(100);
        objectcounter=0;
        // Create a socket to get calls from lisp and to send the result of these call's back
        try
        {
            calledServer = new ServerSocket(socket1);
        }
        catch (IOException e)
        {
            throw new LispConnectionException("Could not listen on port",socket1,socket2);
        }
        // Wait on LISP client connection
        try
        {
            calledClient = calledServer.accept();
            calledClient.setTcpNoDelay(true);
            // Creates the input and output channel
            calledOut = new BufferedOutputStream(calledClient.getOutputStream());
            calledIn = new BufferedInputStream(calledClient.getInputStream());
        }
        catch (IOException e)
        {
            throw new LispConnectionException("Accept on port " + socket1 + " failed",socket1,socket2);
        }
        // Create a socket to send calls to lisp and to get the results back
        try
        {
            callServer = new ServerSocket(socket2);
        }
        catch (IOException e)
        {
            throw new LispConnectionException("Could not listen on port" + socket2,socket1,socket2);
        }
        // Wait on LISP client connection
        try
        {
            callClient = callServer.accept();
            callClient.setTcpNoDelay(true);
            // Creates the input and output channel
            callOut = new BufferedOutputStream(callClient.getOutputStream());
            callIn = new BufferedInputStream(callClient.getInputStream());
        }
        catch (IOException e)
        {
            throw new LispConnectionException("Accept on port " + socket1 + " failed",socket1,socket2);
        }
        // Register this object as object with id 0 , so that lisp can refer to this object
        integerToObject.put(new Integer(0),this);
        objectToInteger.put(this,new Integer(0));
        objectcounter=1;
        // Starts Listening to LISP
        thread=new Thread(this,"processLISP");
        thread.setPriority(Thread.MIN_PRIORITY);
        thread.start(); // Starts the thread
    }

    /** Get a free server socket nr in the range [begin,end] , returns -1 if this fails */
    public static int getFreeSocketPort(int begin,int end)
    {
        if (begin>end)
            return -1;
        ServerSocket socket=null;
        int socketnr=begin;
        while (socketnr<=end)
        {
            try
            {
                socket=new ServerSocket(socketnr);
                // If this succeeds we have a free socket nr, so we close it now
                socket.close();
                // If this is still ok we return the found socket nr
                return socketnr;
            }
            catch (Exception e1)
            {
                // Failed to get the socket, try to close the socket just to be sure
                try
                {
                    socket.close();
                }
                catch (Exception e2)
                {
                };
            };
            socketnr++;
        };
        return -1;
    }
    /** Get the thread where the process incoming thread is running */
    public Thread getThread()
    {
        return thread;
    }
    // Cleans up everything created by server
    
    public void cleanup()
    {
        try
        {
            // Tell LISP to stop processing input
            try
            {
                callOut.write('A');
                callOut.flush();
                
            }
            catch (IOException e)
            {
            };
            
            // Cleanup Java
            thread.interrupt(); // Stop thread
            callOut.close();
            callIn.close();
            calledOut.close();
            calledIn.close();
            callClient.close();
            callServer.close();
            calledClient.close();
            calledServer.close();
        }
        catch (IOException e)
        {
            System.exit(-1);
        };
    }
    
    /* Method needed by the runnable interface */
    
    public void run()
    {
        while (!Thread.interrupted())
        {
            try
            {
                this.processLisp();
            }
            catch (IOException e)
            {
                MessageHandler.report("IO errors during processing LISP input",e);
                return;
            }
        };
    }
    /** Call a method via SwingUtilities.invokeLater to ensure that it is invoked in the Swing GUI thread */
    private void callInSwingThread(final Method method,final Object object,final Object[] arglist)
    {
        Runnable caller=new Runnable()
        {
            public void run()
            {
                try
                {
                    returnToLisp( method.invoke(object,arglist),calledOut);
                }
                catch (Exception e)
                {
                    MessageHandler.report("Error processing lisp function",e);
                }
            }
        };
        try
        {
            SwingUtilities.invokeAndWait(caller);
        }
        catch (Exception e)
        {
        };
    }
    /** Process the commands send by the Lisp System */
    protected void processLisp() throws IOException
    {
        String classname="",methodname="",membername="";
        Object[] arglist;
        Class[] arglistclass=null;
        Method method;
        Constructor constructor;
        Object object;
        Object data;
        char c;
        
        long nrargs;
        Class cls;
        try
        {
            switch ((c=(char) (calledIn.read())))
            {
                case 'S':
                case 'C':
                {
                    // Call of a static class method
                    // Get the class name
                    classname=(String) readLispData(calledIn);
                    // Get the method name
                    methodname=(String) readLispData(calledIn);
                    // Get the number of args
                    nrargs=((Long) readLispData(calledIn)).longValue();
                    // Read all the args
                    arglist=new Object[(int) nrargs];
                    arglistclass=new Class[(int) nrargs];
                    for (int i=1;i<=nrargs;i++)
                    {
                        data=readLispData(calledIn);
                        if (data instanceof BaseType)
                        {
                            arglist[i-1]=((BaseType) data).getValue();
                            arglistclass[i-1]=((BaseType) data).getType();
                        }
                        else
                        {
                            arglist[i-1]=data;
                            arglistclass[i-1]=data.getClass();
                        };
                    };
                    // Get the class and method and call the method
                    cls=Class.forName(classname); // get the class
                    method=cls.getMethod(methodname,arglistclass); // get the method
                    if (c=='S')
                        callInSwingThread(method,null,arglist);
                    else
                        returnToLisp(method.invoke(null,arglist),calledOut);
                    break;
                }
                case 'c':
                {
                    // Get a static member of a class
                    classname=(String) readLispData(calledIn);
                    // Get the member name
                    membername=(String)readLispData(calledIn);
                    // Get the class
                    cls=Class.forName(classname);
                    // Get the member
                    returnToLisp(cls.getField(membername).get(null),calledOut);
                    break;
                }
                case 's':
                case 'O':
                {
                    // Call of a method of a object
                    // Get the object
                    object=readLispData(calledIn);
                    methodname=(String) readLispData(calledIn);
                    // Get the number of args
                    nrargs=((Long) readLispData(calledIn)).longValue();
                    // Read all the args
                    arglist=new Object[(int) nrargs];
                    arglistclass=new Class[(int) nrargs];
                    for (int i=1;i<=nrargs;i++)
                    {
                        data=readLispData(calledIn);
                        if (data instanceof BaseType)
                        {
                            arglist[i-1]=((BaseType) data).getValue();
                            arglistclass[i-1]=((BaseType) data).getType();
                        }
                        else
                        {
                            arglist[i-1]=data;
                            arglistclass[i-1]=data.getClass();
                        };
                    };
                    // Get the method of the object
                    cls=object.getClass();
                    classname=cls.toString();
                    method=cls.getMethod(methodname,arglistclass); // Get the method
                    if (c=='s')
                        callInSwingThread(method,object,arglist);
                    else
                        returnToLisp(method.invoke(object,arglist),calledOut);
                    break;
                }
                case 'o':
                {
                    // Get the member of a object
                    // Get the object
                    object=readLispData(calledIn);
                    membername=(String) readLispData(calledIn);
                    // Get the class of the object
                    cls=object.getClass();
                    // Return the member value
                    returnToLisp(cls.getField(membername).get(object),calledOut);
                    break;
                }
                case 'N':
                {
                    // Call of a new method to create a object
                    // Get the class name
                    classname=(String) readLispData(calledIn);
                    // Get the number of args
                    nrargs=((Long) readLispData(calledIn)).longValue();
                    // Read all the args
                    arglist=new Object[(int) nrargs];
                    arglistclass=new Class[(int) nrargs];
                    for (int i=1;i<=nrargs;i++)
                    {
                        data=readLispData(calledIn);
                        if (data instanceof BaseType)
                        {
                            arglist[i-1]=((BaseType) data).getValue();
                            arglistclass[i-1]=((BaseType) data).getType();
                        }
                        else
                        {
                            arglist[i-1]=data;
                            arglistclass[i-1]=data.getClass();
                        };
                    };
                    // Create the object
                    cls=Class.forName(classname);
                    constructor=cls.getConstructor(arglistclass);
                    returnToLisp(constructor.newInstance(arglist),calledOut);
                    break;
                }
            };
        }
        catch (ClassNotFoundException e)
        {
            returnToLisp(null,calledOut);
            MessageHandler.report("Could not find class = "+classname+" to call a static method from",e);
        }
        catch (NoSuchMethodException e)
        {
            String args="(";
            if (arglistclass!=null)
            {
                for (int i=0;i<arglistclass.length;i++)
                {
                    args=args+arglistclass[i].getName();
                };
            };
            args=args+")";
            returnToLisp(null,calledOut);
            MessageHandler.report("Could not find method "+classname+"."+methodname+args,e);
        }
        catch (SecurityException e)
        {
            returnToLisp(null,calledOut);
            MessageHandler.report("You are not allowed to call method "+classname+"."+methodname,e);
        }
        catch (IllegalAccessException e)
        {
            returnToLisp(null,calledOut);
            MessageHandler.report("Illegal Access Exception when calling method "+classname+"."+methodname,e);
        }
        catch (InvocationTargetException e)
        {
            returnToLisp(null,calledOut);
            MessageHandler.report("Invocation Target Exception",e);
        }
        catch (InstantiationException e)
        {
            returnToLisp(null,calledOut);
            MessageHandler.report("Instantiation Exception",e);
        }
        catch (NoSuchFieldException e)
        {
            returnToLisp(null,calledOut);
            MessageHandler.report("Could not find member = "+membername,e);
        }
        catch (Exception e)
        {
        };
    }
    
    /* Return data back to the Lisp System */
    
    protected void returnToLisp(Object obj,BufferedOutputStream out) throws IOException
    {
        Class objclass;
        
        if (obj == null)
        {
            returnToLispBool(false,out);
        }
        else
        {
            
            objclass=obj.getClass();
            
            if (objclass.isArray())
            {
                // Return as array to lisp
                returnToLispArray((Object[]) obj,out);
            }
            else if (objclass==LinkedList.class)
            {
                // Return as a list to lisp
                returnToLispList((LinkedList) obj,out);
            }
            else if (objclass == Boolean.class)
            {
                // return as boolean value to lisp
                returnToLispBool(((Boolean) obj).booleanValue(),out);
            }
            else if (objclass == Short.class)
            {
                // return as number to lisp
                returnToLispNumber(((Short) obj).toString(),out);
            }
            else if (objclass == Integer.class)
            {
                // return as number to lisp
                returnToLispNumber(((Integer) obj).toString(),out);
            }
            else if (objclass == Long.class)
            {
                // return as number to lisp
                returnToLispNumber(((Long) obj).toString(),out);
            }
            else if (objclass == Float.class)
            {
                // return as number to lisp
                returnToLispNumber(((Float) obj).toString(),out);
            }
            else if (objclass == Double.class)
            {
                // return as number to lisp
                returnToLispNumber(((Double) obj).toString(),out);
            }
            else if (objclass == Byte.class)
            {
                // return as number to lisp
                returnToLispNumber(((Byte) obj).toString(),out);
            }
            else if (objclass == Character.class)
            {
                // return as character to lisp
                returnToLispChar(((Character) obj).charValue(),out);
            }
            else if (objclass == String.class)
            {
                // return as string to lisp
                returnToLispString((String) obj,out);
            }
            else if (objclass == LispObject.class)
            {
                // Return a lisp onject
                returnToLispLispObject((LispObject) obj,out);
            }
            else
            {
                // return as a object to lisp
                returnToLispObject(obj,out);
            };
        };
    }
    
    
    /* Send a array to LISP */
    
    protected void returnToLispArray(Object[] obj,BufferedOutputStream out) throws IOException
    {
        int len;
        
        len=obj.length;
        out.write('A');
        out.write((String.valueOf(len)).getBytes());
        out.write('\n');
        out.flush();
        
        for (int i=0;i<len;i++)
        {
            returnToLisp(obj[i],out);
        };
    }
    
    /* Return a List to LISP */
    
    protected void returnToLispList(LinkedList lst,BufferedOutputStream out) throws IOException
    {
        long len;
        
        len=lst.size();
        ListIterator lstiterator;
        out.write('L');
        out.write((String.valueOf(len)).getBytes());
        out.write('\n');
        out.flush();
        lstiterator=lst.listIterator();
        for (int i=1;i<=len;i++)
        {
            returnToLisp(lstiterator.next(),out);
        };
    }
    /* Send return data to the lisp system */
    
    protected void returnToLispBool(boolean bool,BufferedOutputStream out) throws IOException
    {
        out.write('B');
        if (bool)
        {
            out.write('t');
        }
        else
        {
            out.write("nil".getBytes());
        };
        out.write('\n');
        out.flush();
    }
    /* Send numeric data to the Lisp system */
    
    protected void returnToLispNumber(String number,BufferedOutputStream out) throws IOException
    {
        out.write('N');
        out.write(number.getBytes());
        out.write('\n');
        out.flush();
    }
    
    /* Returns a character to the lisp system */
    
    protected void returnToLispChar(char chr,BufferedOutputStream out) throws IOException
    {
        String str;
        out.write('C');
        str="#\\"+chr;
        out.write(str.getBytes());
        out.write('\n');
        out.flush();
    }
    
    /* Returns a string to the lisp system */
    
    protected void returnToLispString(String str,BufferedOutputStream out) throws IOException
    {
        out.write('S');
        // Write length string + string value
        out.write((String.valueOf(str.length())).getBytes());
        out.write('\n');
        out.write(str.getBytes());
        out.flush();
    }
    /* Returns a object to the lisp system */
    
    protected void returnToLispObject(Object obj,BufferedOutputStream out) throws IOException
    {
        int objectid;
        Integer integer=(Integer) objectToInteger.get(obj);
        if (integer==null)
        {
            // Remember the returned object so LISP can always refer to it
            objectcounter=objectcounter+1;
            objectid=objectcounter;
            integerToObject.put(new Integer(objectcounter),obj);
            objectToInteger.put(obj,new Integer(objectcounter));
        }
        else
        {
            objectid=integer.intValue();
        };
        // Return object identifier to LISP
        out.write('O');
        out.write((String.valueOf(objectid)).getBytes());
        out.write('\n');
        out.flush();
    }
    
    /* Return a LISP object to LISP */
    
    protected void returnToLispLispObject(LispObject obj,BufferedOutputStream out) throws IOException
    {
        out.write('l');
        out.write((String.valueOf(obj.getId())).getBytes());
        out.write('\n');
        out.flush();
    }
    
    /* Send comamands to the Lisp system */
    public synchronized Object callLisp(String lispcmd,boolean nodisplay) throws IOException
    {
        //Object res;
        if (nodisplay)
            callOut.write('S');
        else
            callOut.write('s');
        returnToLisp(lispcmd,callOut);
        callOut.flush();
        /* Get the result of the LISP call */
        return readLispData(callIn);
    }
    
    /* Accept a LIST and send  it  as LISP command */
    
    public synchronized Object callLisp(LinkedList lst,boolean nodisplay) throws IOException
    {
        // Object res;
        ListIterator lstiterator;
        Object function;
        
        lstiterator=lst.listIterator();
        
        // Indicate a lisp call to LISP
        if (nodisplay)
            callOut.write('C');
        else
            callOut.write('c');
        // Give the number of arguments
        callOut.write((String.valueOf(lst.size()-1)).getBytes());
        callOut.write('\n');
        // First element in lst is a function  ,
        function=lstiterator.next();
        returnToLisp(function,callOut);
        while (lstiterator.hasNext())
        {
            returnToLisp(lstiterator.next(),callOut);
        };
        /* Get the result of the LISP call */
        return readLispData(callIn);
    }
    
    /* Accept a Function and a arglist and send it as a LISP command */
    public synchronized Object callLisp(Object function,Object[] args,boolean nodisplay) throws IOException
    {
        // Indicate a lisp call to LISP
        if (nodisplay)
            callOut.write('C');
        else
            callOut.write('c');
        // Give the number of arguments
        callOut.write((String.valueOf(args.length)).getBytes());
        callOut.write('\n');
        //Send out the function
        returnToLisp(function,callOut);
        for (int i=0;i<args.length;i++)
        {
            returnToLisp(args[i],callOut);
        };
        /* Get the result of the LISP call */
        return readLispData(callIn);
    }
    /* Read the data send by LISP */
    protected Object readLispData(BufferedInputStream strm) throws IOException
    {
        char c;
        switch (c=(char) strm.read())
        {
            case 'N':
            {
                return readLispDataNumber(strm);
            }
            case 'B':
            {
                return readLispDataBool(strm);
            }
            case 'S':
            {
                return readLispDataString(strm);
            }
            case 'A':
            {
                return readLispDataArray(strm);
            }
            case 'L':
            {
                return readLispDataList(strm);
            }
            case 'l':
            {
                return readLispDataLispObject(strm);
            }
            case 'O':
            {
                return readLispDataObject(strm);
            }
            case 'C':
            {
                return readLispDataCast(strm);
            }
        }
        MessageHandler.warn("Expected a character N,B,S,A,L,I,O or C but received a <"+c+">");
        return new Integer(0); // Dummy return
    }
    
    /* Read a data from lisp to be casted to  a elementary java type */
    
    protected Object readLispDataCast(BufferedInputStream strm) throws IOException
    {
        Long value;
        Double dbl;
        switch((char) strm.read())
        {
            case 'S':
            {
                return new ShortType(readLispData(strm));
            }
            case 'I':
            {
                return new IntType(readLispData(strm));
            }
            case 'B':
            {
                return new ByteType(readLispData(strm));
            }
            case 'L':
            {
                return new LongType(readLispData(strm));
            }
            case 'F':
            {
                return new FloatType(readLispData(strm));
            }
            case 'D':
            {
                return new DoubleType(readLispData(strm));
            }
            case 'C':
            {
                return new CharType(readLispData(strm));
            }
            case 'b':
            {
                return new BooleanType(readLispData(strm));
            }
        };
        return new Integer(0);
    }
    
    
    /* Read a array of data from lisp */
    
    protected Object[] readLispDataArray(BufferedInputStream strm) throws IOException
    {
        long len;
        Object[] res;
        // Read the length of the array
        len=((Long) readLispDataNumber(strm)).longValue();
        // Read now all the items in the array
        res=new Object[(int) len];
        for (int i=0;i<len;i++)
        {
            res[i]=readLispData(strm);
        }
        return res;
    }
    
    protected LinkedList readLispDataList(BufferedInputStream strm) throws IOException
    {
        long len;
        LinkedList lst;
        // read the length of the list
        len=((Long) readLispDataNumber(strm)).longValue();
        // read now all the items in the list
        lst=new LinkedList();
        for (int i=0;i<len;i++)
        {
            lst.add(readLispData(strm));
        }
        return lst;
    }
    
    /* Read Java Object from LISP */
    protected Object readLispDataObject(BufferedInputStream strm)
    {
        long id;
        // Read the object identifier
        id=((Long) readLispDataNumber(strm)).longValue();
        // return the object
        return integerToObject.get(new Integer((int) id));
    }
    
    /* Read Lisp Object from LISP */
    
    protected Object readLispDataLispObject(BufferedInputStream strm) throws IOException
    {
        return new LispObject(((Long) readLispData(strm)).longValue(),this);
    }
    
    /* Read Number from Lisp */
    
    protected Object readLispDataNumber(BufferedInputStream strm)
    {
        StringBuffer buffer=new StringBuffer(10);
        char c;
        Long longResult;
        Double doubleResult;
        String str;
        
        
        try
        {
            // Read until we reach the null character
            while ((c=(char) strm.read())!='\0')
            {
                buffer.append(c);
            };
            // Try to interpret the string as a number
            str=buffer.toString();
            try
            {
                longResult=Long.valueOf(str);
                return longResult;
            }
            catch (NumberFormatException e3)
            {
                try
                {
                    doubleResult=Double.valueOf(str);
                    return doubleResult;
                }
                catch (NumberFormatException e5)
                {
                    MessageHandler.report("Data from lisp is not a number when it should be, data="+str,e5);
                };
            };
        }
        catch (IOException e)
        {
        };
        return new Long(0);
    }
    
    /* Read boolean data from LISP */
    
    protected Boolean readLispDataBool(BufferedInputStream strm) throws IOException
    {
        if ((Character.toLowerCase((char) strm.read()))=='t')
        {
            return new Boolean(true);
        }
        else
        {
            return new Boolean(false);
        }
    }
    
    /* Read String data from lisp */
    
    protected String readLispDataString(BufferedInputStream strm) throws IOException
    {
        StringBuffer buffer=new StringBuffer(10);
        int length=0;
        char c;
        
        // First read the the length of the string
        while ((c=(char) strm.read())!='\0')
        {
            buffer.append(c);
        };
        length=Integer.parseInt(buffer.toString());
        // Now we read the string
        buffer=new StringBuffer(100);
        for (int i=1;i<=length;i++)
            buffer.append((char) strm.read());
        return buffer.toString();
    }
    
    /** Test code
     * public static int testInt(Long n)
     * {
     * return (int) n.longValue();
     * }
     *
     * public static short testShort(Long n)
     * {
     * return (short) n.shortValue();
     * }
     *
     * public static long testLong(Long n)
     * {
     * return n.longValue();
     * }
     *
     * public static float testFloat(Long n)
     * {
     * return (float) n.longValue();
     * }
     *
     * public static double testDouble(Long n)
     * {
     * return (double) n.longValue();
     * }
     *
     * public static byte testByte(Long n)
     * {
     * return (byte) 10;
     * }
     *
     * public static char testChar(Long n)
     * {
     * return 'A';
     * }
     *
     * public static Long[] testArray(Long n1,Long n2)
     * {
     * Long[] ret=new Long[2];
     * System.out.println("testArray");
     * ret[0]=n1;
     * ret[1]=n2;
     * return ret;
     * }
     *
     * public static LinkedList testList(Long n1,Long n2)
     * {
     * LinkedList lst=new LinkedList();
     *
     * lst.add(n1);
     * lst.add(n2);
     *
     * return lst;
     * }
     *
     * public static Object ObjectTest(Object obj)
     * {
     * System.out.println("Object :" + obj);
     * return obj;
     * }
     *
     * public static void main(String args[])
     * {
     * LispConnection connection;
     * Object obj;
     * LinkedList lst;
     * try
     * {
     * connection=new LispConnection(4444,4445);
     * System.out.println("Call to LISP");
     * connection.callLisp("(princ 'OK)");
     * obj=connection.callLisp("(test)");
     * System.out.println("Result of (test) call = "+obj);
     * lst=new LinkedList();
     * lst.add(obj);
     * lst.add(new Integer(10));
     * obj=connection.callLisp(lst);
     * System.out.println("Result of (cos 10) = "+obj);
     * // connection.cleanup();
     * }
     * catch (LispConnectionException e)
     * {
     * }
     * catch (IOException e)
     * {
     * }
     * }
     */
}