/*
 * LispAnalyser.java
 *
 * Created on March 18, 2001, 1:25 PM
 */

package IDE.LispSyntax;

import java.util.*;

import IDE.Main;
import IDE.Project.LispRepository;
import IDE.CodeEditor.*;
import IDE.CodeEditor.Parser.*;

/**
 * Class to represent the SyntaxElements of lisp code
 * More specifically, tries to figure out the following:
 * - lexical variables defined in the code (for autocompletion)
 * - function and macrocall's (for syntax highlighting and and autocompletion)
 * - special function and macrocall's (for syntax highlighting)
 * - defined functions,macros and methods (for autocompletion)
 *
 * @author  marc
 * @version
 */
public abstract class LispAnalyzer
{
	/**
	 * Static strings containing the keywords starting a lisp definition
	 */
	/** Interned DEFUN symbol */
	private static String DEFUN="DEFUN".intern();
	/** Interned DEFMETHOD symbol */
	private static String DEFMETHOD="DEFMETHOD".intern();
	/** Interned DEFGENERIC symbol */
	private static String DEFGENERIC="DEFGENERIC".intern();
	/** Interned DEFMACRO symbol */
	private static String DEFMACRO="DEFMACRO".intern();
	/** Interned DEFCLASS symbol */
	private static String DEFCLASS="DEFCLASS".intern();
	/** Interned SUBTYPEP symbol */
	private static String SUBTYPEP="SUBTYPEP".intern();
	/** Interned COERCE symbol */
	private static String COERCE="COERCE".intern();
	/** Interned TYPEP symbol */
	private static String TYPEP="TYPEP".intern();
	/** Interned HANDLER-BIND symbol */
	private static String HANDLERBIND="HANDLER-BIND".intern();
	/** Interned HANDLER-CASE symbol */
	private static String HANDLERCASE="HANDLER-CASE".intern();
	/** Interned CHECK-TYPE symbol */
	private static String CHECKTYPE="CHECK-TYPE".intern();
	/** Interned CTYPECASE symbol */
	private static String CTYPECASE="CTYPECASE".intern();
	/** Interned ETYPECASE symbol */
	private static String ETYPECASE="ETYPECASE".intern();
	/** Interned TYPECASE symbol */
	private static String TYPECASE="TYPECASE".intern();
	/** Interned symbol */
	private static String MAKECONDITION="MAKE-CONDITION".intern();
	/** Predefined types in LISP */
	private static String[] predefinedTypes=new String[] {"arithmetic-error",
		"function","simple-condition","array",                         
	    "generic-function","simple-error","atom",
	    "hash-table","simple-string",            
		"base-char","integer","simple-type-error",    
		"base-string","keyword","simple-vector",              
		"bignum","list","simple-warning",     
		"bit","logical-pathname","single-float",              
		"bit-vector","long-float","standard-char",              
		"broadcast-stream","method","standard-class",             
		"built-in-class","method-combination","standard-generic-function",  
		"cell-error","nil","standard-method",
		"character","null","standard-object",         
		"class","number","storage-condition",          
		"compiled-function","package","stream",          
		"complex","package-error","stream-error",               
		"concatenated-stream","parse-error","string",              
		"condition","pathname","string-stream",              
		"cons", "print-not-readable","structure-class",            
		"control-error","program-error","structure-object",           
		"division-by-zero","random-state","style-warning",            
		"double-float","ratio","symbol",
		"echo-stream", "rational","synonym-stream",           
		"end-of-file","reader-error",
		"error", "readtable","two-way-stream",
		"extended-char", "real","type-error",         
		"file-error","restart","unbound-slot",            
		"file-stream", "sequence","unbound-variable",          
		"fixnum","serious-condition","undefined-function",         
		"float","short-float","unsigned-byte",
		"floating-point-inexact","signed-byte","vector",                  
		"floating-point-invalid-operation", "simple-array","warning",                    
		"floating-point-overflow","simple-base-string",            
		"floating-point-underflow","simple-bit-vector"};                              

 
    /** Create a LispAnalyser */
    public LispAnalyzer()
    {
    }
    /** Do the analyse */
    public abstract boolean analyze(LispListElement list,CodeElementIterator listelements, AnalyzerList todo);
    /** Find the possible hints for a e given element and add them to the hints vector (vector of strings) */
    public abstract boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo);
    /** Handle the todo vector (remaining analyze to do)  */
    protected boolean analyzeNext(AnalyzerList todo)
    {
        if (todo.nullp())
        {
            return true;
        }
        else
        {
            return todo.firstAnalyzer().analyze(todo.firstList(),todo.firstListElements(),todo.rest());
        }
    }
    /** Handle the todo vector (remaining hints to be discoverd) */
    protected boolean findHintsNext(AnalyzerList todo,CodeElement element,Vector hints)
    {
    	if (todo.nullp())
    		return true;
    	else
    		return todo.firstAnalyzer().findHints(element,todo.firstList(),todo.firstListElements(),hints,todo.rest());
    }
	/** Get a vector of all the known function and macros */
	public static Vector getFunctionsAndMacros()
	{
		Vector result=new Vector();
		String[] buildins=LispAnalyzerGen.getLispKeywords();
		String[] definitions=LispRepository.lisprepository.getDefinitionsOfType(new String[] {"defgeneric","defmethod","defun","defmacro"});
		for (int i=0;i<buildins.length;i++)
			result.add(buildins[i].toLowerCase());
		for (int i=0;i<definitions.length;i++)
			result.add(definitions[i]);
		return result;
	}
	/**
	 *  Get a vector of all the possible lisp forms that can be evaluated 
	 * This includes the variables valid at the context of the element
	 */
	public static Vector getValidForms(CodeElement element)
	{
		Vector result=new Vector();
		// If the element is part of a lexical context defining variables then these variables are legal lisp forms
		result.addAll(LispListElement.getAllVariablesFor(element));
		// Find the possible variables, constants and parameters defined in the current project
		String[] defs=LispRepository.lisprepository.getDefinitionsOfType(new String[]{"defparameter","defconstant","defvar"});
		for (int i=0;i<defs.length;i++)
			result.add(defs[i]);
		// Add the possibility to do a call of a new form
		result.add("()");
		return result;
	}
	/** 
	 * Check if the element is positioned after a DEFUN,DEFPARAMETER,DEFGENERIC or DEFMETHOD 
	 * and giving back then a list of undefined functions gives a empty vector back otherwise
	 * @author marc
	 *
	 */ 
	public static Vector getUndefinedFunctions(CodeElement element)
	{
		if (element.getParent() instanceof LispListElement)
		{
			CodeElementVector elements=((LispListElement) element.getParent()).getElements();
			if (elements.size()>1)
				if (elements.codeElementAt(0) instanceof LispAtomElement)
				{
					String str=((LispAtomElement) elements.codeElementAt(0)).getSymbol();
					if ((str==DEFUN || str==DEFGENERIC || str==DEFMACRO || str==DEFMETHOD) && elements.codeElementAt(1)==element)
					{
						// Get the current document in focus
						CodeEditor current=Main.currentEditor();
						if (current!=null)
						{
							SortedSet undefines=LispRepository.lisprepository.getUndefinedCallers();
							String[] keywords=LispAnalyzerGen.getLispKeywords();
							for (int i=0;i<keywords.length;i++)
								undefines.remove(keywords[i]);
							return new Vector(undefines);
						}
					}
			}
		}
		return new Vector();
	}
	/**
	 * Check if the element is positioned on position that stands for a type
	 * and give back a list of types (with a ' added before) and lexical variables, parameters ....
	 * @param element Marker element to indicate a position
	 * @return
	 */
	public static Vector getTypes(CodeElement element)
	{
		CodeElementVector elements,superelements,supersuperelements;
		LispListElement parent,grantparent;
		String first;
		if (element.getParent() instanceof LispListElement)
		{
			parent=(LispListElement) element.getParent();
			elements=parent.getElements();
			if (elements.size()>1 && elements.codeElementAt(0) instanceof LispAtomElement)
			{
				first=((LispAtomElement) elements.codeElementAt(0)).getSymbol();
				if ((first==SUBTYPEP && (elements.elementAtIs(1,element) || elements.elementAtIs(2,element))) ||
				(first==COERCE && elements.elementAtIs(2,element)) ||
				(first==TYPEP && elements.elementAtIs(2,element)) ||
				(first==MAKECONDITION && elements.elementAtIs(1,element)))
				{
					return getQuotedTypes();
				}
				else if (first==CHECKTYPE && elements.elementAtIs(2,element))
				{
					return getUnquotedTypes();					
				}
			}
			if (parent.getParent() instanceof LispListElement)
			{
				grantparent=(LispListElement) parent.getParent();
				superelements=grantparent.getElements();
				if (superelements.size()>1 && superelements.codeElementAt(0) instanceof LispAtomElement)
				{
					first=((LispAtomElement) superelements.codeElementAt(0)).getSymbol();
					if (((first==CTYPECASE || first==ETYPECASE || first==TYPECASE || first==HANDLERCASE) && superelements.indexOf(parent,0)>1))
					{
						return getUnquotedTypes();
					}
					else if (first==DEFCLASS && superelements.elementAtIs(2,parent))
					{
						return getClassTypes();
					}
				};
				if (grantparent.getParent() instanceof LispListElement)
				{
					supersuperelements=((LispListElement) grantparent.getParent()).getElements();
					if (supersuperelements.size()>1 && supersuperelements.codeElementAt(0) instanceof LispAtomElement)
					{
						first=((LispAtomElement) supersuperelements.codeElementAt(0)).getSymbol();
						if ((first==HANDLERBIND && supersuperelements.elementAtIs(1,grantparent) && elements.elementAtIs(0,element)))
							return getUnquotedTypes();
					}
				};
			};
		};
		return new Vector();
	}
	public static Vector getQuotedTypes()
	{
		Vector result=new Vector();
		String[] defs=LispRepository.lisprepository.getDefinitionsOfType(new String[]{"defclass","deftype","defstruct"});
		for (int i=0;i<defs.length;i++)
			result.add("'"+defs[i]);
		for (int i=0;i<predefinedTypes.length;i++)
			result.add("'"+predefinedTypes[i]);
		result.add("t");
		return result;
	}
	public static Vector getUnquotedTypes()
	{
		Vector result=new Vector();
		String[] defs=LispRepository.lisprepository.getDefinitionsOfType(new String[]{"defclass","deftype","defstruct"});
		for (int i=0;i<defs.length;i++)
			result.add(defs[i]);
		for (int i=0;i<predefinedTypes.length;i++)
			result.add(predefinedTypes[i]);
		result.add("t");
		return result;		
	}
	public static Vector getClassTypes()
	{
		Vector result=new Vector();
		String[] defs=LispRepository.lisprepository.getDefinitionsOfType(new String[]{"defclass"});
		for (int i=0;i<defs.length;i++)
			result.add(defs[i]);
		return result;
	}
}

/**
 * Analyzer who checks if the next element is a token with the token text equal to a given string
 */

class TokenLispAnalyzer extends LispAnalyzer
{
    /** String to test for  equality */
    private String string;
    
    public TokenLispAnalyzer(String string)
    {
        super();
        this.string=string.intern();
    }
    /*
     *  (non-Javadoc)
     * @see IDE.LispSyntax.LispAnalyzer#analyze(IDE.CodeEditor.Parser.LispListElement, IDE.CodeEditor.CodeElementIterator, IDE.LispSyntax.AnalyzerList)
     */
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        CodeElement token;
        String str;
        
        token=listelements.next();
        if (token instanceof LispAtomElement)
        {
            if (((LispAtomElement) token).isEqualTo(string) && analyzeNext(todo))
            {
                list.addKeyWord(token);
                return true;
            }
            else
            {
                return false;
            }
        }
        else
        {
            return false;
        }
    }
    /*
     *  (non-Javadoc)
     * @see IDE.LispSyntax.LispAnalyzer#findHints(IDE.CodeEditor.CodeElement, IDE.CodeEditor.Parser.LispListElement, IDE.CodeEditor.CodeElementIterator, java.util.Vector, IDE.LispSyntax.AnalyzerList)
     */
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		CodeElement token;
		String str;
		token=listelements.next();
		if (token==element)
		{
			hints.add(string.toLowerCase());
			return true;
		}
		else
		{
			if (token instanceof LispAtomElement)
			{
				if (((LispAtomElement) token).isEqualTo(string) && findHintsNext(todo,element,hints))
					return true;
				else
					return false;	
			}
			else
			{
				return false;
			}
		}
	}
}

/**
 * Analyzer to check that there is a next element , no mather what it is
 */

class VarLispAnalyzer extends LispAnalyzer
{
    public VarLispAnalyzer()
    {
        super();
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
		return listelements.next()!=null && analyzeNext(todo);
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		CodeElement next=listelements.next();
		if (next==element)
		{
			hints.addAll(getUndefinedFunctions(element));
			hints.addAll(getTypes(element));
			return true;
		}
		else
		{
			return next!=null && findHintsNext(todo,element,hints);
		}
	}
}

/*
 * Analyzer used to process the last element in a dotted list
 */

class AfterDotAnalyzer extends LispAnalyzer
{
    private LispAnalyzer analyzer;
    public AfterDotAnalyzer(LispAnalyzer analyzer)
    {
        super();
        this.analyzer=analyzer;
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
    	return analyzer.analyze(list,ListOfLispListElement.getIteratorOnListOfIterator(listelements),todo);
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		return analyzer.findHints(element,list,ListOfLispListElement.getIteratorOnListOfIterator(listelements),hints,todo);
	}
}

/**
 * Analyzer to check that the next element is a string
 */

class StringLispAnalyzer extends LispAnalyzer
{
    public StringLispAnalyzer()
    {
        super();
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        CodeElement string;
        
        string=listelements.next();
        if (string instanceof LispStringElement && analyzeNext(todo))
            return true;
        else
            return false;
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
			CodeElement string=listelements.next();
			if (string==element)
			{
				hints.add("\"\"");
				return true;		
			}
			else if (string instanceof LispStringElement && findHintsNext(todo,element,hints))
			{
				return true;
			}
			else
			{
				return false;
			}
	}
}

/** Analyzer to check that we have the caller of a function or macro */

class CallLispAnalyzer extends LispAnalyzer
{
    public CallLispAnalyzer()
    {
        super();
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        CodeElement token;
        CodeElement first;
        token=listelements.next();
        if (token instanceof LispListElement)
        {
            if (analyzeNext(todo))
            {
                first=((LispListElement) token).getFirst();
                if (first instanceof LispAtomElement)
                {
                    list.addCaller(token);
                };
                return true;
            }
            else
            {
                return false;
            }
        }
        else if (token instanceof LispFunctionElement)
        {
            if (analyzeNext(todo))
            {
                return true;
            }
            else
            {
                return false;
            }
        }
        else
        {
            return token!=null && analyzeNext(todo);
        }
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		CodeElement token;
		CodeElement first;
        
		 token=listelements.next();
		 if (token==element)
		 {
		 	hints.addAll(getValidForms(element));
		 	return true;
		 }
		 else if (token instanceof LispListElement)
		 {
			first=((LispListElement) token).getFirst();
		 	if (first instanceof LispAtomElement || first instanceof HintMarkerElement)
			{
			 	if (token.findHints(element,hints))
			 	{
			 		list.addCaller(token);
			 		return findHintsNext(todo,element,hints);
			 	}
			 	else
			 	{
			 		return false;
			 	}
			 }
			 else
			 {
			 	return false;
			 }
		 }
		 else if (token instanceof LispFunctionElement)
		 {
		 	if (token.findHints(element,hints))
		 	{
		 		if (findHintsNext(todo,element,hints))
		 			return true;
		 		else
		 			return false;
		 	}
		 	else
		 	{
		 		return false;
		 	}
		 }
		 else
		 {
			 return token!=null && findHintsNext(todo,element,hints);
		 }		
	}
}

/** Analyzer to check that we have a variable defined  (remark . is not a variable) */

class VariableLispAnalyzer extends LispAnalyzer
{
    public VariableLispAnalyzer()
    {
        super();
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        CodeElement token;
        token=listelements.next();
        if (token instanceof LispAtomElement && analyzeNext(todo))
        {
            list.addVariable(token);
            return true;
        }
        else
        {
            return false;
        }
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		CodeElement token=listelements.next();
		if (token==element)
		{
			hints.add("<variable>");
			return true;
		}
		else if (token instanceof LispAtomElement && findHintsNext(todo,element,hints))
		{
			list.addVariable(token);
			return true;
		}
		else
		{
			return false;
		}
	}	    
}

/**
 * Analyzer which notes that it is called, used in tryLispAnalyzer to figure out when
 * we have to stop adding tries to the todo list
 * This analyzer does not actually analyze something , it just notes it is called and
 * then passes on the rest of the analyzing to todo
 */

class TestCallLispAnalyzer extends LispAnalyzer
{
    private boolean iscalled;
    
    public TestCallLispAnalyzer()
    {
        // We don't have to do any analyzing, only indicate that we are called so
        // we don't need super(_list,_listelements)
        iscalled=false;
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        // Indicate we are called
        iscalled=true;
        // Let todo analyze the rest and returns it
        return analyzeNext(todo);
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		// Indicate we are called
		iscalled=true;
		return findHintsNext(todo,element,hints);
	}
	 // Reset the iscalled flag to false, so that we can use this object again to test
    // if it does its (dummy) analyze step
    public void setNotCalled()
    {
        iscalled=false;
    }
    // Check if this function is called
    public boolean isCalled()
    {
        return iscalled;
    }
}

/** Analyzer to try zero or more applies of the same analyzer */
class TryLispAnalyzer extends LispAnalyzer
{
    /** Analyzer which must be tried zero or more times before the rest */
    LispAnalyzer[] trier;
    
    public TryLispAnalyzer(LispAnalyzer[] _try)
    {
        super();
        trier=_try;
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        SavePoint savepoint; 
        // Analyzer to check if it is called
       TestCallLispAnalyzer testCall=new TestCallLispAnalyzer();
        AnalyzerList trytodo;
        // Make a savepoint to rollback to (we can make only one savepoint, because we don't work with a stack)  */
        savepoint=new SavePoint(list,listelements);
        // Test if we need not use the trier
        if (analyzeNext(todo))
            // We don't need to use the trier
            return true;
        else
            // Do a rollback of modifications in listelements and list
            savepoint.rollback();
        // We must try the trier one or more times untill we have a failing of the last analyze of the trier
        // Add a special analyzer to check if it is called
       	trytodo=todo.cons(testCall,list,listelements);
        // Now try to add different instances to the trier until we have a success or until testCall is not
        // called anymore (to avoid infinite looping we do not loop more then the maximum number of elements in listelements
        int m=listelements.getNextCount();
        for (int i=0;i<m;i++)
        {
            // Add a extra analyze step by trier
            trytodo=trytodo.append(trier,list,listelements);
            // After this call testCall.isCalled() returns false until it is called
            testCall.setNotCalled();
            // Try to do a analyze with the trytodo
            if (analyzeNext(trytodo))
            {
                return true;
            }
            else if (testCall.isCalled())
            {
                // We failed, but isCalled is still called so we can try one more trier
                savepoint.rollback();
            }
            else
            {
                // We failed because one of the triers failed , so we give up
                savepoint.rollback();
                return false;
            }
        };
        return false;
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		boolean result=false;
		SavePoint savepoint;
		// Analyzer to check if it is called
		TestCallLispAnalyzer testCall=new TestCallLispAnalyzer();
		AnalyzerList trytodo;
		// Make a savepoint to rollback to (we can make only one savepoint, because we don't work with a stack)  */
		savepoint=new SavePoint(list,listelements);
		// Find the possible hints without a try
		result=findHintsNext(todo,element,hints) || result;
		// Do a rollback of modifications in listelements and list
		savepoint.rollback();
		// We must try the trier one or more times untill we have a failing of the last analyze of the trier
		// Add a spacial analyzer to check if it is called
		trytodo=todo.cons(testCall,list,listelements);
		// Now try to add different instances to the trier until we have a success or until testCall is not
		// called anymore
		int m=listelements.getNextCount();
		for (int i=0;i<=m;i++)
		{
			// Add a extra analyze step by trier
			trytodo=trytodo.append(trier,list,listelements);
			// After this call testCall.isCalled() returns false until it is called
			testCall.setNotCalled();
			// Try to do a analyze with the trytodo
			result=findHintsNext(trytodo,element,hints) || result;
			// Test if testCall is still called
			if (testCall.isCalled())
			{
				savepoint.rollback();
			}
			else
			{
				// We failed because one of the triers failed , so we give up
				savepoint.rollback();
				return result;
			}
		}
		return false;
	}
}

/** Analyzer to try zero or one applies of the same analyzer */

class TryOnceLispAnalyzer extends LispAnalyzer
{
    LispAnalyzer[] trier;
    public TryOnceLispAnalyzer(LispAnalyzer[] _try)
    {
        super();
        trier=_try;
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        SavePoint savepoint;
        // Set a savepoint so we can rollback if needed
        // Check if we can do without the trier
        savepoint=new SavePoint(list,listelements);
        // Try without the trier
        if (analyzeNext(todo))
        {
            return true;
        }
        else
        {
            savepoint.rollback();
            return analyzeNext(todo.append(trier,list,listelements));
        }
    }
    
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		boolean result=false;
		SavePoint savepoint;
		// Set a savepoint so we can rollback if needed
		// Check if we can do without the trier
		savepoint=new SavePoint(list,listelements);
		// Try without the trier
		result=findHintsNext(todo,element,hints) || result;
		savepoint.rollback();
		return findHintsNext(todo.append(trier,list,listelements),element,hints) || result;
	}
}

/** Analyzer to try different Analyzers  */

class ChoiceLispAnalyzer extends LispAnalyzer
{
    LispAnalyzer[][] choices;
    public ChoiceLispAnalyzer(LispAnalyzer[][] _choices)
    {
        super();
        choices=_choices;
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        SavePoint savepoint;
        savepoint=new SavePoint(list,listelements);
        for (int i=0;i<choices.length;i++)
        {
            if (analyzeNext(todo.append(choices[i],list,listelements)))
                return true;
            else
                savepoint.rollback();
        };
        return false;
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		boolean result=false;
		SavePoint savepoint;

		savepoint=new SavePoint(list,listelements);
		for (int i=0;i<choices.length;i++)
		{
			result=findHintsNext(todo.append(choices[i],list,listelements),element,hints) || result;
			savepoint.rollback();
		};
		return result;
	}
}
/** Analyzer of a branch */
class BranchLispAnalyzer extends LispAnalyzer
{
    LispAnalyzer[] branch;
    
    public BranchLispAnalyzer(LispAnalyzer[] _branch)
    {
        super();
        branch=_branch;
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        Object lst;
        lst=listelements.next();
        if (lst instanceof LispListElement)
            return analyzeNext(todo.append(branch,list,((LispListElement) lst).getElements().makeIterator()));
        else
            return false;
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		Object lst;
        
		lst=listelements.next();
		if (lst==element)
		{
			hints.add("()");
			return true;
		}
		else if (lst instanceof LispListElement)
		{
			return findHintsNext(todo.append(branch,list,((LispListElement) lst).getElements().makeIterator()),element,hints);
		}
		else
		{
			return false;
		}	
	}
}

/** Analyzer based on a sequence of analyzers which must be executed after each other */
class SequenceLispAnalyzer extends LispAnalyzer
{
    LispAnalyzer sequence[];
    
    public SequenceLispAnalyzer(LispAnalyzer[] _sequence)
    {
        super();
        sequence=_sequence;
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        return analyzeNext(todo.append(sequence,list,listelements));
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		return findHintsNext(todo.append(sequence,list,listelements),element,hints);
	}
}

/** Analyzer to check if we are at the end of the list */

class EndLispAnalyzer extends LispAnalyzer
{
    public EndLispAnalyzer()
    {
        super();
    }
    public boolean analyze(LispListElement list,CodeElementIterator listelements,AnalyzerList todo)
    {
        if (listelements.peek()==null)
            return analyzeNext(todo);
        else
            return false;
    }
	public boolean findHints(CodeElement element,LispListElement list,CodeElementIterator listelements,Vector hints,AnalyzerList todo)
	{
		if (listelements.peek()==null)
			return findHintsNext(todo,element,hints);
		else
			return false;
	}   
}