/*
 * ReportError.java
 *
 * Created on May 20, 2001, 8:59 AM
 */

package IDE.Messaging;

import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.util.*;
import javax.swing.*;
import javax.swing.text.*;
import IDE.Utilities.*;
import IDE.CodeEditor.*;

/**
 * This class represent the message area of the IDE
 * It contains 3 components
 * - errorandwarnings : A combo box containing the errors and warnings generated
 * - messagefield : A text label containing the information messages
 * - input : A instruction field for input
 *
 * @author  marc
 * @version
 */
public class MessageHandler extends JSplitPane implements ItemListener
{
    /** Contains the log file to write debugging informtion in */
    private static File logfile;
    /**
     * Refer to the MessageHandler itself so we can make a instance of the
     * handler (only one may be made) and refer to it from static methods
     */
    private static MessageHandler message=null;
    /** Label field to display messages in */
    private static JLabel messagefield;
    /** Drop downbox for the errors and warnings */
    private static JComboBox errorandwarnings;
    /** A instruction field for throws input in the message area */
    private static InstructionField instructionfield;
    /** A splitpane to contain the messagefield and the input field */
    private static JSplitPane messageInput;
    /** Display the current mode */
    private static JLabel modeIndicator;
    /** Create message handler */
    public MessageHandler()
    {
        logfile=new File(System.getProperty("user.home")+File.separator+".Jabberwocky"+File.separator+"error.log");
        messagefield=new JLabel();
        messagefield.setBorder(BorderFactory.createEtchedBorder());
        modeIndicator=new JLabel();
        modeIndicator.setBorder(BorderFactory.createEtchedBorder());
        errorandwarnings=new JComboBox();
        errorandwarnings.setOpaque(true);
        setOpaque(true);
        errorandwarnings.addItem("");
        errorandwarnings.addItemListener(this);
        errorandwarnings.setBorder(BorderFactory.createEtchedBorder());
        instructionfield=new InstructionField();
        messageInput=new JSplitPane();
        
        messageInput.add(messagefield,JSplitPane.LEFT);
        messageInput.add(instructionfield,JSplitPane.RIGHT);
        add(errorandwarnings,JSplitPane.LEFT);
        
        JSplitPane rightPane=new JSplitPane();
        rightPane.add(messageInput,JSplitPane.LEFT);
        rightPane.add(modeIndicator,JSplitPane.RIGHT);
        rightPane.setResizeWeight(1); // Left componnent get all the available space
        rightPane.setDividerSize(3);
        add(rightPane,JSplitPane.RIGHT);
        
        messageInput.setDividerLocation(0.5);
        messageInput.setResizeWeight(0.5);
        messageInput.setDividerSize(3);
        setDividerLocation(0.25);
        setResizeWeight(0.25);
        setDividerSize(3);
        message=this;
        
        resetSize();
    }
    /** Log the error to the error log file */
    public static void logError(String error,Exception e)
    {
        try
        {
            BufferedWriter writer=new BufferedWriter(new FileWriter(Util.getCanonicalFileName(logfile),true));
            writer.write(error+"\n");
            if (e!=null)
            {
                StackTraceElement[] elements=e.getStackTrace();
                for (int i=0;i<elements.length;i++)
                    writer.write(elements[i].toString()+"\n");
            };
            writer.close();
        }
        catch (Exception ex)
        {
            // No way to report this error
        };
    }
    /**
     * Report the error text and the event stack trace, used for reporting
     * unexpected events in the program, do this in the thread of the GUI
     * so we can call this from another thread
     */
    public static void report(final String errorstr,final Exception e)
    {
        Runnable doreport=new Runnable()
        {
            public void run()
            {
                String error=getTime()+" "+errorstr;
                errorandwarnings.setBackground(Color.green);
                errorandwarnings.setForeground(Color.red);
                errorandwarnings.addItem(error);
                errorandwarnings.setSelectedItem(error);
                e.printStackTrace();
                resetSize();
            }
        };
        if (message!=null)
            SwingUtilities.invokeLater(doreport);
        logError(errorstr,e);
    }
    /**
     * Report the error text and the event stack trace
     * and exit the program with status -1, used to report fatal errors
     * and recover gracefull from them
     */
    public static void reportAndExit(String errorstr,Exception e)
    {
        if (message!=null)
            report(errorstr,e);
        System.exit(-1);
    }
    /**
     * Used to talk about a detectable syntax error in a source
     * @param errorstr
     * @param document
     * @param position
     */
    public static void codeError(String errorstr,CodeDocument document,Position position)
    {
    	warn(errorstr+" in "+document.toString()+" at "+position.getOffset());
    }
	/**
	 * Used to talk about a detectable syntax warning in a source
	 * @param errorstr Warning to give
	 * @param document
	 * @param position
	 */
	public static void codeWarn(String errorstr,CodeDocument document,Position position)
	{
		warn(errorstr+" in "+document.toString()+" at "+position.getOffset());
	}   
	 /**
     * Report user induced errors to the user, do this in the thread from the GUI
     * so we can call this from another thread
     */
    public static void warn(final String errorstr)
    {
        Runnable dowarn=new Runnable()
        {
            public void run()
            {
                String error=getTime()+" "+errorstr;
                errorandwarnings.addItem(error);
                errorandwarnings.setSelectedItem(error);
                errorandwarnings.setBackground(Color.cyan);
                errorandwarnings.setForeground(Color.red);
                resetSize();
            }
        };
        if (message!=null)
            SwingUtilities.invokeLater(dowarn);
    }
    /** Display the new mode, occurs when the mode is changed */
    public static void displayMode(final String mode)
    {
        Runnable domode=new Runnable()
        {
            public void run()
            {
                modeIndicator.setText(mode);
            }
        };
        SwingUtilities.invokeLater(domode);
    }
    /**
     * Give a message to the user, do this in the thread from the GUI
     * se we can call this from another thread
     */
    public static void say(final String str)
    {
        Runnable dosay=new Runnable()
        {
            public void run()
            {
                if (str.length()>0)
                {
                    errorandwarnings.setBackground(Color.lightGray);
                    errorandwarnings.setForeground(Color.black);
                };
                messagefield.setText(str);
                messagefield.setToolTipText(str);
                messagefield.repaint();
                resetSize();
            }
        };
        if (message!=null)
            SwingUtilities.invokeLater(dosay);
    }
    /*
     * Clear the error zone
     */
    public static void clear()
    {
        errorandwarnings.setBackground(Color.lightGray);
        errorandwarnings.setForeground(Color.black);
        errorandwarnings.setSelectedIndex(0);
    }
    /** Get the time of the day, used for reporting errors */
    private static String getTime()
    {
        Calendar calendar=Calendar.getInstance();
        return calendar.get(Calendar.HOUR_OF_DAY)+":"+calendar.get(Calendar.MINUTE)+":"+calendar.get(Calendar.SECOND);
    }
    /** Reset the minimum and preferred sizes */
    private static void resetSize()
    {
        Font font=messagefield.getFont();
        FontMetrics fm=messagefield.getFontMetrics(font);
        Dimension dim=new Dimension(50,fm.getHeight()+8);
        // Set minimum sizes
        messagefield.setMinimumSize(dim);
        errorandwarnings.setMinimumSize(dim);
        instructionfield.setMinimumSize(dim);
        messageInput.setMinimumSize(dim);
        // Set the preferred sizea
        messagefield.setPreferredSize(dim);
        errorandwarnings.setPreferredSize(dim);
        instructionfield.setPreferredSize(dim);
        messageInput.setPreferredSize(dim);
    }
    /** Called when the combobox is changed */
    public void itemStateChanged(ItemEvent ev)
    {
        String message=(String) errorandwarnings.getSelectedItem();
        errorandwarnings.setToolTipText(message);
    }
}
