/*
 * Created on Aug 18, 2003
 */
 
package IDE.Project;

import javax.swing.*;
import javax.swing.text.*;
import java.util.*;

/**
 * @author marc 
 * Represents Lisp packages
 */
public class LispPackage implements INode,Comparable
{
	/** Name of the package, null for the default package */
	private String name;
	/** The documentation of this package */
	private String documentation;
	/** Definition instance indicating where the package is created (null for the default package) */
	private IDefinition definition;
	/** The default LispPackage */
	public static final LispPackage defaultPackage=new LispPackage(null,null);
	/** The interned symbols in this package */
	private final ArrayList interned=new ArrayList();
	/** The exported symbols in this package */
	private final ArrayList exported=new ArrayList();
	/** The shadowed symbols in this package */
	private final ArrayList shadowed=new ArrayList();
	/** The packages used by this package */
	private final ArrayList usedPackages=new ArrayList();
	/** The node containing the interned definitions */
	public final static InternedNode internedNode=new InternedNode();
	/** The node containing the exported defintions */
	public final static ExportedNode exportedNode=new ExportedNode();
	/** The node containing the shadowed definitions */
	public final static ShadowedNode shadowedNode=new ShadowedNode();
	/** The node containing the used packages */
	public final static UsedNode usedNode=new UsedNode();
	public LispPackage(String name,IDefinition definition)
	{
		this.name=name;
		this.definition=definition;
	}
	/**
	 * Set the documentation of this package
	 * @param documentation The documentation of this package
	 */
	public void setDocumentation(final String documentation,Position begin,int length)
	{
		final String oldDocumentation=this.documentation;
		new AbstractDefinition(begin,length)
		{
			public void undoDefinition()
			{
				LispPackage.this.documentation=oldDocumentation;
			}
		};
		this.documentation=documentation;
	}
	public void usePackage(final LispPackage usedPackage,Position begin,int length)
	{
		// Undo used package
		new AbstractDefinition(begin,length)
		{
			public void undoDefinition()
			{
				usedPackages.remove(usedPackage);
			}
		};
		// Add used package
		usedPackages.add(usedPackage);
		// Create a path for the used package
		FileReference.registerTreePath(new INode[] {LispRepository.lisprepository,this,usedNode,usedPackage});
	}
	/**
	 * Intern the symbol in this LispPackage
	 * @param symbol to be interned
	 * @param begin Begin position of the text that defines the intern of the symbol
	 * @param length Length of the text that defines the intern of the symbol
	 * @return The interned symbol node 
	 */
	public LispSymbolDefinitionNode internSymbol(final LispSymbol symbol,String type,Position begin,int length)
	{
		// Create the lispnode and interns it
		LispSymbolDefinitionNode lispnode=new LispSymbolDefinitionNode(symbol,type,begin,length)
		{
			public void undoDefinition()
			{
				interned.remove(this);
				symbol.packagesInterned.remove(LispPackage.this);
			}
		};
		interned.add(lispnode);
		symbol.packagesInterned.add(this);
		return lispnode; 
	}
	/**
	 * Intern the symbol in this LispPackage
	 * @param symbol to be interned
	 * @param begin Begin position of the text that defines the intern of the symbol
	 * @param length Length of the text that defines the intern of the symbol
	 * @return The interned symbol node 
	 */
	public LispSymbolDefinitionNode internSymbol(final LispSymbol symbol,Position begin,int length)
	{
		// Create the lispnode and interns it
		LispSymbolDefinitionNode lispnode=new LispSymbolDefinitionNode(symbol,begin,length)
		{
			public void undoDefinition()
			{
				interned.remove(this);
				symbol.packagesInterned.remove(LispPackage.this);
			}
		};
		interned.add(lispnode);
		symbol.packagesInterned.add(this);
		return lispnode; 
	}
	/** Export the symbol */
	public LispSymbolDefinitionNode exportSymbol(final LispSymbol symbol,Position begin,int length)
	{
		// Create the lispnode and export it
		LispSymbolDefinitionNode lispnode=new LispSymbolDefinitionNode(symbol,begin,length)
		{
			public void undoDefinition()
			{
				exported.remove(this);
				symbol.packagesExported.remove(LispPackage.this);
			}
		};
		exported.add(lispnode);
		symbol.packagesExported.add(this);
		return lispnode;
	}
	/** Shadowimport of the symbol */
	public LispSymbolDefinitionNode shadowImportSymbol(final LispSymbol symbol,Position begin,int length)
	{
		// Create the lispnode and export it
		LispSymbolDefinitionNode lispnode=new LispSymbolDefinitionNode(symbol,begin,length)
		{
			public void undoDefinition()
			{
				shadowed.remove(this);
				symbol.packagesShadowed.remove(LispPackage.this);
			}
		};
		shadowed.add(lispnode);
		symbol.packagesShadowed.add(this);
		return lispnode;
	}
	/**
	 * Get the arraylist of the used package 
	 * @return The array list of the used package
	 */
	protected ArrayList getUsedPackages()
	{
		return usedPackages;
	}
	/**
	 * Returns a ArrayList of the concrete definitions of interned symbols
	 * @return A Array of concrete definitions
	 */
	protected ArrayList getDefinedInternedSymbols()
	{
		return getDefinedSymbols(interned);
	}
	/**
	 * Returns a ArrayList of the concrete definitions of exported symbols
	 * @return A Array of concrete definitions
	 */
	protected ArrayList getDefinedExportedSymbols()
	{
		return getDefinedSymbols(exported);
	}
	/**
	 * Returns a ArrayList of the concrete definitions of shadowed symbols
	 * @return A Array of concrete definitions
	 */
	protected ArrayList getDefinedShadowdSymbols()
	{
		return getDefinedSymbols(shadowed);
	}
	/**
	 * Given a ArrayList of definitions get a ArrayList of all the concrete definitions associated with this symbol
	 * @param symbols A array list of symbols to get the concrete definition for
	 * @return The array of all the symbols
	 */
	private ArrayList getDefinedSymbols(ArrayList symbols)
	{
		ArrayList definitions=new ArrayList();
		LispSymbolDefinitionNode[] symboldefinitions;
		int m=symbols.size();
		for (int i=0;i<m;i++)
		{
			symboldefinitions=((LispSymbolDefinitionNode) symbols.get(i)).getSymbol().getDefinitions();
			for (int j=0;j<symboldefinitions.length;j++)
				definitions.add(symboldefinitions[j]);
		};
		return definitions;
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getUniqueName()
	 */
	public String getUniqueName() 
	{
		if (name==null)
			return "common-lisp-user";
		else
			return name;
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getDisplayName()
	 */
	public String getDisplayName() 
	{
		return getUniqueName();
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getIconFile()
	 */
	public String getIconFile() 
	{
		return "package.png";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getPopupMenu()
	 */
	public JPopupMenu getPopupMenu() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getTooltip()
	 */
	public String getTooltip() 
	{
		if (documentation==null)
			return "Lisp package "+getDisplayName();
		else
			return documentation;
	}

	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object o) 
	{
		return getUniqueName().compareTo(((LispPackage) o).getUniqueName());
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#doubleClick()
	 */
	public void doubleClick() 
	{
		// TODO Auto-generated method stub	
	}
}
class InternedNode implements INode
{

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getUniqueName()
	 */
	public String getUniqueName() 
	{
		return "Interned Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getDisplayName()
	 */
	public String getDisplayName() 
	{
		return "Interned Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getIconFile()
	 */
	public String getIconFile() 
	{
		return "directory.png";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getPopupMenu()
	 */
	public JPopupMenu getPopupMenu() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getTooltip()
	 */
	public String getTooltip() 
	{
		return "Interned Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#doubleClick()
	 */
	public void doubleClick() {
		// TODO Auto-generated method stub
		
	}
	
}
class ExportedNode implements INode
{
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getUniqueName()
	 */
	public String getUniqueName() 
	{
		return "Exported Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getDisplayName()
	 */
	public String getDisplayName() 
	{
		return "Exported Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getIconFile()
	 */
	public String getIconFile() 
	{
		return "directory.png";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getPopupMenu()
	 */
	public JPopupMenu getPopupMenu() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getTooltip()
	 */
	public String getTooltip() 
	{
		return "Exported Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#doubleClick()
	 */
	public void doubleClick() {
		// TODO Auto-generated method stub
		
	}
	
}
class ShadowedNode implements INode
{

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getUniqueName()
	 */
	public String getUniqueName() 
	{
		return "Shadowed Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getDisplayName()
	 */
	public String getDisplayName() 
	{
		return "Shadowed Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getIconFile()
	 */
	public String getIconFile() 
	{
		return "directory.png";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getPopupMenu()
	 */
	public JPopupMenu getPopupMenu() {
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getTooltip()
	 */
	public String getTooltip() 
	{
		return "Shadowed Symbols";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#doubleClick()
	 */
	public void doubleClick() {
		// TODO Auto-generated method stub
		
	}
}
	class UsedNode implements INode
	{

		/* (non-Javadoc)
		 * @see IDE.Project.INode#getUniqueName()
		 */
		public String getUniqueName() 
		{
			return "Used Packages";
		}

		/* (non-Javadoc)
		 * @see IDE.Project.INode#getDisplayName()
		 */
		public String getDisplayName() 
		{
			return "Used Packages";
		}

		/* (non-Javadoc)
		 * @see IDE.Project.INode#getIconFile()
		 */
		public String getIconFile() 
		{
			return "directory.png";
		}

		/* (non-Javadoc)
		 * @see IDE.Project.INode#getPopupMenu()
		 */
		public JPopupMenu getPopupMenu() {
			// TODO Auto-generated method stub
			return null;
		}

		/* (non-Javadoc)
		 * @see IDE.Project.INode#getTooltip()
		 */
		public String getTooltip() 
		{
			return "Used packages";
		}

		/* (non-Javadoc)
		 * @see IDE.Project.INode#doubleClick()
		 */
		public void doubleClick() {
			// TODO Auto-generated method stub
			
		}
}