/*
 * Created on Aug 18, 2003
 */
 
package IDE.Project;

import javax.swing.*;
import javax.swing.text.*;
import java.util.*;

import IDE.Main;
import IDE.CodeEditor.CodeEditor;
import IDE.CodeEditor.Parser.*;
import IDE.Utilities.Util;

/**
 * @author marc
 * Represents the repository that contains all the Lisp definitions, only one can exist
 */
public class LispRepository extends AbstractRepository implements INode
{
	/** Mapping of the name of a package to a package instance to maintain a unique list of packages */
	private SortedMap nameToPackage=new TreeMap();
	/** Mapping of a LispSymbol to itself to maintain a unique list of symbols */
	private SortedMap symbolToSymbol=new TreeMap();
	/** The only instance of the lisp repository */
	public final static LispRepository lisprepository=new LispRepository();
	/** A mapping of a path to a mapping of a package to the list of its callers */
	private SortedMap pathToPackageToCallers=new TreeMap();
	/** The current package (part of the state during analyzing a lisp source) */
	private static LispPackage currentPackage;
	/** Contains a mapping of package to all the callers (functions, macros in a form) for the current package (is part of the state) */
	private static SortedMap packageToCallers;
	private LispRepository()
	{
		nameToPackage.put(LispPackage.defaultPackage.getUniqueName(),LispPackage.defaultPackage);
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getUniqueName()
	 */
	public String getUniqueName() 
	{
		return "Lisp";
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getDisplayName()
	 */
	public String getDisplayName() 
	{
		return "Lisp";
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getIconFile()
	 */
	public String getIconFile() 
	{
		return "big_projects.png";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getPopupMenu()
	 */
	public JPopupMenu getPopupMenu() 
	{
		// TODO Auto-generated method stub
		return null;
	}

	/* (non-Javadoc)
	 * @see IDE.Project.INode#getTooltip()
	 */
	public String getTooltip() 
	{
		return "Lisp definitions";
	}

	/* (non-Javadoc)
	 * @see IDE.Project.IRepository#supportFile(java.lang.String)
	 */
	public boolean supportFile(String filename) 
	{
		filename=filename.toUpperCase();
		if (filename.endsWith(".LSP") || filename.endsWith(".LISP"))
			return true;
		else
			return false;
	}

	/* (non-Javadoc)
	 * @see IDE.Project.IRepository#createParser(boolean)
	 */
	public Parser createParser(boolean assync) 
	{
		return new LispParser(assync);
	}
	/**
	 * Tries to find a package with the name name and if not found creates a new one
	 * and register it so that it will be found the next time
	 * @param lu_name Name of the package
	 * @return The package with name name
	 */
	private LispPackage getOrCreatePackage(final String lu_name,Position begin,int length)
	{
		final String name=lu_name.toLowerCase();
		LispPackage result=(LispPackage) nameToPackage.get(name);
		if (result==null)
		{
			IDefinition definition=new AbstractDefinition(begin,length)
			{
				public void undoDefinition()
				{
					nameToPackage.remove(name);
				}
			};
			result=new LispPackage(name,definition);
			nameToPackage.put(name,result);
			// Register a treepath displaying this package
			FileReference.registerTreePath(new INode[] {this,result});
		};
		return result;
	}
	/**
	 * Tries to find a symbol and if not found creates a new one and register it so it will be found the next time
	 * @param name Name of the symbol
	 * @param homepackage The home package of this symbol 
	 * @return The symbol found or created 
	 */
	private LispSymbol getOrCreateSymbol(String lu_name,final LispPackage homepackage)
	{
		final String name=lu_name.toLowerCase();
		final String packageName=homepackage.getUniqueName();
		LispSymbol symbol=(LispSymbol) symbolToSymbol.get(LispSymbol.getKeySymbol(name,packageName));
		if (symbol==null)
		{
			symbol=new LispSymbol(name,homepackage);
			symbolToSymbol.put(symbol,symbol);
			// Create a definition with a nul reference to its position and length (we don't associate a INode with it, we only 
			// need the undo
			new AbstractDefinition(null,-1)
			{
				public void undoDefinition()
				{
					symbolToSymbol.remove(LispSymbol.getKeySymbol(name,packageName));
				}
			};
		};
		return symbol;
	}
	/**
     * Export of the symbols in exportedsymbols for a given namespace
	 * @param exportedsymbols
	 * @param from Package to export symbols from
	 * @param begin Begin position of the defining text of the export
	 * @param length Length of the defining text of the export
	 */
	public void exportSymbols(String[] exportedsymbols, String from, Position begin, int length) 
	{
		LispSymbol symbol;
		if (from==null)
			from=LispPackage.defaultPackage.getUniqueName();
		LispPackage lisppackage=getOrCreatePackage(from,begin,length);
		for (int i=0;i<exportedsymbols.length;i++)
		{
			symbol=getOrCreateSymbol(exportedsymbols[i],lisppackage);
			lisppackage.exportSymbol(symbol,begin,length);
		};
	}

	/**
	 * @param packageToImportTo The package where the symbols must be interned on
	 * @param originatingPackage The package containing the symbol
	 * @param symbols Array of the symbols to import
	 * @param begin Begin position of the defining text
	 * @param length Length of the definining text
	 */
	public void importSymbols(String packageToImportTo, String originatingPackage, String[] symbols, Position begin, int length)
	{
		LispPackage inPackage=getOrCreatePackage(packageToImportTo,begin,length);
		LispPackage outPackage=getOrCreatePackage(originatingPackage,begin,length);
		LispSymbol symbol;
		for (int i=0;i<symbols.length;i++)
		{
			symbol=getOrCreateSymbol(symbols[i],outPackage);
			inPackage.internSymbol(symbol,begin,length);
		};
	}

	/**
	 * @param packageAndSymbols Array of packages and the interned symbols
	 * @param packageToImportTo Package to intern the symbol on
	 * @param begin Begin position of the definition
	 * @param length Length of the definining text
	 */
	public void importSymbols(String[][] packageAndSymbols, String packageToImportTo, Position begin, int length) 
	{
		LispPackage inPackage=getOrCreatePackage(packageToImportTo,begin,length);
		LispPackage outPackage;
		LispSymbol symbol;
		for (int i=0;i<packageAndSymbols.length;i++)
		{
			outPackage=getOrCreatePackage(packageAndSymbols[i][0],begin,length);
			symbol=getOrCreateSymbol(packageAndSymbols[i][1],outPackage);
			inPackage.internSymbol(symbol,begin,length);
		};
	}
	/**
	 * Defines the nicknames for a given package
	 * @param namePackage Name of the package to set the nicknames for
	 * @param nicknames Array of strings representing the nicknames
	 * @param begin Begin position of the definition
	 * @param length Length of the definition
	 */
	public void setNickNamesForPackage(final String namePackage,final String[] nicknames,Position begin,int length)
	{
		final LispPackage lisppackage=getOrCreatePackage(namePackage,begin,length);
		// Create definition fo undo of this action
		new AbstractDefinition(begin,length)
		{
			public void undoDefinition()
			{
				for (int i=0;i<nicknames.length;i++)
					if (namePackage.compareTo(nicknames[i])!=0)
						nameToPackage.remove(nicknames[i]);
			}
		};
		// Create the nicknames
		for (int i=0;i<nicknames.length;i++)
			if (namePackage.compareTo(nicknames[i])!=0)
				nameToPackage.put(nicknames[i],lisppackage);
	}
	/**
	 * Set the documentation for a given package
	 * @param namePackage Name of the package to set the documentation string
	 * @param documentation The documentation string of the package
	 * @param begin Begin position of the definition
	 * @param length Length of the definition
	 */
	public void setDocumentationForPackage(String namePackage,String documentation,Position begin,int length)
	{
		LispPackage lisppackage=getOrCreatePackage(namePackage,begin,length);
		lisppackage.setDocumentation(documentation,begin,length);
	}
	/**
	 * Do a shadow import of symbols in a package
	 * @param namePackage The package to do the shadow import in
	 * @param fromPackage The package containing the symbols
	 * @param symbols Array of the symbols to shadow import
	 * @param begin Begin position of the document
	 * @param length Length of the definition
	 */
	public void shadowImportSymbols(String namePackage, String fromPackage, String[] symbols, Position begin, int length) 
	{
		LispPackage inPackage=getOrCreatePackage(namePackage,begin,length);
		LispPackage outPackage=getOrCreatePackage(fromPackage,begin,length);
		LispSymbol symbol;
		for (int i=0;i<symbols.length;i++)
		{
			symbol=getOrCreateSymbol(symbols[i],outPackage);
			inPackage.shadowImportSymbol(symbol,begin,length);
		};
	}

	/**
	 * Let a package use another package
	 * @param namePackage Package that uses another package
	 * @param usedPackage Package that is used by namePAckage
	 * @param begin Begin position of the definition
	 * @param length Length of the definition
	 */
	public void useNameSpace(String namePackage, String usedPackage, Position begin, int length) 
	{
		LispPackage lisppackage=getOrCreatePackage(namePackage,begin,length);
		LispPackage used=getOrCreatePackage(usedPackage,begin,length);
		lisppackage.usePackage(used,begin,length);
	}
	/**
	 * Called to indicate that we must make another package current
	 * @param namePackage Name of the package to make current
	 * @param begin Begin position of the definition
	 * @param length Length of the definition
	 */
	public void inPackage(String namePackage,Position begin,int length)
	{
		// Set the current package
		currentPackage=getOrCreatePackage(namePackage,begin,length);
		// Create a definition to mark the position
		FileReference.registerDefinition(new LispInPackageDefinition(namePackage,begin,length));
	}
	/**
	 * @param type Type of the definition
	 * @param name Name of the symbol that is associated with the definition
	 * @param position  Begin position of the defining text
	 * @param length Length of the defining text
	 */
	public void addDefinition(String type, String name, Position position, int length) 
	{
		// Get or create the symbol
		LispSymbol symbol=getOrCreateSymbol(name,currentPackage);
		symbol.registerUsage(type,position,length); // Register the usage of the symbol
		// Intern the symbol in the correct package
		LispSymbolDefinitionNode node=currentPackage.internSymbol(symbol,type,position,length);
		// The treepath for the different projects
		ArrayList[] treepaths=Projects.calculateTreePathsForDefinitionNode(node);
		INode[] treepath;
		LispTypeNode typenode=LispTypeNode.getOrCreateType(type);
		for (int i=0;i<treepaths.length;i++)
		{
			treepath=new INode[treepaths[i].size()+2];
			for (int j=0;-1<j && j<treepath.length-2;j++)
				treepath[j]=(INode) treepaths[i].get(j);
			treepath[treepath.length-2]=typenode;
			treepath[treepath.length-1]=node;
			FileReference.registerTreePath(treepath);
		};
	}
	/**
	 * @param name The name of the package to create
	 * @param begin The position in a CodeDocument pointing to the start of statement defining a package
	 * @param length The length of the text defining a package
	 */
	public void createPackage(String name,Position begin,int length) 
	{
		getOrCreatePackage(name,begin,length);
	}
	/**
	 * Register a call in the current package
	 * @param caller
	 */
	public void registerCaller(String caller)
	{
		String packageName=currentPackage.getUniqueName();
		SortedSet callers=(SortedSet) packageToCallers.get(packageName);
		if (callers==null)
		{
			callers=new TreeSet();
			packageToCallers.put(packageName,callers);
		};
		callers.add(caller);
	}
	/* (non-Javadoc)
	 * @see IDE.Project.IRepository#beginAnalyze()
	 */
	public void beginAnalyze() 
	{
		currentPackage=LispPackage.defaultPackage;
		packageToCallers=new TreeMap();
	}
	/* (non-Javadoc)
	 * @see IDE.Project.IRepository#endAnalyze()
	 */
	public void endAnalyze() 
	{
		// Find all the callers  in the analyzed source and register them
		final String path=FileReference.getPath();
		new AbstractDefinition()
		{
			public void undoDefinition()
			{
				pathToPackageToCallers.remove(path);
			}
		};
		pathToPackageToCallers.put(path,packageToCallers);
		packageToCallers=null;
		// Loop over all the Symbols 
		Iterator iter=symbolToSymbol.keySet().iterator();
		LispSymbol symbol;
		LispSymbolDefinitionNode[] lispnodes;
		LispPackage[] packages;
		while (iter.hasNext())
		{
			symbol=(LispSymbol) iter.next();
			lispnodes=symbol.getDefinitionsFor(FileReference.getPath());
			// Creates the treepaths for the interened nodes
			packages=symbol.getPackagesInternedIn();
			for (int i=0;i<packages.length;i++)
			{
				for (int j=0;j<lispnodes.length;j++)
					FileReference.registerTreePath(new INode[]{this,packages[i],LispPackage.internedNode,symbol,lispnodes[j]});
				if (lispnodes.length==0)
					FileReference.registerTreePath(new INode[]{this,packages[i],LispPackage.internedNode,symbol});
			}
			// Creates the treepaths for the exported nodes
			packages=symbol.getPackagesExportedIn();
			for (int i=0;i<packages.length;i++)
			{
				for (int j=0;j<lispnodes.length;j++)
					FileReference.registerTreePath(new INode[]{this,packages[i],LispPackage.exportedNode,symbol,lispnodes[j]});
				if (lispnodes.length==0)
					FileReference.registerTreePath(new INode[]{this,packages[i],LispPackage.exportedNode,symbol});
			}
			// Creates the treepaths for the shadowed nodes
			packages=symbol.getPackagesShadowedIn();
			for (int i=0;i<packages.length;i++)
			{
				for (int j=0;j<lispnodes.length;j++)
					FileReference.registerTreePath(new INode[]{this,packages[i],LispPackage.shadowedNode,symbol,lispnodes[j]});
				if (lispnodes.length==0)
					FileReference.registerTreePath(new INode[]{this,packages[i],LispPackage.shadowedNode,symbol});
			}
		};
	}
	/**
	 * Find the current package for position position in a source with path path.
	 * @param path The path of the source
	 * @param position The position where we have to find the current package for
	 * @return A string representing the current package
	 */
	public String findCurrentPackage(String path,int position)
	{
		IDefinition[] inPackageDefinitions=FileReference.getDefinitionsOfType(path,LispInPackageDefinition.class);
		int i;
		for (i=0;i<inPackageDefinitions.length && inPackageDefinitions[i].getOffset()<position;i++);
		if (i-1<inPackageDefinitions.length && i>0)
			return ((LispInPackageDefinition) inPackageDefinitions[i-1]).getPackageName().toLowerCase();
		else
			return LispPackage.defaultPackage.getUniqueName().toLowerCase();
	}
	public SortedSet getUndefinedCallers()
	{
		CodeEditor current=Main.currentEditor();
		if (current==null)
		{
			return getUndefinedCallers(null);
		}
		else
		{
			return getUndefinedCallers(findCurrentPackage(Util.getCanonicalFileName(current.getDocument().getFile()),current.cursor.cursorOffset()));
		}
	}
	/**
	 * Get all the callers that are not defined as a defun,defmacro,defmethod,defgeneric in the given package
	 * @param packageName The name of the package where we want to find the undefined for
	 * @return
	 */
	public SortedSet getUndefinedCallers(String packageName)
	{
		SortedSet callers=getCallers(packageName);
		// Remove all the know definitions from the callers in this package
		String[] definitions=getDefinitionsOfType(packageName,new String[] {"defun","defmacro","defgeneric","defmethod"});
		for (int i=0;i<definitions.length;i++)
			callers.remove(definitions[i]);
		// The remaining elements are the undefined callers
		String[] result=new String[callers.size()];
		Iterator iter=callers.iterator();
		return callers;
	}
	/** 
	 * Get a SortedSet of strings containing all the callers in a certain package
	 * @param packageName The name of the package containing all the calls
	 * @return A sorted set of the callers name in the given package
	 */
	private SortedSet getCallers(String packageName)
	{
		if (packageName==null)
			packageName=LispPackage.defaultPackage.getUniqueName();
		SortedSet result=new TreeSet();
		SortedMap packageToCallers;
		String name;
		Iterator pathIter=pathToPackageToCallers.values().iterator();
		Iterator packageIter;
		while (pathIter.hasNext())
		{
			packageToCallers=(SortedMap) pathIter.next();
			packageIter=packageToCallers.keySet().iterator();
			while (packageIter.hasNext())
			{
				name=(String) packageIter.next();
				if (name.compareTo(packageName)==0)
					result.addAll((TreeSet) packageToCallers.get(name));
			}
		}
		return result;
	}
	/**
	 * Find all the definition names of the type mentioned in the array of types (for example new String[] {"defun","defmacro"} 
	 * finds all the defined macros. 
	 * @param types Array of strings representing the types of the definitions
	 * @return A array of definition names matching the types mentioned
	 */
	public String[] getDefinitionsOfType(String types[])
	{
		CodeEditor current=Main.currentEditor();
		if (current==null)
			return getDefinitionsOfType(null,types);
		else
			return getDefinitionsOfType(findCurrentPackage(Util.getCanonicalFileName(current.getDocument().getFile()),current.cursor.cursorOffset()),types);
	}
	/**
	 * Find all the definition names of the type mentioned in the array of types (for example new String[] {"defun","defmacro"} 
	 * finds all the defined macros. The packageName will point to the current package and is used to add a the correct package 
	 * prefix to the definition string.
	 * @param packageName the current package context from where the list of definitions is asked
	 * @param types Array of strings representing the types of the definitions
	 * @return A array of definition names matching the types mentioned
	 */
	public synchronized String[] getDefinitionsOfType(String packageName,String[] types)
	{
		Arrays.sort(types); // Ensure that types is sorted to make a  binary search possible
		SortedSet definitions=new TreeSet(DefinitionNameComparator.comparator);
		ArrayList definitionsFound;
		LispSymbolDefinitionNode definition;
		LispPackage lisppackage;
		Iterator iter;
		int m;
		// First work with the current package 
		LispPackage  currentPackage=(packageName==null) ? LispPackage.defaultPackage : (LispPackage) nameToPackage.get(packageName);
		if (currentPackage==null)
			return new String[0];
		// Interned symbols in the current package
		definitionsFound=currentPackage.getDefinedInternedSymbols();
		m=definitionsFound.size();
		for (int i=0;i<m;i++)
		{
			definition=(LispSymbolDefinitionNode) definitionsFound.get(i);
			if (types.length==0 || Arrays.binarySearch(types,definition.getType())>=0)
				definitions.add(definition.getSymbolName());
		};
		// Shadowed imported symbols in the current package
		definitionsFound=currentPackage.getDefinedShadowdSymbols();
		m=definitionsFound.size();
		for (int i=0;i<m;i++)
		{
			definition=(LispSymbolDefinitionNode) definitionsFound.get(i);
			if (types.length==0 || Arrays.binarySearch(types,definition.getType())>=0)
				definitions.add(definition.getSymbolName());
		};
		// The exported symbols of all the used packages
		ArrayList usedPackages=currentPackage.getUsedPackages();
		m=usedPackages.size();
		for (int i=0;i<m;i++)
		{
			lisppackage=(LispPackage) usedPackages.get(i);
			definitionsFound=lisppackage.getDefinedExportedSymbols();
			m=definitionsFound.size();
			for (int j=0;j<m;j++)
			{
				definition=(LispSymbolDefinitionNode) definitionsFound.get(j);
				if (types.length==0 || Arrays.binarySearch(types,definition.getType())>=0)
					definitions.add(definition.getSymbolName());
			};			
		}
		// Now consider all the other known packages
		iter=nameToPackage.values().iterator();
		SortedSet interned=new TreeSet(),exported=new TreeSet();
		String p_name;
		while (iter.hasNext())
		{
			lisppackage=(LispPackage) iter.next();
			if (lisppackage==currentPackage)
				continue; // We have already treated the current package
			p_name=lisppackage.getUniqueName();
			interned.clear();
			exported.clear();
			// Build the list of all the interned or shadowed imported symbols
			definitionsFound=lisppackage.getDefinedInternedSymbols();
			m=definitionsFound.size();
			for (int i=0;i<m;i++)
			{
				definition=(LispSymbolDefinitionNode) definitionsFound.get(i);
				if (types.length==0 || Arrays.binarySearch(types,definition.getType())>=0)
					interned.add(definition.getSymbolName());
			};
			definitionsFound=lisppackage.getDefinedShadowdSymbols();
			m=definitionsFound.size();
			for (int i=0;i<m;i++)
			{
				definition=(LispSymbolDefinitionNode) definitionsFound.get(i);
				if (types.length==0 || Arrays.binarySearch(types,definition.getType())>=0)
					interned.add(definition.getSymbolName());
			};
			// Build the list of all the exported symbols (removing them from the interned symbols)
			definitionsFound=lisppackage.getDefinedExportedSymbols();
			m=definitionsFound.size();
			for (int i=0;i<m;i++)
			{
				definition=(LispSymbolDefinitionNode) definitionsFound.get(i);
				if (types.length==0 || Arrays.binarySearch(types,definition.getType())>=0)
				{
					exported.add(definition.getSymbolName());
					interned.remove(definition.getSymbolName());
				};
			};
			// Now for the intered symbols not exported add the prefix package::
			Iterator internedIter=interned.iterator();
			while (internedIter.hasNext())
				definitions.add(p_name+"::"+(String) internedIter.next());
			// For the exported symbols add the prefix package:
			Iterator externedIter=exported.iterator();
			while (externedIter.hasNext())
				definitions.add(p_name+"::"+(String) externedIter.next());
		};
		// Build the result array
		iter=definitions.iterator();
		String[] result=new String[definitions.size()];
		for (int i=0;i<result.length;i++)
			result[i]=(String) iter.next();
		return result;
	}
	
	/* (non-Javadoc)
	 * @see IDE.Project.INode#doubleClick()
	 */
	public void doubleClick() 
	{
		// TODO Auto-generated method stub
		
	}
	/* (non-Javadoc)
	 * @see IDE.Project.IRepository#getExtensions()
	 */
	public String[] getExtensions() 
	{
		return new String[]{"lsp","lisp"};
	}
	/* (non-Javadoc)
	 * @see IDE.Project.IRepository#getHintComparator()
	 */
	public Comparator getHintComparator() 
	{
		return DefinitionNameComparator.comparator;
	}
	/* (non-Javadoc)
	 * @see IDE.Project.IRepository#getDefaultExtension()
	 */
	public String getDefaultExtension() 
	{
		return "lisp";
	}
}

