/*
 * Created on Aug 30, 2003
 */
 
package IDE.Project;

import java.util.*;

import IDE.CodeEditor.Parser.*;

import javax.swing.*;
import javax.swing.text.*;

/**
 * @author marc
 * Represents a lisp symbol
 */
public class LispSymbol implements Comparable,INode
{
	/** Name of the symbol */
	private String name;
	/** Package the symbol is defined in */
	private String homepackage;
	/** Definitions associating a class with this symbol */
	private ArrayList classDefinitions;
	/** Definitions associating a constant with this symbol */
	private ArrayList constantDefinitions;
	/** Definitions associating a generic function with this symbol */
	private ArrayList genericDefinitions;
	/** Definitions associating a macro with this symbol */
	private ArrayList macroDefinitions;
	/** Definitions associating a parameter with this symbol */
	private ArrayList parameterDefinitions;
	/** Definitions associating a struct with this symbol */
	private ArrayList structDefinitions;
	/** Definitions associating a type with this symbol */
	private ArrayList typeDefinitions;
	/** Definitions associating a function with this symbol */
	private ArrayList functionDefinitions;
	/** Definitions associating a variable with this symbol */
	private ArrayList variableDefinitions;
	/** Definitions associating a method with a symbol */
	private ArrayList methodDefinitions;
	/** Packages symbol is interned in */
	public final SortedSet packagesInterned=new TreeSet();
	/** Packages symbol is exported in */
	public final SortedSet packagesExported=new TreeSet();
	/** Packages symbol is shadowed in */
	public final SortedSet packagesShadowed=new TreeSet();
	/** Symbol used for comparisation */
	private final static LispSymbol keySymbol=new LispSymbol(null,null);
	public LispSymbol(String name,LispPackage homepackage)
	{
		this.name=name;
		if (homepackage!=null)
			this.homepackage=homepackage.getUniqueName();
	}
	/**
	 * Get a Symbol usefull for searching in a SortedSet of symbols
	 * @param name The name of the symbol
	 * @param homepackage The home package of the symbol
	 * @return
	 */
	public static LispSymbol getKeySymbol(String name,String homepackage)
	{
		keySymbol.name=name;
		keySymbol.homepackage=homepackage;
		return keySymbol;
	}
	/**
	 * Returns the packages this symbol is interned in
	 * @return The packages this symbol is interned in
	 */
	public LispPackage[] getPackagesInternedIn()
	{
		LispPackage[] res=new LispPackage[packagesInterned.size()];
		Iterator iter=packagesInterned.iterator();
		for (int i=0;i<res.length;i++)
			res[i]=(LispPackage) iter.next();
		return res;
	}
	/**
	 * Returns the packages this symbol is exported in
	 * @return The packages this symbol is exported in
	 */
	public LispPackage[] getPackagesExportedIn()
	{
		LispPackage[] res=new LispPackage[packagesExported.size()];
		Iterator iter=packagesExported.iterator();
		for (int i=0;i<res.length;i++)
			res[i]=(LispPackage) iter.next();
		return res;
	}
	/**
	 * Returns the packages this symbol is shadowed in
	 * @return The packages this symbol is shadowed in
	 */
	public LispPackage[] getPackagesShadowedIn()
	{
		LispPackage[] res=new LispPackage[packagesShadowed.size()];
		Iterator iter=packagesShadowed.iterator();
		for (int i=0;i<res.length;i++)
			res[i]=(LispPackage) iter.next();
		return res;
	}
	/**
	 * Returns the list of all the definitions of this symbol
	 * @return A array of the definition nodes for this symbol
	 */
	public LispSymbolDefinitionNode[] getDefinitions()
	{
		ArrayList definitions=new ArrayList();
		if (classDefinitions!=null)
			definitions.addAll(classDefinitions);
		if (constantDefinitions!=null)
			definitions.addAll(constantDefinitions);
		if (genericDefinitions!=null)
			definitions.addAll(genericDefinitions);
		if (macroDefinitions!=null)
			definitions.addAll(macroDefinitions);
		if (parameterDefinitions!=null)
			definitions.addAll(parameterDefinitions);
		if (structDefinitions!=null)
			definitions.addAll(structDefinitions);
		if (typeDefinitions!=null)
			definitions.addAll(typeDefinitions);
		if (functionDefinitions!=null)
			definitions.addAll(functionDefinitions);
		if (variableDefinitions!=null)
			definitions.addAll(variableDefinitions);
		if (methodDefinitions!=null)
			definitions.addAll(methodDefinitions);
		LispSymbolDefinitionNode[] result=new LispSymbolDefinitionNode[definitions.size()];
		for (int i=0;i<result.length;i++)
			result[i]=(LispSymbolDefinitionNode) definitions.get(i);
		return result;
		
	}
	/**
	 * Get the list of the definition nodes associated with this symbol
	 * @param path Path where the definition node must be defined in
	 * @return A array of definition nodes for this symbol
	 */
	public LispSymbolDefinitionNode[] getDefinitionsFor(String path)
	{
		ArrayList definitions=new ArrayList();
		findDefinitionsFor(path,classDefinitions,definitions);
		findDefinitionsFor(path,constantDefinitions,definitions);
		findDefinitionsFor(path,genericDefinitions,definitions);
		findDefinitionsFor(path,macroDefinitions,definitions);
		findDefinitionsFor(path,parameterDefinitions,definitions);
		findDefinitionsFor(path,structDefinitions,definitions);
		findDefinitionsFor(path,typeDefinitions,definitions);
		findDefinitionsFor(path,functionDefinitions,definitions);
		findDefinitionsFor(path,variableDefinitions,definitions);
		findDefinitionsFor(path,methodDefinitions,definitions);
		LispSymbolDefinitionNode[] result=new LispSymbolDefinitionNode[definitions.size()];
		for (int i=0;i<result.length;i++)
			result[i]=(LispSymbolDefinitionNode) definitions.get(i);
		return result;
	}
	private void findDefinitionsFor(String path,ArrayList in,ArrayList out)
	{
		if (in==null)
			return;
		int m=in.size();
		for (int i=0;i<m;i++)
			if (((LispSymbolDefinitionNode) in.get(i)).getPath().compareTo(path)==0)
				out.add(in.get(i));
	}
	/**
	 * Register the usage of a symbol for a certain type (defun,defmacro ...)
	 * @param type Type of the usage
	 * @param begin Begin position of the definition text of the usage
	 * @param length Length of the definition text of the usage
	 */
	public void registerUsage(String type,Position begin,int length)
	{
		LispSymbolDefinitionNode definition;
		if (type==LispAtomElement.DEFCLASS)
		{
			if (classDefinitions==null)
				classDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					classDefinitions.remove(this);
				}
			};
			classDefinitions.add(definition);			
		}
		else if (type==LispAtomElement.DEFCONSTANT)
		{
			if (constantDefinitions==null)
				constantDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					constantDefinitions.remove(this);
				}
			};
			constantDefinitions.add(definition);
		}
		else if (type==LispAtomElement.DEFGENERIC)
		{
			if (genericDefinitions==null)
				genericDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					genericDefinitions.remove(this);
				}
			};
			genericDefinitions.add(definition);
		}
		else if (type==LispAtomElement.DEFMACRO)
		{
			if (macroDefinitions==null)
				macroDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					macroDefinitions.remove(this);
				}
			};
			macroDefinitions.add(definition);
		}
		else if (type==LispAtomElement.DEFMETHOD)
		{
			if (methodDefinitions==null)
				methodDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					methodDefinitions.remove(this);
				}
			};
			methodDefinitions.add(definition);
		}
		else if (type==LispAtomElement.DEFPARAMETER)
		{
			if (parameterDefinitions==null)
				parameterDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					parameterDefinitions.remove(this);
				}
			};
			parameterDefinitions.add(definition);
		}
		else if (type==LispAtomElement.DEFSTRUCT)
		{
			if (structDefinitions==null)
				structDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					structDefinitions.remove(this);
				}
			};
			structDefinitions.add(definition);
		}
		else if (type==LispAtomElement.DEFTYPE)
		{
			if (typeDefinitions==null)
				typeDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					typeDefinitions.remove(this);
				}
			};
			typeDefinitions.add(definition);
		}
		else if (type==LispAtomElement.DEFVAR)
		{
			if (variableDefinitions==null)
				variableDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					variableDefinitions.remove(this);
				}
			};
			variableDefinitions.add(definition);
		}
		else if (type==LispAtomElement.DEFUN)
		{
			if (functionDefinitions==null)
				functionDefinitions=new ArrayList();
			definition=new LispSymbolDefinitionNode(this,type,begin,length)
			{
				public void undoDefinition() 
				{
					functionDefinitions.remove(this);
				}
			};
			functionDefinitions.add(definition);
		}
	}
	/* (non-Javadoc)
	 * @see java.lang.Comparable#compareTo(java.lang.Object)
	 */
	public int compareTo(Object o) 
	{
		LispSymbol symbol=(LispSymbol) o;
		int c=homepackage.compareTo(symbol.homepackage);
		if (c==0)
			return name.compareTo(symbol.name);
		else
			return c;
	}
	/**
	 * Gets the name of this symbol
	 * @return The name of the symbol
	 */
	public String getName() 
	{
		return name;
	}
	/**
	 * Gets the home package of this symbol
	 * @return The home package of this symbol
	 */
	public String getHomePackage()
	{
		return homepackage;
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getUniqueName()
	 */
	public String getUniqueName() 
	{
		return name;
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getDisplayName()
	 */
	public String getDisplayName() 
	{
		return name;
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getIconFile()
	 */
	public String getIconFile() 
	{
		return "big_defun.png";
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getPopupMenu()
	 */
	public JPopupMenu getPopupMenu() {
		// TODO Auto-generated method stub
		return null;
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#getTooltip()
	 */
	public String getTooltip() 
	{
		return homepackage+"::"+name;
	}
	/* (non-Javadoc)
	 * @see IDE.Project.INode#doubleClick()
	 */
	public void doubleClick() {
		// TODO Auto-generated method stub
	}
}
