/*
 * Created on Aug 18, 2003
 */
 
package IDE.Project;

import IDE.CodeEditor.*;
import IDE.CodeEditor.Parser.*;
import IDE.Utilities.*;
import IDE.WorkArea.WorkArea;

import java.util.*;
import java.io.*;
/**
 * @author marc
 * Class to hold all the repositories know
 */
public class Repositories
{
	/** Timer thread to scan files by all the repositories */
	private static Timer timer=new Timer();
	/** Set of opened douments */
	private static SortedMap pathToOpenDocuments=new TreeMap();
	/** List of all the registered repositories */
	private static SortedMap  repositories=new TreeMap();
	static 
	{
		// The LispRepository is always registered
		registerRepository(LispRepository.lisprepository);	
	}
	/** Avoid that a instance can be created */
	private Repositories()
	{
	}
	/**
	 * Returns the default file extension to use for the document
	 * @param doc The document where we search a file extension for
	 * @return The file extension for the document
	 */
	public static String getDefaultExtension(CodeDocument doc)
	{
		Parser parser=doc.getParser();
		if (parser==null)
			return ".lisp";
		else
			return parser.getRepository().getDefaultExtension();
	}
	/**
	 * Register a repository 
	 * @param repository The repository to register
	 */
	public static void registerRepository(AbstractRepository repository)
	{
		repositories.put(repository.getUniqueName(),repository);
		ProjectExplorer.addNode(new INode[] {repository});
	}
	/**
	 * Given a file creates a parser for this file based on the repository who first tells he can support this file
	 * @param file File to create a parser for
	 * @param async Create a async or sync parser
	 * @return The created parser
	 */
	public static Parser createParser(File file,boolean async)
	{
		if (file==null)
			return LispRepository.lisprepository.createParser(async);
		Iterator iter=repositories.values().iterator();
		IRepository repository;
		while (iter.hasNext())
		{
			repository=(IRepository) iter.next();
			if (repository.supportFile(Util.getCanonicalFileName(file)))
				return repository.createParser(async);
		}
		return LispRepository.lisprepository.createParser(async);
	}
	/**
	 * Checks if there is a repository supporting this file
	 * @param file File to check if its supported
	 * @return true if there exist a repository supporting this file
	 */
	public static boolean supportsFile(File file)
	{
		Iterator iter=repositories.values().iterator();
		while (iter.hasNext())
			if (((IRepository) iter.next()).supportFile(Util.getCanonicalFileName(file)))
				return true;
		return false;
	}
	
	/**
	 * Check if a document is already open and returns this document, otherwise 
	 * creates a new document and code pane containing the given file and returns the created document
	 * @param path The filename of that must be opened
	 * @return The open (opened) document
	 */
	public static CodeDocument openFile(String path)
	{
		CodeDocument doc=(CodeDocument) pathToOpenDocuments.get(path);
		if (doc==null)
		{
			File file=new File(path);
			CodePane pane=CodePane.createCodePane(createParser(file,true),file);
			WorkArea.addPane(pane);
			return pane.getEditor().getDocument();
		}
		else
		{
			return doc;
		}
	}
	/**
	 * Open the document (register it as a opened document)
	 * @param doc Document to open
	 */
	public static void openDocument(CodeDocument doc)
	{
		String path=Util.getCanonicalFileName(doc.getFile());
		if (pathToOpenDocuments.containsKey(path))
			return; // Document is already open
		// Open the document (the first adding of the document happens automatical by the first parse)
		pathToOpenDocuments.put(path,doc);
		
	}
	/**
	 * Close the document (removes it registration and close all file references of the file the document represents)
	 * @param doc Document to close
	 */
	public static void closeDocument(CodeDocument doc)
	{
		String path=Util.getCanonicalFileName(doc.getFile());
		pathToOpenDocuments.remove(path);
		if (Projects.isInProject(path))
		{
			// Ensure that no references to the document are left in the IDefinition elements
			FileReference.closeFile(path);
		}
		else 
		{
			// Remove all the references to the open file
			FileReference.removeFileReference(path);			
		}
	}
	/**
	 * Change the file associated with a document
	 * @param oldpath (old filename of the document)
	 * @param doc Document to changed
	 */
	public static void renameDocument(CodeDocument doc,File oldpath)
	{
		String path=Util.getCanonicalFileName(oldpath);
		// Remove all the old references
		pathToOpenDocuments.remove(path);
		FileReference.removeFileReference(path);
		// Add the new references
		path=Util.getCanonicalFileName(doc.getFile());
		pathToOpenDocuments.put(path,doc);
		scanFile(doc.getFile());
	}
	/**
	 * Checks if a file is opened in a document
	 * @param path Path name of the file to check
	 * @return true if the file is opened in a document false otherwise
	 */
	public static boolean isOpen(String path)
	{
		return pathToOpenDocuments.containsKey(path);
	}
	/**
	 * Try to find a open document containing the content of the file with pat path
	 * @param path The path of the file to search the open document for
	 * @return The open document found (or null if the file is not open)
	 */
	public static CodeDocument getOpenDocument(String path)
	{
		return (CodeDocument) pathToOpenDocuments.get(path);
	}
	/**
	 * Get a filter for the valid files that can be analyzed by the repositories
	 * @return
	 */
	public static javax.swing.filechooser.FileFilter getChooserFilter()
	{
		return new FileChooserFilter();
	}
	/**
	 * Get a filter for the file extensions
	 * @return
	 */
	public static javax.swing.filechooser.FileFilter getFileExtensionsFilter()
	{
		return new FileExtentionsFilter();
	}
	/**
	 * Scans the file on a Timer thread to 
	 * @param file
	 */
	public static void scanFile(final File file)
	{
		timer.schedule(new TimerTask()
		{
			public void run()
			{
				scanFileImp(file);
			}
		},10);
	}
	/**
	 * Do the actual work of scanning a file for all the repositories who support this
	 * @param file
	 */
	private static void scanFileImp(File file)
	{
		Iterator iter=repositories.values().iterator();
		IRepository repository;
		String path=Util.getCanonicalFileName(file);
		Parser parser;
		CodeDocument document;
		while (iter.hasNext())
		{
			repository=(IRepository) iter.next();
			if (repository.supportFile(path))
			{
				parser=repository.createParser(false);
				document=new CodeDocument(parser,file);
				FileReference.closeFile(path); // Close the open file
			};
		};
	}
	/**
	 * Do a scan of all the open documents
	 */
	public static void scanOpenDocuments()
	{
		Iterator iter=repositories.keySet().iterator();
		while (iter.hasNext())
			scanFile(new File((String) iter.next()));
	}
	/**
	 * Returns a array of all the extensions supported by the repositories
	 * @return A array of all the extensions supported by the repositories
	 */
	public static String[] getExtensions()
	{
		SortedSet extensions=new TreeSet();
		Iterator iter=repositories.values().iterator();
		String[] exts;
		while (iter.hasNext())
		{
			exts=((IRepository) iter.next()).getExtensions();
			for (int i=0;i<exts.length;i++)
				extensions.add(exts[i]);
		};
		String[] result=new String[extensions.size()];
		iter=extensions.iterator();
		int i=0;
		while (iter.hasNext())
		{
			result[i]=(String) iter.next();
			i++;
		};
		return result;
	}
}
/** Inner class to represent a filter to see only the files with known extensions (used in FileChooser */
class FileChooserFilter extends javax.swing.filechooser.FileFilter
{
	public boolean accept(File file)
	{
		if (file.isDirectory())
			return true;
		else if (file.isFile() && Repositories.supportsFile(file))
			return true;
		else
			return false;
	}
	public String getDescription()
	{
		// TODO Must be implemented
		return "This must still be implemented";
	}
}
/** Inner class to represent a filter to see only the files with the know extensions (used in FIle classes) */
class FileExtentionsFilter extends javax.swing.filechooser.FileFilter
{
	private HashSet seen;
       
	public FileExtentionsFilter()
	{
		seen=new HashSet();
	}
	public boolean accept(File file)
	{
		try
		{
			file=file.getCanonicalFile();
			if (!seen.contains(file))
			{
				seen.add(file);
				if (file.isDirectory())
					return true;
				else if (file.isFile() && Repositories.supportsFile(file))
					return true;
				else
					return false;
			}
			else
			{
				return false;
			}
		}
		catch (IOException e)
		{
			return false;
		}
	}
	/* (non-Javadoc)
	 * @see javax.swing.filechooser.FileFilter#getDescription()
	 */
	public String getDescription() 
	{
		// TODO Auto-generated method stub
		return "Must still be implemented";
	}
}	
