/*
 * Recording.java
 *
 * Created on May 4, 2002, 7:31 PM
 */

package IDE.Recording;

import java.awt.event.*;
import java.util.*;
import java.io.*;
import IDE.*;
import IDE.CodeEditor.*;
import IDE.Configuration.*;
import IDE.Messaging.*;

/**
 * Class containing static methods for the recording of actions performed
 * @author  marc
 */
public class Recording implements IIDEAddin
{
    /** Indicates that we are replaying */
    private static boolean replaying=false;
    /** Current set of macro */
    private static Macro currentMacro=null;
    /** Name to macro mapping */
    private static HashMap nameToMacro=new HashMap();
    /** Boolean variable indicating recording is taking place */
    private static boolean recording=false;
    /** Instance needed so main can call saveState() when program is exited */
    private static Recording self=null;
    /** Ensure no one can create instance directly */
    private Recording()
    {
        self=this;
    }
    /** Get the only instance of Recording */
    public static Recording getRecording()
    {
        if (self==null)
            return new Recording();
        else
            return self;
    }
    /** Method to start recording */
    public static void startRecording()
    {
        MessageHandler.say("Start recording");
        currentMacro=new Macro();
        recording=true;
    }
    /** Methods to stop recording */
    public static void stopRecording()
    {
        nameToMacro.put("*",currentMacro);
        currentMacro.setName("*");
        Main.getMacroExplorer().addMacroNode("*");
        recording=false;
    }
    public static void stopRecording(String name)
    {
        nameToMacro.put(name,currentMacro);
        currentMacro.setName(name);
        Main.getMacroExplorer().addMacroNode(name);
        recording=false;
    }
    /** Abort recording */
    public static void abortRecording()
    {
        currentMacro=null;
        recording=false;
    }
    /** Record the insertion of a DeadKey character */
    public static void recordInsertDeadKeyEvent(char unicodechar)
    {
        if (recording)
            currentMacro.recordInsertDeadKeyEvent(unicodechar);
    }
    
    /** Record a action and its event */
    public static void recordActionEvent(AbstractIDEAction action)
    {
        if (recording)
            currentMacro.recordActionEvent(action);
    }
    /** Record a question and its event */
    public static void recordQuestionEvent(IQuestion question,int keycode,int modifiers,String text,int type)
    {
        if (recording)
            currentMacro.recordQuestionEvent(question,keycode,modifiers,text,type);
    }
    /** Record a KeyEvent, because of a KeyPressed */
    public static void recordKeyEventPressed(KeyEvent ev)
    {
        if (recording)
            currentMacro.recordKeyEvent(ev.getID(),ev.getWhen(),ev.getModifiers(),ev.getKeyCode(),ev.getKeyChar(),ev.getKeyLocation(),RecordKeyEvent.KeyPressed);
    }
    /** Record a KeyEvent Because of a KeyType */
    public static void recordKeyEventTyped(KeyEvent ev)
    {
        if (recording)
            currentMacro.recordKeyEvent(ev.getID(),ev.getWhen(),ev.getModifiers(),ev.getKeyCode(),ev.getKeyChar(),ev.getKeyLocation(),RecordKeyEvent.KeyTyped);
    }
    /** Remove the Macro stored in the nameToMacro */
    public static void removeMacro(String name)
    {
        nameToMacro.remove(name);
    }
    /** Replay the recording */
    public static void replay()
    {
        CodeEditor current=Main.currentEditor();
        if (currentMacro !=null)
        {
            replaying=true;
            currentMacro.replay();
            replaying=false;
        }
        if (current!=null)
            current.grabFocus();
    }
    public static void replay(String name)
    {
        CodeEditor current=Main.currentEditor();
        Macro macro=(Macro) nameToMacro.get(name);
        if (macro!=null)
        {
            replaying=true;
            currentMacro=macro;
            currentMacro.replay();
            replaying=false;
        };
        if (current!=null)
            current.grabFocus();
    }
    /** Return true if replaying is active */
    public static boolean isReplaying()
    {
        return replaying;
    }
    /** Return the set of macro for macro with the given name */
    public static Macro getMacro(String name)
    {
        return (Macro) nameToMacro.get(name);
    }
    
    /**
     * This method will be called just before exit and allow the addin to save its state, the addin must be registered via a call
     * to Main.registerAddin()
     * Use the XMLConfigClass(Field,Method or Object) objects to serialize this class
     */
    public void saveState(Writer writer)
    {
        XMLConfigClass cls=new XMLConfigClass(self.getClass().getName());
        XMLConfigMethod method;
        nameToMacro.remove("*"); // Remove the currentMacro
        Iterator itermacro=nameToMacro.values().iterator();
        Iterator records;
        Object record;
        Macro macro;
        RecordActionEvent actionevent;
        RecordQuestionEvent questionevent;
        RecordKeyEvent keyevent;
        while (itermacro.hasNext())
        {
            macro=(Macro) itermacro.next();
            if (macro.getName()!=null)
            {
                records=macro.getRecordIterator();
                while (records.hasNext())
                {
                    record=records.next();
                    if (record instanceof RecordActionEvent)
                    {
                        actionevent=(RecordActionEvent) record;
                        method=new XMLConfigMethod("addActionRecord");
                        method.addArgument(macro.getName());
                        method.addArgument(Mode.getNameAction(actionevent.getAction()));
                        cls.addMethod(method);
                    }
                    else if (record instanceof RecordQuestionEvent)
                    {
                        questionevent=(RecordQuestionEvent) record;
                        method=new XMLConfigMethod("addQuestionRecord");
                        method.addArgument(macro.getName());
                        method.addArgument(InstructionField.getQuestionName(questionevent.getQuestion()));
                        method.addArgument(questionevent.getKeyCode());
                        method.addArgument(questionevent.getModifiers());
                        method.addArgument(questionevent.getText());
                        method.addArgument(questionevent.getType());
                        cls.addMethod(method);
                    }
                    else if (record instanceof RecordKeyEvent)
                    {
                        keyevent=(RecordKeyEvent) record;
                        method=new XMLConfigMethod("addKeyEventRecord");
                        method.addArgument(macro.getName());
                        method.addArgument(keyevent.getId());
                        method.addArgument(keyevent.getWhen());
                        method.addArgument(keyevent.getModifiers());
                        method.addArgument(keyevent.getKeycode());
                        method.addArgument(keyevent.getKeychar());
                        method.addArgument(keyevent.getKeylocation());
                        method.addArgument(keyevent.getType());
                        cls.addMethod(method);
                    }
                };
            }
        };
        method=new XMLConfigMethod("updateMacroExplorer");
        cls.addMethod(method);
        cls.serialize(writer);
    }
    /** Methods to be called by the deserialization code */
    /** Creates a Question record */
    public static void addQuestionRecord(String macroname,String questionname,int keycode,int modifiers,String text,int type)
    {
        Macro macro=(Macro) nameToMacro.get(macroname);
        if (macro==null)
        {
            macro=new Macro();
            nameToMacro.put(macroname,macro);
            macro.setName(macroname);
        };
        macro.recordQuestionEvent(InstructionField.getQuestion(questionname),keycode,modifiers,text,type);
    }
    /** Creates a action record */
    public static void addActionRecord(String macroname,String actionname)
    {
        Macro macro=(Macro) nameToMacro.get(macroname);
        AbstractManagedAction action=Mode.getAction(actionname);
        if (action instanceof AbstractIDEAction)
        {
            if (macro==null)
            {
                macro=new Macro();
                nameToMacro.put(macroname,macro);
                macro.setName(macroname);
            };
            macro.recordActionEvent((AbstractIDEAction) action);
        }
    }
    /** Creates a keyevent record  */
    public static void addKeyEventRecord(String macroname,int id,long when,int modifiers,int keycode,char keychar,int keylocation,int type)
    {
        Macro macro=(Macro) nameToMacro.get(macroname);
        if (macro==null)
        {
            macro=new Macro();
            nameToMacro.put(macroname,macro);
            macro.setName(macroname);
        };
        macro.recordKeyEvent(id,when,modifiers,keycode,keychar,keylocation,type);
    }
    /** Update the macro explorer */
    public static void updateMacroExplorer()
    {
        Iterator iter=nameToMacro.keySet().iterator();
        while (iter.hasNext())
        {
            Main.getMacroExplorer().addMacroNode((String) iter.next());
        }
    }
}
