/*
 * WatchPane.java
 *
 * Created on December 1, 2001, 1:56 PM
 */

package IDE.Watch;

import java.awt.*;
import java.awt.event.*;
import java.util.*;
import javax.swing.*;
import javax.swing.tree.*;
import IDE.Configuration.*;
import IDE.CodeEditor.*;
import IDE.CodeEditor.Parser.*;
import IDE.Interaction.*;

/**
 * Pane containing the result of watches
 * @author  marc
 */
public class Watches extends JPanel implements MouseListener
{
    private JTree tree;
    private DefaultTreeModel treemodel;
    private DefaultMutableTreeNode root;
    private JScrollPane scrollpane;
    private JPopupMenu topnodemenu;
    private JPopupMenu nodemenu;
    
    /** Creates new instance of a WatchPane */
    public Watches()
    {
        root=new DefaultMutableTreeNode("Watches");
        treemodel=new DefaultTreeModel(root);
        tree=new JTree(treemodel);
        tree.setCellRenderer(new WatchRenderer());
        scrollpane=new JScrollPane(tree);
        this.setLayout(new BorderLayout());
        this.add(scrollpane,BorderLayout.CENTER);
        tree.addMouseListener(this);
        // Create the popup menus
        createPopupMenus();
    }
    /** Repaint the tree if the fonts are changed */
    public void repaintAll()
    {
        treemodel.nodeStructureChanged(root);
        repaint();
    }
    /** Refresh the values of the watchpoints passed on */
    public static void refreshWatches(final LinkedList nodeResultList)
    {
        // Do all the work in a separate thread to avoid holding the lisp process
        Thread thread=new Thread(new Runnable()
        {
            public void run()
            {
                Iterator iter=nodeResultList.iterator();
                Object element;
                Object first,second;
                WatchNode watch;
                Object value;
                while (iter.hasNext())
                {
                    element=iter.next();
                    if (element instanceof LinkedList && ((LinkedList) element).size()==2)
                    {
                        first=((LinkedList) element).get(0);
                        second=((LinkedList) element).get(1);
                        if (first instanceof WatchNode)
                            refreshWatchNode((WatchNode) first,second instanceof LinkedList ? (LinkedList) second : null);
                    }
                }
            }
        });
        thread.run();
    }
    /** Refresh the node with the new expression  */
    private static void refreshWatchNode(final WatchNode node,final LinkedList list)
    {
        Runnable updateNode=new Runnable()
        {
            public void run()
            {
               node.refresh(list);
            }
        };
        SwingUtilities.invokeLater(updateNode);
    }
    /** Creates the popupmenus for the toplevel node and the watch nodes */
    private void createPopupMenus()
    {
        // Topnode menu
        topnodemenu=new JPopupMenu("Watches");
        topnodemenu.add(Mode.getAction("addwatch")).setIcon(null);
        //topnodemenu.add(Mode.getAction("refreshwatch")).setIcon(null);
        // Node menu
        nodemenu=new JPopupMenu("Watch");
        nodemenu.add(Mode.getAction("editwatch")).setIcon(null);
        nodemenu.add(Mode.getAction("removewatch")).setIcon(null);
        //nodemenu.add(Mode.getAction("refreshwatch")).setIcon(null);
    }
    /** Add a watch point for the value of a leaf */
    public void addWatchedLeaf(CodeEditor editor,IInteraction interaction)
    {
        CodeDocument document=editor.getDocument();
        Parser parser=document.getParser();
        String str=document.getLeafAt(editor.cursor.cursorOffset()).getString(document);
        if (str!=null)
        {
            WatchNode node=new WatchNode(str,str,interaction,parser.cloneParser(true),treemodel);
            tree.expandRow(0);
        };
    }
    /** Add a watch point to the watchpane (interactively) */
    public void addWatchPoint(CodeEditor editor,IInteraction interaction)
    {
        Parser parser=editor.getDocument().getParser();
        Region region=parser.findHooks(editor.cursor.cursorOffset());
        String str=(region==null) ? "" : editor.getDocument().getString(region.begin,region.end-region.begin+1) ;
        str=CodeDialogBox.showDialog("Watch Expression",parser.cloneParser(false),str);
        if (str!=null)
        {
            WatchNode node=new WatchNode(str,str,interaction,parser.cloneParser(true),treemodel);
            tree.expandRow(0);
        };
    }
    /** Add a watchpoint non interactively  (this watchpoint can not be modified or deleted, used to set predefined watch points) */
    public void addWatchPoint(String title,String watch,IInteraction interaction)
    {
        if (watch!=null)
        {
            WatchNode node=new WatchNode(title,watch,interaction,null,treemodel);
            tree.expandRow(0);
        };
    }
    /** Edit a watchpoint */
    public void editWatchPoint()
    {
        Object node=tree.getLastSelectedPathComponent();
        if (node instanceof WatchNode)
            ((WatchNode) node).editWatch();
    }
    /** Remove all the watchpoints using the interaction interaction */
    public void removeWatchPointsForInteraction(IInteraction interaction)
    {
        WatchNode node;
        int i=0;
        while (i<treemodel.getChildCount(root))
        {
            node=((WatchNode) treemodel.getChild(root,i));
            if (node.getInteraction()==interaction)
            {
                treemodel.removeNodeFromParent(node);
                node.cleanup();
            }
            else
            {
                i++;
            };
        };
    }
    /** Remove a watch point */
    public void removeWatchPoint()
    {
        Object node=tree.getLastSelectedPathComponent();
        if (node instanceof WatchNode)
        {
            treemodel.removeNodeFromParent((WatchNode) node);
            tree.repaint();
            ((WatchNode) node).cleanup();
        }
    }
    /** Check if we have a popup event for the tree and display then the popup menu */
    private void checkpopup(MouseEvent ev)
    {
        if (ev.isPopupTrigger())
        {
            tree.clearSelection();
            tree.addSelectionPath(tree.getClosestPathForLocation(ev.getX(),ev.getY()));
            Object node=tree.getLastSelectedPathComponent();
            if (node instanceof WatchNode)
                nodemenu.show(tree,ev.getX(),ev.getY());
            else
                topnodemenu.show(tree,ev.getX(),ev.getY());
        };
    }
    /** Listen to mouse events */
    public void mouseExited(MouseEvent ev)
    {
    }
    
    public void mouseReleased(MouseEvent ev)
    {
        checkpopup(ev);
    }
    
    public void mousePressed(MouseEvent ev)
    {
        checkpopup(ev);
    }
    
    public void mouseClicked(MouseEvent ev)
    {
        checkpopup(ev);
    }
    
    public void mouseEntered(java.awt.event.MouseEvent mouseEvent)
    {
    }
}
